@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp OpenSSL interface
- * Copyright (c) 2008-2010 TJ Saunders
+ * Copyright (c) 2008-2014 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,18 +14,19 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
  *
- * $Id: crypto.c,v 1.14.2.2 2010/05/18 21:43:35 castaglia Exp $
+ * $Id: crypto.c,v 1.33 2014-01-28 17:26:17 castaglia Exp $
  */
 
 #include "mod_sftp.h"
 #include "crypto.h"
+#include "umac.h"
 
 /* In OpenSSL 0.9.7, all des_ functions were renamed to DES_ to avoid 
  * clashes with older versions of libdes. 
@@ -60,6 +61,11 @@ struct sftp_cipher {
    * be explicitly requested via SFTPCiphers.
    */
   int enabled;
+
+  /* Is this cipher usable when FIPS is enabled?  If FALSE, then this
+   * cipher must NOT be advertised to clients in the KEXINIT.
+   */
+  int fips_allowed;
 };
 
 /* Currently, OpenSSL does NOT support AES CTR modes (not sure why).
@@ -78,25 +84,24 @@ static struct sftp_cipher ciphers[] = {
   /* The handling of NULL openssl_name and get_type fields is done in
    * sftp_crypto_get_cipher(), as special cases.
    */
-
 #if OPENSSL_VERSION_NUMBER > 0x000907000L
-  { "aes256-ctr",	NULL,		0,	NULL,			TRUE },
-  { "aes192-ctr",	NULL,		0,	NULL,			TRUE },
-  { "aes128-ctr",	NULL,		0,	NULL,			TRUE },
+  { "aes256-ctr",	NULL,		0,	NULL,	TRUE, TRUE },
+  { "aes192-ctr",	NULL,		0,	NULL,	TRUE, TRUE },
+  { "aes128-ctr",	NULL,		0,	NULL,	TRUE, TRUE },
 
 # ifndef HAVE_AES_CRIPPLED_OPENSSL
-  { "aes256-cbc",	"aes-256-cbc",	0,	EVP_aes_256_cbc,	TRUE },
-  { "aes192-cbc",	"aes-192-cbc",	0,	EVP_aes_192_cbc,	TRUE },
+  { "aes256-cbc",	"aes-256-cbc",	0,	EVP_aes_256_cbc, TRUE, TRUE },
+  { "aes192-cbc",	"aes-192-cbc",	0,	EVP_aes_192_cbc, TRUE, TRUE },
 # endif /* !HAVE_AES_CRIPPLED_OPENSSL */
 
-  { "aes128-cbc",	"aes-128-cbc",	0,	EVP_aes_128_cbc,	TRUE },
+  { "aes128-cbc",	"aes-128-cbc",	0,	EVP_aes_128_cbc, TRUE, TRUE },
 #endif
 
-  { "blowfish-ctr",	NULL,		0,	NULL,			TRUE },
-  { "blowfish-cbc",	"bf-cbc",	0,	EVP_bf_cbc,		TRUE },
-  { "cast128-cbc",	"cast5-cbc",	0,	EVP_cast5_cbc,		TRUE },
-  { "arcfour256",	"rc4",		1536,	EVP_rc4,		TRUE },
-  { "arcfour128",	"rc4",		1536,	EVP_rc4,		TRUE },
+  { "blowfish-ctr",	NULL,		0,	NULL,	TRUE, FALSE },
+  { "blowfish-cbc",	"bf-cbc",	0,	EVP_bf_cbc, TRUE, FALSE },
+  { "cast128-cbc",	"cast5-cbc",	0,	EVP_cast5_cbc, TRUE, FALSE },
+  { "arcfour256",	"rc4",		1536,	EVP_rc4, TRUE, FALSE },
+  { "arcfour128",	"rc4",		1536,	EVP_rc4, TRUE, FALSE },
 
 #if 0
   /* This cipher is explicitly NOT supported because it does not discard
@@ -106,13 +111,13 @@ static struct sftp_cipher ciphers[] = {
    * require explicit configuration via SFTPCiphers, and would generate
    * warnings about its unsafe use.
    */
-  { "arcfour",		"rc4",		0,	EVP_rc4,		FALSE },
+  { "arcfour",		"rc4",		0,	EVP_rc4, FALSE, FALSE },
 #endif
 
-  { "3des-ctr",		NULL,		0,	NULL,			TRUE },
-  { "3des-cbc",		"des-ede3-cbc",	0,	EVP_des_ede3_cbc,	TRUE },
-  { "none",		"null",		0,	EVP_enc_null,		FALSE },
-  { NULL, NULL, 0, NULL, FALSE }
+  { "3des-ctr",		NULL,		0,	NULL, TRUE, TRUE },
+  { "3des-cbc",		"des-ede3-cbc",	0,	EVP_des_ede3_cbc, TRUE, TRUE },
+  { "none",		"null",		0,	EVP_enc_null, FALSE, TRUE },
+  { NULL, NULL, 0, NULL, FALSE, FALSE }
 };
 
 struct sftp_digest {
@@ -131,16 +136,33 @@ struct sftp_digest {
    * explicitly requested via SFTPDigests.
    */
   int enabled;
+
+  /* Is this MAC usable when FIPS is enabled?  If FALSE, then this digest must
+   * NOT be advertised to clients in the KEXINIT.
+   */
+  int fips_allowed;
 };
 
 static struct sftp_digest digests[] = {
-  { "hmac-sha1",	"sha1",		EVP_sha1,	0,	TRUE },
-  { "hmac-sha1-96",	"sha1",		EVP_sha1,	12,	TRUE },
-  { "hmac-md5",		"md5",		EVP_md5,	0,	TRUE },
-  { "hmac-md5-96",	"md5",		EVP_md5,	12,	TRUE },
-  { "hmac-ripemd160",	"rmd160",	EVP_ripemd160,	0,	TRUE },
-  { "none",		"null",		EVP_md_null,	0,	FALSE },
-  { NULL, NULL, NULL, 0, FALSE }
+  /* The handling of NULL openssl_name and get_type fields is done in
+   * sftp_crypto_get_digest(), as special cases.
+   */
+#ifdef HAVE_SHA256_OPENSSL
+  { "hmac-sha2-256",	"sha256",		EVP_sha256,	0, TRUE, TRUE },
+#endif /* SHA256 support in OpenSSL */
+#ifdef HAVE_SHA512_OPENSSL
+  { "hmac-sha2-512",	"sha512",		EVP_sha512,	0, TRUE, TRUE },
+#endif /* SHA512 support in OpenSSL */
+  { "hmac-sha1",	"sha1",		EVP_sha1,	0, 	TRUE, TRUE },
+  { "hmac-sha1-96",	"sha1",		EVP_sha1,	12,	TRUE, TRUE },
+  { "hmac-md5",		"md5",		EVP_md5,	0,	TRUE, FALSE },
+  { "hmac-md5-96",	"md5",		EVP_md5,	12,	TRUE, FALSE },
+  { "hmac-ripemd160",	"rmd160",	EVP_ripemd160,	0,	TRUE, FALSE },
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+  { "umac-64@openssh.com", NULL,	NULL,		8,	TRUE, FALSE },
+#endif /* OpenSSL-0.9.7 or later */
+  { "none",		"null",		EVP_md_null,	0,	FALSE, TRUE },
+  { NULL, NULL, NULL, 0, FALSE, FALSE }
 };
 
 static const char *trace_channel = "ssh2";
@@ -173,7 +195,7 @@ static int init_bf_ctr(EVP_CIPHER_CTX *ctx, const unsigned char *key,
     /* Allocate our data structure. */
     bce = calloc(1, sizeof(struct bf_ctr_ex));
     if (bce == NULL) {
-      pr_log_pri(PR_LOG_ERR, MOD_SFTP_VERSION ": Out of memory!");
+      pr_log_pri(PR_LOG_ALERT, MOD_SFTP_VERSION ": Out of memory!");
       _exit(1);
     }
 
@@ -216,7 +238,7 @@ static int cleanup_bf_ctr(EVP_CIPHER_CTX *ctx) {
 }
 
 static int do_bf_ctr(EVP_CIPHER_CTX *ctx, unsigned char *dst,
-    const unsigned char *src, unsigned int len) {
+    const unsigned char *src, size_t len) {
   struct bf_ctr_ex *bce;
   unsigned int n;
   unsigned char buf[BF_BLOCK];
@@ -280,7 +302,7 @@ static const EVP_CIPHER *get_bf_ctr_cipher(void) {
 
   memset(&bf_ctr_cipher, 0, sizeof(EVP_CIPHER));
 
-  bf_ctr_cipher.nid = NID_undef;
+  bf_ctr_cipher.nid = NID_bf_cbc;
   bf_ctr_cipher.block_size = BF_BLOCK;
   bf_ctr_cipher.iv_len = BF_BLOCK;
   bf_ctr_cipher.key_len = 32;
@@ -293,6 +315,8 @@ static const EVP_CIPHER *get_bf_ctr_cipher(void) {
   return &bf_ctr_cipher;
 }
 
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+
 /* 3DES CTR mode implementation */
 
 struct des3_ctr_ex {
@@ -322,7 +346,7 @@ static int init_des3_ctr(EVP_CIPHER_CTX *ctx, const unsigned char *key,
     /* Allocate our data structure. */
     dce = calloc(1, sizeof(struct des3_ctr_ex));
     if (dce == NULL) {
-      pr_log_pri(PR_LOG_ERR, MOD_SFTP_VERSION ": Out of memory!");
+      pr_log_pri(PR_LOG_ALERT, MOD_SFTP_VERSION ": Out of memory!");
       _exit(1);
     }
 
@@ -370,7 +394,7 @@ static int cleanup_des3_ctr(EVP_CIPHER_CTX *ctx) {
 }
 
 static int do_des3_ctr(EVP_CIPHER_CTX *ctx, unsigned char *dst,
-    const unsigned char *src, unsigned int len) {
+    const unsigned char *src, size_t len) {
   struct des3_ctr_ex *dce;
   unsigned int n;
   unsigned char buf[8];
@@ -427,7 +451,7 @@ static const EVP_CIPHER *get_des3_ctr_cipher(void) {
 
   memset(&des3_ctr_cipher, 0, sizeof(EVP_CIPHER));
 
-  des3_ctr_cipher.nid = NID_undef;
+  des3_ctr_cipher.nid = NID_des_ede3_ecb;
   des3_ctr_cipher.block_size = 8;
   des3_ctr_cipher.iv_len = 8;
   des3_ctr_cipher.key_len = 24;
@@ -436,12 +460,13 @@ static const EVP_CIPHER *get_des3_ctr_cipher(void) {
   des3_ctr_cipher.do_cipher = do_des3_ctr;
 
   des3_ctr_cipher.flags = EVP_CIPH_CBC_MODE|EVP_CIPH_VARIABLE_LENGTH|EVP_CIPH_ALWAYS_CALL_INIT|EVP_CIPH_CUSTOM_IV;
+#ifdef OPENSSL_FIPS
+  des3_ctr_cipher.flags |= EVP_CIPH_FLAG_FIPS;
+#endif /* OPENSSL_FIPS */
 
   return &des3_ctr_cipher;
 }
 
-#if OPENSSL_VERSION_NUMBER > 0x000907000L
-
 /* AES CTR mode implementation */
 struct aes_ctr_ex {
   AES_KEY key;
@@ -460,7 +485,7 @@ static int init_aes_ctr(EVP_CIPHER_CTX *ctx, const unsigned char *key,
     /* Allocate our data structure. */
     ace = calloc(1, sizeof(struct aes_ctr_ex));
     if (ace == NULL) {
-      pr_log_pri(PR_LOG_ERR, MOD_SFTP_VERSION ": Out of memory!");
+      pr_log_pri(PR_LOG_ALERT, MOD_SFTP_VERSION ": Out of memory!");
       _exit(1);
     }
 
@@ -503,7 +528,7 @@ static int cleanup_aes_ctr(EVP_CIPHER_CTX *ctx) {
 }
 
 static int do_aes_ctr(EVP_CIPHER_CTX *ctx, unsigned char *dst,
-    const unsigned char *src, unsigned int len) {
+    const unsigned char *src, size_t len) {
   struct aes_ctr_ex *ace;
 # if OPENSSL_VERSION_NUMBER <= 0x0090704fL
   unsigned int n;
@@ -560,7 +585,54 @@ static const EVP_CIPHER *get_aes_ctr_cipher(int key_len) {
 
   memset(&aes_ctr_cipher, 0, sizeof(EVP_CIPHER));
 
+#ifdef OPENSSL_FIPS
+  /* Set the NID depending on the key len. */
+  switch (key_len) {
+    case 16:
+      aes_ctr_cipher.nid = NID_aes_128_cbc;
+      break;
+
+    case 24:
+      aes_ctr_cipher.nid = NID_aes_192_cbc;
+      break;
+
+    case 32:
+      aes_ctr_cipher.nid = NID_aes_256_cbc;
+      break;
+
+    default:
+      aes_ctr_cipher.nid = NID_undef;
+  }
+#else
+  /* Setting this nid member to something other than NID_undef causes
+   * interesting problems on an OpenSolaris system, using the provided
+   * OpenSSL installation's pkcs11 engine via:
+   *
+   *  SFTPCryptoDevice pkcs11
+   *
+   * for the mod_sftp config.  I'm not sure why; I need to look into this
+   * issue more.
+   *
+   * For posterity, the issues seen when using the above config are
+   * described below.  After sending the NEWKEYS request, mod_sftp
+   * would log the following, upon receiving the next message from sftp(1):
+   *
+   *  <ssh2:20>: SSH2 packet len = 1500737511 bytes
+   *  <ssh2:20>: SSH2 packet padding len = 95 bytes
+   *  <ssh2:20>: SSH2 packet payload len = 1500737415 bytes
+   *  <ssh2:20>: payload len (1500737415 bytes) exceeds max payload len (262144), ignoring payload
+   *  client sent buggy/malicious packet payload length, ignoring
+   *
+   * and sftp(1), for its side, would report:
+   *
+   *  debug1: send SSH2_MSG_SERVICE_REQUEST
+   *  Disconnecting: Bad packet length.
+   *  debug1: Calling cleanup 0x807cc14(0x0)
+   *  Couldn't read packet: Error 0
+   */
   aes_ctr_cipher.nid = NID_undef;
+#endif /* OPENSSL_FIPS */
+
   aes_ctr_cipher.block_size = AES_BLOCK_SIZE;
   aes_ctr_cipher.iv_len = AES_BLOCK_SIZE;
   aes_ctr_cipher.key_len = key_len;
@@ -569,9 +641,66 @@ static const EVP_CIPHER *get_aes_ctr_cipher(int key_len) {
   aes_ctr_cipher.do_cipher = do_aes_ctr;
 
   aes_ctr_cipher.flags = EVP_CIPH_CBC_MODE|EVP_CIPH_VARIABLE_LENGTH|EVP_CIPH_ALWAYS_CALL_INIT|EVP_CIPH_CUSTOM_IV;
+#ifdef OPENSSL_FIPS
+  aes_ctr_cipher.flags |= EVP_CIPH_FLAG_FIPS;
+#endif /* OPENSSL_FIPS */
 
   return &aes_ctr_cipher;
 }
+
+static int update_umac(EVP_MD_CTX *ctx, const void *data, size_t len) {
+  int res;
+
+  if (ctx->md_data == NULL) {
+    struct umac_ctx *umac;
+
+    umac = umac_new((unsigned char *) data);
+    if (umac == NULL) {
+      return 0;
+    }
+
+    ctx->md_data = umac;
+    return 1;
+  }
+
+  res = umac_update(ctx->md_data, (unsigned char *) data, (long) len);
+  return res;
+}
+
+static int final_umac(EVP_MD_CTX *ctx, unsigned char *md) {
+  unsigned char nonce[8];
+  int res;
+
+  res = umac_final(ctx->md_data, md, nonce);
+  return res;
+}
+
+static int delete_umac(EVP_MD_CTX *ctx) {
+  struct umac_ctx *umac;
+
+  umac = ctx->md_data;
+  umac_delete(umac);
+  ctx->md_data = NULL;
+
+  return 1;
+}
+
+static const EVP_MD *get_umac_digest(void) {
+  static EVP_MD umac_digest;
+
+  memset(&umac_digest, 0, sizeof(EVP_MD));
+
+  umac_digest.type = NID_undef;
+  umac_digest.pkey_type = NID_undef;
+  umac_digest.md_size = 8;
+  umac_digest.flags = 0UL;
+  umac_digest.update = update_umac;
+  umac_digest.final = final_umac;
+  umac_digest.cleanup = delete_umac;
+  umac_digest.block_size = 32;
+
+  return &umac_digest;
+}
 #endif /* OpenSSL older than 0.9.7 */
 
 const EVP_CIPHER *sftp_crypto_get_cipher(const char *name, size_t *key_len,
@@ -582,20 +711,20 @@ const EVP_CIPHER *sftp_crypto_get_cipher(const char *name, size_t *key_len,
     if (strcmp(ciphers[i].name, name) == 0) {
       const EVP_CIPHER *cipher;
 
-      if (strcmp(name, "blowfish-ctr") == 0) {
+      if (strncmp(name, "blowfish-ctr", 13) == 0) {
         cipher = get_bf_ctr_cipher();
 
-      } else if (strcmp(name, "3des-ctr") == 0) {
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+      } else if (strncmp(name, "3des-ctr", 9) == 0) {
         cipher = get_des3_ctr_cipher();
 
-#if OPENSSL_VERSION_NUMBER > 0x000907000L
-      } else if (strcmp(name, "aes256-ctr") == 0) {
+      } else if (strncmp(name, "aes256-ctr", 11) == 0) {
         cipher = get_aes_ctr_cipher(32);
 
-      } else if (strcmp(name, "aes192-ctr") == 0) {
+      } else if (strncmp(name, "aes192-ctr", 11) == 0) {
         cipher = get_aes_ctr_cipher(24);
 
-      } else if (strcmp(name, "aes128-ctr") == 0) {
+      } else if (strncmp(name, "aes128-ctr", 11) == 0) {
         cipher = get_aes_ctr_cipher(16);
 #endif /* OpenSSL older than 0.9.7 */
 
@@ -604,7 +733,7 @@ const EVP_CIPHER *sftp_crypto_get_cipher(const char *name, size_t *key_len,
       }
 
       if (key_len) {
-        if (strcmp(name, "arcfour256") != 0) {
+        if (strncmp(name, "arcfour256", 11) != 0) {
           *key_len = 0;
 
         } else {
@@ -633,7 +762,19 @@ const EVP_MD *sftp_crypto_get_digest(const char *name, uint32_t *mac_len) {
 
   for (i = 0; digests[i].name; i++) {
     if (strcmp(digests[i].name, name) == 0) {
-      const EVP_MD *digest = digests[i].get_type();
+      const EVP_MD *digest = NULL;
+
+#if OPENSSL_VERSION_NUMBER > 0x000907000L
+      if (strncmp(name, "umac-64@openssh.com", 12) == 0) {
+        digest = get_umac_digest();
+#else
+      if (FALSE) {
+#endif /* OpenSSL older than 0.9.7 */
+
+      } else {
+        digest = digests[i].get_type();
+      }
+
       if (mac_len) {
         *mac_len = digests[i].mac_len;
       }
@@ -665,7 +806,20 @@ const char *sftp_crypto_get_kexinit_cipher_list(pool *p) {
 
       for (j = 0; ciphers[j].name; j++) {
         if (strcmp(c->argv[i], ciphers[j].name) == 0) {
-          if (strcmp(c->argv[i], "none") != 0) {
+#ifdef OPENSSL_FIPS
+          if (FIPS_mode()) {
+            /* If FIPS mode is enabled, check whether the cipher is allowed
+             * for use.
+             */
+            if (ciphers[j].fips_allowed == FALSE) {
+              pr_trace_msg(trace_channel, 5,
+                "cipher '%s' is disabled in FIPS mode, skipping",
+                ciphers[j].name);
+              continue;
+            }
+          }
+#endif /* OPENSSL_FIPS */
+          if (strncmp(c->argv[i], "none", 5) != 0) {
             if (EVP_get_cipherbyname(ciphers[j].openssl_name) != NULL) {
               res = pstrcat(p, res, *res ? "," : "",
                 pstrdup(p, ciphers[j].name), NULL);
@@ -673,12 +827,12 @@ const char *sftp_crypto_get_kexinit_cipher_list(pool *p) {
             } else {
               /* The CTR modes are special cases. */
 
-              if (strcmp(ciphers[j].name, "blowfish-ctr") == 0 ||
-                  strcmp(ciphers[j].name, "3des-ctr") == 0
+              if (strncmp(ciphers[j].name, "blowfish-ctr", 13) == 0 ||
+                  strncmp(ciphers[j].name, "3des-ctr", 9) == 0
 #if OPENSSL_VERSION_NUMBER > 0x000907000L
-                  || strcmp(ciphers[j].name, "aes256-ctr") == 0 ||
-                  strcmp(ciphers[j].name, "aes192-ctr") == 0 ||
-                  strcmp(ciphers[j].name, "aes128-ctr") == 0
+                  || strncmp(ciphers[j].name, "aes256-ctr", 11) == 0 ||
+                  strncmp(ciphers[j].name, "aes192-ctr", 11) == 0 ||
+                  strncmp(ciphers[j].name, "aes128-ctr", 11) == 0
 #endif
                   ) {
                 res = pstrcat(p, res, *res ? "," : "",
@@ -704,7 +858,21 @@ const char *sftp_crypto_get_kexinit_cipher_list(pool *p) {
 
     for (i = 0; ciphers[i].name; i++) {
       if (ciphers[i].enabled) {
-        if (strcmp(ciphers[i].name, "none") != 0) {
+#ifdef OPENSSL_FIPS
+          if (FIPS_mode()) {
+            /* If FIPS mode is enabled, check whether the cipher is allowed
+             * for use.
+             */
+            if (ciphers[i].fips_allowed == FALSE) {
+              pr_trace_msg(trace_channel, 5,
+                "cipher '%s' is disabled in FIPS mode, skipping",
+                ciphers[i].name);
+              continue;
+            }
+          }
+#endif /* OPENSSL_FIPS */
+
+        if (strncmp(ciphers[i].name, "none", 5) != 0) {
           if (EVP_get_cipherbyname(ciphers[i].openssl_name) != NULL) {
             res = pstrcat(p, res, *res ? "," : "",
               pstrdup(p, ciphers[i].name), NULL);
@@ -712,12 +880,12 @@ const char *sftp_crypto_get_kexinit_cipher_list(pool *p) {
           } else {
             /* The CTR modes are special cases. */
 
-            if (strcmp(ciphers[i].name, "blowfish-ctr") == 0 ||
-                strcmp(ciphers[i].name, "3des-ctr") == 0
+            if (strncmp(ciphers[i].name, "blowfish-ctr", 13) == 0 ||
+                strncmp(ciphers[i].name, "3des-ctr", 9) == 0
 #if OPENSSL_VERSION_NUMBER > 0x000907000L
-                || strcmp(ciphers[i].name, "aes256-ctr") == 0 ||
-                strcmp(ciphers[i].name, "aes192-ctr") == 0 ||
-                strcmp(ciphers[i].name, "aes128-ctr") == 0
+                || strncmp(ciphers[i].name, "aes256-ctr", 11) == 0 ||
+                strncmp(ciphers[i].name, "aes192-ctr", 11) == 0 ||
+                strncmp(ciphers[i].name, "aes128-ctr", 11) == 0
 #endif
                 ) {
               res = pstrcat(p, res, *res ? "," : "",
@@ -763,15 +931,37 @@ const char *sftp_crypto_get_kexinit_digest_list(pool *p) {
 
       for (j = 0; digests[j].name; j++) {
         if (strcmp(c->argv[i], digests[j].name) == 0) {
-          if (strcmp(c->argv[i], "none") != 0) {
-            if (EVP_get_digestbyname(digests[j].openssl_name) != NULL) {
+#ifdef OPENSSL_FIPS
+          if (FIPS_mode()) {
+            /* If FIPS mode is enabled, check whether the MAC is allowed
+             * for use.
+             */
+            if (digests[j].fips_allowed == FALSE) {
+              pr_trace_msg(trace_channel, 5,
+                "digest '%s' is disabled in FIPS mode, skipping",
+                digests[j].name);
+              continue;
+            }
+          }
+#endif /* OPENSSL_FIPS */
+
+          if (strncmp(c->argv[i], "none", 5) != 0) {
+            if (digests[j].openssl_name != NULL &&
+                EVP_get_digestbyname(digests[j].openssl_name) != NULL) {
               res = pstrcat(p, res, *res ? "," : "",
                 pstrdup(p, digests[j].name), NULL);
 
             } else {
-              pr_trace_msg(trace_channel, 3,
-                "unable to use '%s' digest: Unsupported by OpenSSL",
-                digests[j].name);
+              /* The umac-64 digest is a special case. */
+              if (strncmp(digests[j].name, "umac-64@openssh.com", 12) == 0) {
+                res = pstrcat(p, res, *res ? "," : "",
+                  pstrdup(p, digests[j].name), NULL);
+
+              } else {
+                pr_trace_msg(trace_channel, 3,
+                  "unable to use '%s' digest: Unsupported by OpenSSL",
+                  digests[j].name);
+              }
             }
 
           } else {
@@ -787,15 +977,37 @@ const char *sftp_crypto_get_kexinit_digest_list(pool *p) {
 
     for (i = 0; digests[i].name; i++) {
       if (digests[i].enabled) {
-        if (strcmp(digests[i].name, "none") != 0) {
-          if (EVP_get_digestbyname(digests[i].openssl_name) != NULL) {
+#ifdef OPENSSL_FIPS
+          if (FIPS_mode()) {
+            /* If FIPS mode is enabled, check whether the digest is allowed
+             * for use.
+             */
+            if (digests[i].fips_allowed == FALSE) {
+              pr_trace_msg(trace_channel, 5,
+                "digest '%s' is disabled in FIPS mode, skipping",
+                digests[i].name);
+              continue;
+            }
+          }
+#endif /* OPENSSL_FIPS */
+
+        if (strncmp(digests[i].name, "none", 5) != 0) {
+          if (digests[i].openssl_name != NULL &&
+              EVP_get_digestbyname(digests[i].openssl_name) != NULL) {
             res = pstrcat(p, res, *res ? "," : "",
               pstrdup(p, digests[i].name), NULL);
 
           } else {
-            pr_trace_msg(trace_channel, 3,
-              "unable to use '%s' digest: Unsupported by OpenSSL",
-              digests[i].name);
+            /* The umac-64 digest is a special case. */
+            if (strncmp(digests[i].name, "umac-64@openssh.com", 12) == 0) {
+              res = pstrcat(p, res, *res ? "," : "",
+                pstrdup(p, digests[i].name), NULL);
+
+            } else {
+              pr_trace_msg(trace_channel, 3,
+                "unable to use '%s' digest: Unsupported by OpenSSL",
+                digests[i].name);
+            }
           }
 
         } else {
@@ -880,7 +1092,18 @@ void sftp_crypto_free(int flags) {
 #endif
 
     ERR_free_strings();
+
+#if OPENSSL_VERSION_NUMBER >= 0x10000001L
+    /* The ERR_remove_state(0) usage is deprecated due to thread ID
+     * differences among platforms; see the OpenSSL-1.0.0c CHANGES file
+     * for details.  So for new enough OpenSSL installations, use the
+     * proper way to clear the error queue state.
+     */
+    ERR_remove_thread_state(NULL);
+#else
     ERR_remove_state(0);
+#endif /* OpenSSL prior to 1.0.0-beta1 */
+
     EVP_cleanup();
     RAND_cleanup();
   }
@@ -895,7 +1118,7 @@ int sftp_crypto_set_driver(const char *driver) {
 
   crypto_engine = driver;
 
-  if (strcasecmp(driver, "ALL") == 0) {
+  if (strncasecmp(driver, "ALL", 4) == 0) {
     /* Load all ENGINE implementations bundled with OpenSSL. */
     ENGINE_load_builtin_engines();
     ENGINE_register_all_complete();
