@@ -1,6 +1,6 @@
 /*
  * ProFTPD - mod_sftp packet IO
- * Copyright (c) 2008-2010 TJ Saunders
+ * Copyright (c) 2008-2013 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,14 +14,14 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
  *
- * $Id: packet.c,v 1.14 2010/02/15 22:03:52 castaglia Exp $
+ * $Id: packet.c,v 1.46 2013-03-08 16:22:18 castaglia Exp $
  */
 
 #include "mod_sftp.h"
@@ -43,17 +43,29 @@
 
 extern pr_response_t *resp_list, *resp_err_list;
 
+extern module sftp_module;
+
 static uint32_t packet_client_seqno = 0;
 static uint32_t packet_server_seqno = 0;
 
+/* Maximum length of the payload data of an SSH2 packet we're willing to
+ * accept.  Any packets reporting a payload length longer than this will be
+ * ignored/dropped.
+ */
+#define SFTP_PACKET_MAX_PAYLOAD_LEN	(256 * 1024)
+
 /* RFC4344 recommends 2^31 for the client packet sequence number at which
  * we should request a rekey, and 2^32 for the server packet sequence number.
- * Since we're using uin32_t, though, it isn't a big enough data type for those
- * numbers.  We'll settle for the max value for the uint32_t type.
+ *
+ * However, the uint32_t data type may not be unsigned for some
+ * platforms/compilers.  To avoid issues on these platforms, use 2^31-1
+ * for rekeying for both client and server packet sequence numbers.
  */
-#define SFTP_PACKET_REKEY_SEQNO_LIMIT		(uint32_t) -1
-static uint32_t rekey_client_seqno = SFTP_PACKET_REKEY_SEQNO_LIMIT;
-static uint32_t rekey_server_seqno = SFTP_PACKET_REKEY_SEQNO_LIMIT;
+#define SFTP_PACKET_CLIENT_REKEY_SEQNO_LIMIT		2147483647
+#define SFTP_PACKET_SERVER_REKEY_SEQNO_LIMIT		2147483647
+
+static uint32_t rekey_client_seqno = SFTP_PACKET_CLIENT_REKEY_SEQNO_LIMIT;
+static uint32_t rekey_server_seqno = SFTP_PACKET_SERVER_REKEY_SEQNO_LIMIT;
 
 static off_t rekey_client_len = 0;
 static off_t rekey_server_len = 0;
@@ -62,25 +74,52 @@ static off_t rekey_size = 0;
 static int poll_timeout = -1;
 static time_t last_recvd, last_sent;
 
-static const char *version_id = SFTP_ID_STRING "\r\n";
+static const char *server_version = SFTP_ID_DEFAULT_STRING;
+static const char *version_id = SFTP_ID_DEFAULT_STRING "\r\n";
 static int sent_version_id = FALSE;
 
+static void is_client_alive(void);
+
+/* Count of the number of "client alive" messages sent without a response. */
+static unsigned int client_alive_max = 0, client_alive_count = 0;
+static unsigned int client_alive_interval = 0;
+
 static const char *trace_channel = "ssh2";
 
 static int packet_poll(int sockfd, int io) {
   fd_set rfds, wfds;
   struct timeval tv;
-  int res, timeout;
+  int res, timeout, using_client_alive = FALSE;
 
-  timeout = poll_timeout;
-  if (timeout <= 0) {
-    /* Default of 60 seconds */
-    timeout = 60;
+  if (poll_timeout == -1) {
+    /* If we have "client alive" timeout interval configured, use that --
+     * but only if we have already done the key exchange, and are not
+     * rekeying.
+     *
+     * Otherwise, we use the default (i.e. TimeoutIdle).
+     */
+
+    if (client_alive_interval > 0 &&
+        (!(sftp_sess_state & SFTP_SESS_STATE_REKEYING) && 
+         (sftp_sess_state & SFTP_SESS_STATE_HAVE_AUTH))) {
+      timeout = client_alive_interval;
+      using_client_alive = TRUE;
+
+    } else {
+      timeout = pr_data_get_timeout(PR_DATA_TIMEOUT_IDLE);
+    }
+
+  } else {
+    timeout = poll_timeout;
   }
 
   tv.tv_sec = timeout;
   tv.tv_usec = 0;
 
+  pr_trace_msg(trace_channel, 19,
+    "waiting for max of %lu secs while polling socket %d using select(2)",
+    (unsigned long) tv.tv_sec, sockfd);
+
   while (1) {
     pr_signals_handle();
 
@@ -106,20 +145,34 @@ static int packet_poll(int sockfd, int io) {
     }
 
     if (res < 0) {
-      if (errno == EINTR) {
+      int xerrno = errno;
+
+      if (xerrno == EINTR) {
         pr_signals_handle();
         continue;
       }
 
+      pr_trace_msg(trace_channel, 18, "error calling select(2) on fd %d: %s",
+        sockfd, strerror(xerrno));
+
+      errno = xerrno;
       return -1;
 
     } else if (res == 0) {
       tv.tv_sec = timeout;
       tv.tv_usec = 0;
 
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "polling on socket %d timed out after %lu sec, trying again", sockfd,
-        (unsigned long) tv.tv_sec);
+      if (using_client_alive) {
+        is_client_alive();
+
+      } else {
+        pr_trace_msg(trace_channel, 18,
+          "polling on socket %d timed out after %lu sec, trying again", sockfd,
+          (unsigned long) tv.tv_sec);
+        (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+          "polling on socket %d timed out after %lu sec, trying again", sockfd,
+          (unsigned long) tv.tv_sec);
+      }
 
       continue;
     }
@@ -138,7 +191,8 @@ static int packet_poll(int sockfd, int io) {
  * It is the caller's responsibility to ensure that buf is large enough to
  * hold reqlen bytes.
  */
-int sftp_ssh2_packet_sock_read(int sockfd, void *buf, size_t reqlen) {
+int sftp_ssh2_packet_sock_read(int sockfd, void *buf, size_t reqlen,
+    int flags) {
   void *ptr;
   size_t remainlen;
 
@@ -165,13 +219,19 @@ int sftp_ssh2_packet_sock_read(int sockfd, void *buf, size_t reqlen) {
 
     while (res <= 0) {
       if (res < 0) {
-        if (errno == EINTR) {
+        int xerrno = errno;
+
+        if (xerrno == EINTR) {
           pr_signals_handle();
           continue;
         }
 
+        pr_trace_msg(trace_channel, 16,
+          "error reading from client (fd %d): %s", sockfd, strerror(xerrno));
         (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-          "error reading from client (fd %d): %s", sockfd, strerror(errno));
+          "error reading from client (fd %d): %s", sockfd, strerror(xerrno));
+
+        errno = xerrno;
 
         /* We explicitly disconnect the client here, rather than sending
          * a DISCONNECT message, because the errors below all indicate
@@ -190,9 +250,14 @@ int sftp_ssh2_packet_sock_read(int sockfd, void *buf, size_t reqlen) {
             errno == ESHUTDOWN ||
 #endif /* ESHUTDOWNN */
             errno == EPIPE) {
+          xerrno = errno;
+
+          pr_trace_msg(trace_channel, 16,
+            "disconnecting client (%s)", strerror(xerrno));
           (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-            "disconnecting client (%s)", strerror(errno));
-          end_login(1);
+            "disconnecting client (%s)", strerror(xerrno));
+          pr_session_disconnect(&sftp_module, PR_SESS_DISCONNECT_CLIENT_EOF,
+            strerror(xerrno));
         }
 
         return -1;
@@ -202,17 +267,35 @@ int sftp_ssh2_packet_sock_read(int sockfd, void *buf, size_t reqlen) {
          * the uncommunicative client.
          */
 
+        pr_trace_msg(trace_channel, 16, "%s",
+          "disconnecting client (received EOF)");
         (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
           "disconnecting client (received EOF)");
-        end_login(1);
+        pr_session_disconnect(&sftp_module, PR_SESS_DISCONNECT_CLIENT_EOF,
+          NULL);
       }
     }
 
+    /* Generate an event for any interested listeners.  Since the data are
+     * probably encrypted and such, and since listeners won't/shouldn't
+     * have the facilities for handling such data, we only pass the
+     * amount of data read in.
+     */
+    pr_event_generate("ssh2.netio-read", &res);
+
+    session.total_raw_in += reqlen;
     time(&last_recvd);
 
     if (res == remainlen)
       break;
 
+    if (flags & SFTP_PACKET_READ_FL_PESSIMISTIC) {
+      pr_trace_msg(trace_channel, 20, "read %lu bytes, expected %lu bytes; "
+        "pessimistically returning", (unsigned long) res,
+        (unsigned long) remainlen);
+      break;
+    }
+
     pr_trace_msg(trace_channel, 20, "read %lu bytes, expected %lu bytes; "
       "reading more", (unsigned long) res, (unsigned long) remainlen);
     ptr = ((char *) ptr + res);
@@ -222,61 +305,15 @@ int sftp_ssh2_packet_sock_read(int sockfd, void *buf, size_t reqlen) {
   return reqlen;
 }
 
-static void handle_debug_mesg(struct ssh2_packet *pkt) {
-  register unsigned int i;
-  char always_display;
-  char *str;
+static char peek_mesg_type(struct ssh2_packet *pkt) {
+  char mesg_type;
 
-  always_display = sftp_msg_read_bool(pkt->pool, &pkt->payload,
-    &pkt->payload_len);
-  str = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
-
-  /* Ignore the language tag. */
-  (void) sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
-
-  /* Sanity-check the message for control (and other non-printable)
-   * characters.
-   */
-  for (i = 0; i < strlen(str); i++) {
-    if (iscntrl((int) str[i]) ||
-        !isprint((int) str[i])) {
-      str[i] = '?';
-    }
-  }
-
-  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-    "client sent SSH_MSG_DEBUG message '%s'", str);
-
-  if (always_display) {
-    pr_log_debug(DEBUG0, MOD_SFTP_VERSION
-      ": client sent SSH_MSG_DEBUG message '%s'", str);
-  }
-}
-
-static void handle_disconnect_mesg(struct ssh2_packet *pkt) {
-  register unsigned int i;
-  char *explain;
-  uint32_t reason;
-
-  reason = sftp_msg_read_int(pkt->pool, &pkt->payload, &pkt->payload_len);
-  explain = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
-  (void) sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
-
-  /* Sanity-check the message for control characters. */
-  for (i = 0; i < strlen(explain); i++) {
-    if (iscntrl((int) explain[i])) {
-      explain[i] = '?';
-    }
-  }
-
-  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-    "client sent SSH_DISCONNECT message '%s' (%lu)", explain,
-    (unsigned long) reason);
-  end_login(1);
+  memmove(&mesg_type, pkt->payload, sizeof(char));
+  return mesg_type;
 }
 
 static void handle_global_request_mesg(struct ssh2_packet *pkt) {
-  char *buf, *ptr;
+  unsigned char *buf, *ptr;
   uint32_t buflen, bufsz;
   char *request_name;
   int want_reply;
@@ -308,29 +345,75 @@ static void handle_global_request_mesg(struct ssh2_packet *pkt) {
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "error writing REQUEST_FAILURE message: %s", strerror(errno));
     }
-
-    destroy_pool(pkt2->pool);
   }
+
+  destroy_pool(pkt->pool);
 }
 
-static void handle_ignore_mesg(struct ssh2_packet *pkt) {
-  char *str;
-  size_t len;
+static void handle_client_alive_mesg(struct ssh2_packet *pkt, char mesg_type) {
+  const char *mesg_desc;
 
-  str = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
-  len = strlen(str);
+  mesg_desc = sftp_ssh2_packet_get_mesg_type_desc(mesg_type);
 
-  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-    "client sent SSH_MSG_IGNORE message (%u bytes)", (unsigned int) len);
+  pr_trace_msg(trace_channel, 12,
+    "client sent %s message, considering client alive", mesg_desc);
+
+  client_alive_count = 0;
+  destroy_pool(pkt->pool);
 }
 
-static void handle_unimplemented_mesg(struct ssh2_packet *pkt) {
-  uint32_t seqno;
+static void is_client_alive(void) {
+  unsigned int count;
+  unsigned char *buf, *ptr;
+  uint32_t bufsz, buflen, channel_id;
+  struct ssh2_packet *pkt;
+  pool *tmp_pool;
 
-  seqno = sftp_msg_read_int(pkt->pool, &pkt->payload, &pkt->payload_len);
+  if (++client_alive_count > client_alive_max) {
+    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+      "SFTPClientAlive threshold (max %u checks, %u sec interval) reached, "
+      "disconnecting client", client_alive_max, client_alive_interval);    
 
-  pr_trace_msg(trace_channel, 7, "received SSH_MSG_UNIMPLEMENTED for "
-    "packet #%lu", (unsigned long) seqno);
+    /* XXX Generate an event for this? */
+
+    SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_BY_APPLICATION,
+      "client alive threshold reached");
+  }
+
+  /* If we have any opened channels, send a CHANNEL_REQUEST to one of
+   * them.  Otherwise, send a GLOBAL_REQUEST.
+   */
+
+  tmp_pool = make_sub_pool(session.pool);
+
+  bufsz = buflen = 64;
+  ptr = buf = palloc(tmp_pool, bufsz);
+
+  count = sftp_channel_opened(&channel_id);  
+  if (count > 0) {
+    sftp_msg_write_byte(&buf, &buflen, SFTP_SSH2_MSG_CHANNEL_REQUEST);
+    sftp_msg_write_int(&buf, &buflen, channel_id);
+
+    pr_trace_msg(trace_channel, 9,
+      "sending CHANNEL_REQUEST (remote channel ID %lu, keepalive@proftpd.org)",
+      (unsigned long) channel_id);
+
+  } else {
+    sftp_msg_write_byte(&buf, &buflen, SFTP_SSH2_MSG_GLOBAL_REQUEST);
+
+    pr_trace_msg(trace_channel, 9,
+      "sending GLOBAL_REQUEST (keepalive@proftpd.org)");
+  }
+
+  sftp_msg_write_string(&buf, &buflen, "keepalive@proftpd.org");
+  sftp_msg_write_bool(&buf, &buflen, TRUE);
+
+  pkt = sftp_ssh2_packet_create(tmp_pool);
+  pkt->payload = ptr;
+  pkt->payload_len = (bufsz - buflen);
+
+  (void) sftp_ssh2_packet_write(sftp_conn->wfd, pkt);
+  destroy_pool(tmp_pool);
 }
 
 /* Attempt to read in a random amount of data (up to the maximum amount of
@@ -352,18 +435,23 @@ static void read_packet_discard(int sockfd) {
     (unsigned long) buflen);
 
   if (buflen > 0) {
-    sftp_ssh2_packet_sock_read(sockfd, buf, buflen);
+    int flags = SFTP_PACKET_READ_FL_PESSIMISTIC;
+
+    /* We don't necessarily want to wait for the entire random amount of data
+     * to be read in.
+     */
+    sftp_ssh2_packet_sock_read(sockfd, buf, buflen, flags);
   }
 
   return;
 }
 
 static int read_packet_len(int sockfd, struct ssh2_packet *pkt,
-    char *buf, size_t *offset, size_t *buflen, size_t bufsz) {
+    unsigned char *buf, size_t *offset, size_t *buflen, size_t bufsz) {
   uint32_t packet_len = 0, len = 0;
   size_t blocksz; 
   int res;
-  char *ptr = NULL;
+  unsigned char *ptr = NULL;
 
   blocksz = sftp_cipher_get_block_size();
 
@@ -372,17 +460,16 @@ static int read_packet_len(int sockfd, struct ssh2_packet *pkt,
    * how many more bytes there are in the packet.
    */
 
-  res = sftp_ssh2_packet_sock_read(sockfd, buf, blocksz);
+  res = sftp_ssh2_packet_sock_read(sockfd, buf, blocksz, 0);
   if (res < 0)
     return res;
 
   len = res;
-  if (sftp_cipher_read_data(pkt->pool, (unsigned char *) buf, blocksz, &ptr,
-      &len) < 0) {
+  if (sftp_cipher_read_data(pkt->pool, buf, blocksz, &ptr, &len) < 0) {
     return -1;
   }
 
-  memcpy(&packet_len, ptr, sizeof(uint32_t));
+  memmove(&packet_len, ptr, sizeof(uint32_t));
   pkt->packet_len = ntohl(packet_len);
 
   ptr += sizeof(uint32_t);
@@ -392,7 +479,7 @@ static int read_packet_len(int sockfd, struct ssh2_packet *pkt,
    * buffer.
    */
   if (len > 0) {
-    memcpy(buf, ptr, len);
+    memmove(buf, ptr, len);
     *buflen = (size_t) len;
   }
 
@@ -401,11 +488,11 @@ static int read_packet_len(int sockfd, struct ssh2_packet *pkt,
 }
 
 static int read_packet_padding_len(int sockfd, struct ssh2_packet *pkt,
-    char *buf, size_t *offset, size_t *buflen, size_t bufsz) {
+    unsigned char *buf, size_t *offset, size_t *buflen, size_t bufsz) {
 
   if (*buflen > sizeof(char)) {
     /* XXX Assume the data in the buffer is unecrypted, and thus usable. */
-    memcpy(&pkt->padding_len, buf + *offset, sizeof(char));
+    memmove(&pkt->padding_len, buf + *offset, sizeof(char));
 
     /* Advance the buffer past the byte we just read off. */
     *offset += sizeof(char);
@@ -421,8 +508,8 @@ static int read_packet_padding_len(int sockfd, struct ssh2_packet *pkt,
 }
 
 static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
-    char *buf, size_t *offset, size_t *buflen, size_t bufsz) {
-  char *ptr = NULL;
+    unsigned char *buf, size_t *offset, size_t *buflen, size_t bufsz) {
+  unsigned char *ptr = NULL;
   int res;
   uint32_t payload_len = pkt->payload_len, padding_len = pkt->padding_len,
     data_len, len = 0;
@@ -430,8 +517,26 @@ static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
   if (payload_len + padding_len == 0)
     return 0;
 
-  if (payload_len > 0)
+  if (payload_len > 0) {
+    /* We don't want to reject the packet outright yet; but we can ignore
+     * the payload data we're going to read in.  This packet will fail
+     * eventually anyway.
+     */
+    if (payload_len > SFTP_PACKET_MAX_PAYLOAD_LEN) {
+      pr_trace_msg(trace_channel, 20,
+        "payload len (%lu bytes) exceeds max payload len (%lu), "
+        "ignoring payload", (unsigned long) payload_len,
+        (unsigned long) SFTP_PACKET_MAX_PAYLOAD_LEN);
+
+      pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "client sent buggy/malicious packet payload length, ignoring");
+
+      errno = EPERM;
+      return -1;
+    }
+
     pkt->payload = pcalloc(pkt->pool, payload_len);
+  }
 
   /* If there's data in the buffer we received, it's probably already part
    * of the payload, unencrypted.  That will leave the remaining payload
@@ -439,7 +544,7 @@ static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
    */
   if (*buflen > 0) {
     if (*buflen < payload_len) {
-      memcpy(pkt->payload, buf + *offset, *buflen);
+      memmove(pkt->payload, buf + *offset, *buflen);
 
       payload_len -= *buflen;
       *offset = 0;
@@ -447,7 +552,7 @@ static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
 
     } else {
       /* There's enough already for the payload length.  Nice. */
-      memcpy(pkt->payload, buf + *offset, payload_len);
+      memmove(pkt->payload, buf + *offset, payload_len);
 
       *offset += payload_len;
       *buflen -= payload_len;
@@ -464,7 +569,7 @@ static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
    */
   if (*buflen > 0) {
     if (*buflen < padding_len) {
-      memcpy(pkt->padding, buf + *offset, *buflen);
+      memmove(pkt->padding, buf + *offset, *buflen);
 
       padding_len -= *buflen;
       *offset = 0;
@@ -472,7 +577,7 @@ static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
 
     } else {
       /* There's enough already for the padding length.  Nice. */
-      memcpy(pkt->padding, buf + *offset, padding_len);
+      memmove(pkt->padding, buf + *offset, padding_len);
 
       *offset += padding_len;
       *buflen -= padding_len;
@@ -492,56 +597,50 @@ static int read_packet_payload(int sockfd, struct ssh2_packet *pkt,
     return -1;
   }
 
-  res = sftp_ssh2_packet_sock_read(sockfd, buf + *offset, data_len);
+  res = sftp_ssh2_packet_sock_read(sockfd, buf + *offset, data_len, 0);
   if (res < 0) {
     return res;
   }
  
   len = res;
-  if (sftp_cipher_read_data(pkt->pool, (unsigned char *) buf + *offset,
-      data_len, &ptr, &len) < 0) {
+  if (sftp_cipher_read_data(pkt->pool, buf + *offset, data_len, &ptr,
+      &len) < 0) {
     return -1;
   }
 
   if (payload_len > 0) {
-    memcpy(pkt->payload + (pkt->payload_len - payload_len), ptr,
+    memmove(pkt->payload + (pkt->payload_len - payload_len), ptr,
       payload_len);
   }
 
-  memcpy(pkt->padding + (pkt->padding_len - padding_len), ptr + payload_len,
+  memmove(pkt->padding + (pkt->padding_len - padding_len), ptr + payload_len,
     padding_len);
   return 0;
 }
 
-static int read_packet_mac(int sockfd, struct ssh2_packet *pkt, char *buf) {
+static int read_packet_mac(int sockfd, struct ssh2_packet *pkt,
+    unsigned char *buf) {
   int res;
   uint32_t mac_len = pkt->mac_len;
 
   if (mac_len == 0)
     return 0;
 
-  res = sftp_ssh2_packet_sock_read(sockfd, buf, mac_len);
+  res = sftp_ssh2_packet_sock_read(sockfd, buf, mac_len, 0);
   if (res < 0)
     return res;
 
   pkt->mac = palloc(pkt->pool, pkt->mac_len);
-  memcpy(pkt->mac, buf, res);
+  memmove(pkt->mac, buf, res);
 
   return 0;
 }
 
-static char peek_mesg_type(struct ssh2_packet *pkt) {
-  char mesg_type;
-
-  memcpy(&mesg_type, pkt->payload, sizeof(char));
-  return mesg_type;
-}
-
 struct ssh2_packet *sftp_ssh2_packet_create(pool *p) {
   pool *tmp_pool;
   struct ssh2_packet *pkt;
 
-  tmp_pool = pr_pool_create_sz(p, 128);
+  tmp_pool = make_sub_pool(p);
   pr_pool_tag(tmp_pool, "SSH2 packet pool");
 
   pkt = pcalloc(tmp_pool, sizeof(struct ssh2_packet));
@@ -560,7 +659,7 @@ int sftp_ssh2_packet_get_last_recvd(time_t *tp) {
     return -1;
   }
 
-  memcpy(tp, &last_recvd, sizeof(time_t));
+  memmove(tp, &last_recvd, sizeof(time_t));
   return 0;
 }
 
@@ -570,21 +669,21 @@ int sftp_ssh2_packet_get_last_sent(time_t *tp) {
     return -1;
   }
 
-  memcpy(tp, &last_sent, sizeof(time_t));
+  memmove(tp, &last_sent, sizeof(time_t));
   return 0;
 }
 
 char sftp_ssh2_packet_get_mesg_type(struct ssh2_packet *pkt) {
   char mesg_type;
 
-  memcpy(&mesg_type, pkt->payload, sizeof(char));
+  memmove(&mesg_type, pkt->payload, sizeof(char));
   pkt->payload += sizeof(char);
   pkt->payload_len -= sizeof(char);
 
   return mesg_type;
 }
 
-const char *sftp_ssh2_packet_get_mesg_type_desc(char mesg_type) {
+const char *sftp_ssh2_packet_get_mesg_type_desc(unsigned char mesg_type) {
   switch (mesg_type) {
     case SFTP_SSH2_MSG_DISCONNECT:
       return "SSH_MSG_DISCONNECT";
@@ -691,22 +790,30 @@ const char *sftp_ssh2_packet_get_mesg_type_desc(char mesg_type) {
 
 int sftp_ssh2_packet_set_poll_timeout(int timeout) {
   if (timeout <= 0) {
-    timeout = pr_data_get_timeout(PR_DATA_TIMEOUT_IDLE);
+    poll_timeout = -1;
+
+  } else {
+    poll_timeout = timeout;
   }
 
-  poll_timeout = timeout;
+  return 0;
+}
+
+int sftp_ssh2_packet_set_client_alive(unsigned int max, unsigned int interval) {
+  client_alive_max = max;
+  client_alive_interval = interval;
   return 0;
 }
 
 int sftp_ssh2_packet_read(int sockfd, struct ssh2_packet *pkt) {
-  char buf[SFTP_MAX_PACKET_LEN];
+  unsigned char buf[SFTP_MAX_PACKET_LEN];
   size_t buflen, bufsz = SFTP_MAX_PACKET_LEN, offset = 0;
-  char mesg_type;
 
   pr_session_set_idle();
 
   while (1) {
     uint32_t req_blocksz;
+    int res;
 
     pr_signals_handle();
 
@@ -748,7 +855,6 @@ int sftp_ssh2_packet_read(int sockfd, struct ssh2_packet *pkt) {
       return -1;
     }
 
-
     pr_trace_msg(trace_channel, 20, "SSH2 packet padding len = %u bytes",
       (unsigned int) pkt->padding_len);
 
@@ -877,66 +983,17 @@ int sftp_ssh2_packet_read(int sockfd, struct ssh2_packet *pkt) {
 
     packet_client_seqno++;
 
-    pr_timer_reset(PR_TIMER_IDLE, ANY_MODULE);
+    /* Handle the case where timers might be being processed at the
+     * moment.
+     */
+    res = pr_timer_reset(PR_TIMER_IDLE, ANY_MODULE);
+    while (res < 0) {
+      if (errno == EINTR) {
+        pr_signals_handle();
+        res = pr_timer_reset(PR_TIMER_IDLE, ANY_MODULE);
+      }
 
-    mesg_type = peek_mesg_type(pkt);
-
-    pr_trace_msg(trace_channel, 3, "received %s (%d) packet",
-      sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type);
-
-    if (mesg_type == SFTP_SSH2_MSG_DEBUG) {
-      /* Since peeking at the mesg type did not actually update the
-       * payload/len values, do that now.
-       */
-      pkt->payload += sizeof(char);
-      pkt->payload_len -= sizeof(char);
-
-      handle_debug_mesg(pkt);
-      continue;
-    }
-
-    if (mesg_type == SFTP_SSH2_MSG_DISCONNECT) {
-      /* Since peeking at the mesg type did not actually update the
-       * payload/len values, do that now.
-       */
-      pkt->payload += sizeof(char);
-      pkt->payload_len -= sizeof(char);
-
-      handle_disconnect_mesg(pkt);
-      continue;
-    }
-
-    if (mesg_type == SFTP_SSH2_MSG_GLOBAL_REQUEST) {
-      /* Since peeking at the mesg type did not actually update the
-       * payload/len values, do that now.
-       */
-      pkt->payload += sizeof(char);
-      pkt->payload_len -= sizeof(char);
-
-      handle_global_request_mesg(pkt);
-      continue;
-    }
-
-    if (mesg_type == SFTP_SSH2_MSG_IGNORE) {
-      /* Since peeking at the mesg type did not actually update the
-       * payload/len values, do that now.
-       */
-      pkt->payload += sizeof(char);
-      pkt->payload_len -= sizeof(char);
-
-      handle_ignore_mesg(pkt);
-      continue;
-    }
-
-    if (mesg_type == SFTP_SSH2_MSG_UNIMPLEMENTED) {
-      /* Since peeking at the mesg type did not actually update the
-       * payload/len values, do that now.
-       */
-      pkt->payload += sizeof(char);
-      pkt->payload_len -= sizeof(char);
-
-      handle_unimplemented_mesg(pkt);
-      continue;
+      break;
     }
 
     break;
@@ -1009,33 +1066,21 @@ static int write_packet_padding(struct ssh2_packet *pkt) {
   return 0;
 }
 
-/* Reserve space for TWO packets: one potential TAP packet, and one
- * definite packet (i.e. the given packet to be sent).
- */
 #define SFTP_SSH2_PACKET_IOVSZ		12
 static struct iovec packet_iov[SFTP_SSH2_PACKET_IOVSZ];
 static unsigned int packet_niov = 0;
 
-int sftp_ssh2_packet_write(int sockfd, struct ssh2_packet *pkt) {
-  char buf[SFTP_MAX_PACKET_LEN * 2], mesg_type;
+int sftp_ssh2_packet_send(int sockfd, struct ssh2_packet *pkt) {
+  unsigned char buf[SFTP_MAX_PACKET_LEN * 2], mesg_type;
   size_t buflen = 0, bufsz = SFTP_MAX_PACKET_LEN;
   uint32_t packet_len = 0;
-  int res;
+  int res, write_len = 0;
 
   /* Clear the iovec array before sending the data, if possible. */
   if (packet_niov == 0) {
     memset(packet_iov, 0, sizeof(packet_iov));
   }
 
-  if (sent_version_id) {
-    res = sftp_tap_send_packet();
-    if (res < 0) {
-      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "error sending TAP packet: %s", strerror(errno));
-      return res;
-    }
-  }
-
   mesg_type = peek_mesg_type(pkt);
 
   if (sftp_compress_write_data(pkt) < 0) {
@@ -1071,16 +1116,19 @@ int sftp_ssh2_packet_write(int sockfd, struct ssh2_packet *pkt) {
     if (!sent_version_id) {
       packet_iov[packet_niov].iov_base = (void *) version_id;
       packet_iov[packet_niov].iov_len = strlen(version_id);
+      write_len += packet_iov[packet_niov].iov_len;
       packet_niov++;
     }
 
     packet_iov[packet_niov].iov_base = (void *) buf;
     packet_iov[packet_niov].iov_len = buflen;
+    write_len += packet_iov[packet_niov].iov_len;
     packet_niov++;
 
     if (pkt->mac_len > 0) {
       packet_iov[packet_niov].iov_base = (void *) pkt->mac;
       packet_iov[packet_niov].iov_len = pkt->mac_len;
+      write_len += packet_iov[packet_niov].iov_len;
       packet_niov++;
     }
 
@@ -1093,42 +1141,64 @@ int sftp_ssh2_packet_write(int sockfd, struct ssh2_packet *pkt) {
     if (!sent_version_id) {
       packet_iov[packet_niov].iov_base = (void *) version_id;
       packet_iov[packet_niov].iov_len = strlen(version_id);
+      write_len += packet_iov[packet_niov].iov_len;
       packet_niov++;
     }
 
     packet_iov[packet_niov].iov_base = (void *) &packet_len;
     packet_iov[packet_niov].iov_len = sizeof(uint32_t);
+    write_len += packet_iov[packet_niov].iov_len;
     packet_niov++;
 
     packet_iov[packet_niov].iov_base = (void *) &(pkt->padding_len);
     packet_iov[packet_niov].iov_len = sizeof(char);
+    write_len += packet_iov[packet_niov].iov_len;
     packet_niov++;
 
     packet_iov[packet_niov].iov_base = (void *) pkt->payload;
     packet_iov[packet_niov].iov_len = pkt->payload_len;
+    write_len += packet_iov[packet_niov].iov_len;
     packet_niov++;
 
     packet_iov[packet_niov].iov_base = (void *) pkt->padding;
     packet_iov[packet_niov].iov_len = pkt->padding_len;
+    write_len += packet_iov[packet_niov].iov_len;
     packet_niov++;
 
     if (pkt->mac_len > 0) {
       packet_iov[packet_niov].iov_base = (void *) pkt->mac;
       packet_iov[packet_niov].iov_len = pkt->mac_len;
+      write_len += packet_iov[packet_niov].iov_len;
       packet_niov++;
     }
   }
 
   if (packet_poll(sockfd, SFTP_PACKET_IO_WR) < 0) {
+    int xerrno = errno;
+
+    /* Socket not writable?  Clear the array, and try again. */
+    memset(packet_iov, 0, sizeof(packet_iov));
+    packet_niov = 0;
+
+    errno = xerrno;
     return -1;
   }
 
+  /* Generate an event for any interested listeners.  Since the data are
+   * probably encrypted and such, and since listeners won't/shouldn't
+   * have the facilities for handling such data, we only pass the
+   * amount of data to be written out.
+   */
+  pr_event_generate("ssh2.netio-write", &write_len);
+
   /* The socket we accept is blocking, thus there's no need to handle
    * EAGAIN/EWOULDBLOCK errors.
    */
   res = writev(sockfd, packet_iov, packet_niov);
   while (res < 0) {
-    if (errno == EINTR) {
+    int xerrno = errno;
+
+    if (xerrno == EINTR) {
       pr_signals_handle();
 
       res = writev(sockfd, packet_iov, packet_niov);
@@ -1136,28 +1206,38 @@ int sftp_ssh2_packet_write(int sockfd, struct ssh2_packet *pkt) {
     }
 
     (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-      "error writing packet (fd %d): %s", sockfd, strerror(errno));
+      "error writing packet (fd %d): %s", sockfd, strerror(xerrno));
+
+    if (xerrno == ECONNRESET ||
+        xerrno == ECONNABORTED ||
+        xerrno == EPIPE) {
 
-    if (errno == ECONNRESET ||
-        errno == ECONNABORTED ||
-        errno == EPIPE) {
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
-        "disconnecting client (%s)", strerror(errno));
-      end_login(1);
+        "disconnecting client (%s)", strerror(xerrno));
+      pr_session_disconnect(&sftp_module, PR_SESS_DISCONNECT_BY_APPLICATION,
+        strerror(xerrno));
     }
 
     /* Always clear the iovec array after sending the data. */
     memset(packet_iov, 0, sizeof(packet_iov));
     packet_niov = 0;
 
+    errno = xerrno;
     return -1;
   }
 
+  session.total_raw_out += res;
+
   /* Always clear the iovec array after sending the data. */
   memset(packet_iov, 0, sizeof(packet_iov));
   packet_niov = 0;
 
-  sent_version_id = TRUE;
+  if (sent_version_id == FALSE) {
+    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+      "sent server version '%s'", server_version);
+    sent_version_id = TRUE;
+  }
+
   time(&last_sent);
 
   packet_server_seqno++;
@@ -1181,12 +1261,141 @@ int sftp_ssh2_packet_write(int sockfd, struct ssh2_packet *pkt) {
     sftp_kex_rekey();
   }
 
-  pr_trace_msg(trace_channel, 3, "sent %s (%d) packet",
-    sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type);
+  pr_trace_msg(trace_channel, 3, "sent %s (%d) packet (%d bytes)",
+    sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type, res);
  
   return 0;
 }
 
+int sftp_ssh2_packet_write(int sockfd, struct ssh2_packet *pkt) {
+  int res;
+
+  /* For future reference, I tried buffering up the possible TAP message
+   * and the given message using TCP_CORK/TCP_NOPUSH.  I tried this test
+   * on a Mac OSX 10.5 box.  It was for this test that I refactored the
+   * core pr_inet_* functions, creating pr_inet_set_proto_cork().
+   *
+   * Turned out to be a very bad idea.  Using sftp(1), the login process
+   * itself was incredibly slow.  mod_sftp was behaving as if the
+   * "uncorking" call was not causing any flushing of the partially-full
+   * network buffer to be written out, and instead was waiting until the
+   * buffer was full before writing it out.
+   *
+   * Now, this could be due to a bug in Mac OSX's handling of TCP_NOPUSH;
+   * there are articles about such a problem in earlier Mac OSX versions.
+   * I should try this test again, on a Linux (TCP_CORK) and FreeBSD
+   * (TCP_NOPUSH), to see if this can be done at least on those platforms.
+   */
+
+  if (sent_version_id) {
+    res = sftp_tap_send_packet();
+    if (res < 0) {
+      (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+        "error sending TAP packet: %s", strerror(errno));
+    }
+  }
+
+  return sftp_ssh2_packet_send(sockfd, pkt);
+}
+
+void sftp_ssh2_packet_handle_debug(struct ssh2_packet *pkt) {
+  register unsigned int i;
+  char always_display;
+  char *str;
+
+  always_display = sftp_msg_read_bool(pkt->pool, &pkt->payload,
+    &pkt->payload_len);
+  str = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
+
+  /* Ignore the language tag. */
+  (void) sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
+
+  /* Sanity-check the message for control (and other non-printable)
+   * characters.
+   */
+  for (i = 0; i < strlen(str); i++) {
+    if (PR_ISCNTRL(str[i]) ||
+        !PR_ISPRINT(str[i])) {
+      str[i] = '?';
+    }
+  }
+
+  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+    "client sent SSH_MSG_DEBUG message '%s'", str);
+
+  if (always_display) {
+    pr_log_debug(DEBUG0, MOD_SFTP_VERSION
+      ": client sent SSH_MSG_DEBUG message '%s'", str);
+  }
+
+  destroy_pool(pkt->pool);
+}
+
+void sftp_ssh2_packet_handle_disconnect(struct ssh2_packet *pkt) {
+  register unsigned int i;
+  char *explain = NULL, *lang = NULL;
+  const char *reason_str = NULL;
+  uint32_t reason_code;
+
+  reason_code = sftp_msg_read_int(pkt->pool, &pkt->payload, &pkt->payload_len);
+  reason_str = sftp_disconnect_get_str(reason_code);
+  if (reason_str == NULL) {
+    pr_trace_msg(trace_channel, 9,
+      "client sent unknown disconnect reason code %lu",
+      (unsigned long) reason_code);
+    reason_str = "Unknown reason code";
+  }
+
+  explain = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
+
+  /* Not all clients send a language tag. */
+  if (pkt->payload_len > 0) {
+    lang = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
+  }
+
+  /* Sanity-check the message for control characters. */
+  for (i = 0; i < strlen(explain); i++) {
+    if (PR_ISCNTRL(explain[i])) {
+      explain[i] = '?';
+    }
+  }
+
+  /* XXX Use the language tag somehow, if provided. */
+  if (lang != NULL) {
+    pr_trace_msg(trace_channel, 19, "client sent DISCONNECT language tag '%s'",
+      lang);
+  }
+
+  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+    "client at %s sent SSH_DISCONNECT message: %s (%s)",
+    pr_netaddr_get_ipstr(session.c->remote_addr), explain, reason_str);
+  pr_session_disconnect(&sftp_module, PR_SESS_DISCONNECT_CLIENT_QUIT, explain);
+}
+
+void sftp_ssh2_packet_handle_ignore(struct ssh2_packet *pkt) {
+  char *str;
+  size_t len;
+
+  str = sftp_msg_read_string(pkt->pool, &pkt->payload, &pkt->payload_len);
+  len = strlen(str);
+
+  (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+    "client sent SSH_MSG_IGNORE message (%u bytes)", (unsigned int) len);
+
+  destroy_pool(pkt->pool);
+}
+
+void sftp_ssh2_packet_handle_unimplemented(struct ssh2_packet *pkt) {
+  uint32_t seqno;
+
+  seqno = sftp_msg_read_int(pkt->pool, &pkt->payload, &pkt->payload_len);
+
+  pr_trace_msg(trace_channel, 7, "received SSH_MSG_UNIMPLEMENTED for "
+    "packet #%lu", (unsigned long) seqno);
+
+  destroy_pool(pkt->pool);
+}
+
 int sftp_ssh2_packet_handle(void) {
   struct ssh2_packet *pkt;
   char mesg_type;
@@ -1200,11 +1409,45 @@ int sftp_ssh2_packet_handle(void) {
   }
 
   mesg_type = sftp_ssh2_packet_get_mesg_type(pkt);
+  pr_trace_msg(trace_channel, 3, "received %s (%d) packet",
+    sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type);
 
   pr_response_clear(&resp_list);
   pr_response_clear(&resp_err_list);
 
+  /* Note: Some of the SSH messages will be handled regardless of the
+   * sftp_sess_state flags; this is intentional, and is the way that
+   * the protocol is supposed to work.
+   */
+  
   switch (mesg_type) {
+    case SFTP_SSH2_MSG_DEBUG:
+      sftp_ssh2_packet_handle_debug(pkt);
+      break;
+
+    case SFTP_SSH2_MSG_DISCONNECT:
+      sftp_ssh2_packet_handle_disconnect(pkt);
+      break;
+
+    case SFTP_SSH2_MSG_GLOBAL_REQUEST:
+      handle_global_request_mesg(pkt);
+      break;
+
+    case SFTP_SSH2_MSG_REQUEST_SUCCESS:
+    case SFTP_SSH2_MSG_REQUEST_FAILURE:
+    case SFTP_SSH2_MSG_CHANNEL_SUCCESS:
+    case SFTP_SSH2_MSG_CHANNEL_FAILURE:
+      handle_client_alive_mesg(pkt, mesg_type);
+      break;
+
+    case SFTP_SSH2_MSG_IGNORE:
+      sftp_ssh2_packet_handle_ignore(pkt);
+      break;
+
+    case SFTP_SSH2_MSG_UNIMPLEMENTED:
+      sftp_ssh2_packet_handle_unimplemented(pkt);
+      break;
+
     case SFTP_SSH2_MSG_KEXINIT:
       /* The client might be initiating a rekey; watch for this. */
       if (sftp_sess_state & SFTP_SESS_STATE_HAVE_KEX) {
@@ -1215,6 +1458,7 @@ int sftp_ssh2_packet_handle(void) {
       sftp_sess_state &= ~SFTP_SESS_STATE_HAVE_KEX;
 
       if (sftp_kex_handle(pkt) < 0) {
+        pr_event_generate("mod_sftp.ssh2.kex.failed", NULL);
         SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_BY_APPLICATION, NULL);
       }
 
@@ -1246,14 +1490,26 @@ int sftp_ssh2_packet_handle(void) {
 
     case SFTP_SSH2_MSG_USER_AUTH_REQUEST:
       if (sftp_sess_state & SFTP_SESS_STATE_HAVE_SERVICE) {
-        int ok;
 
-        ok = sftp_auth_handle(pkt);
-        if (ok == 1) {
-          sftp_sess_state |= SFTP_SESS_STATE_HAVE_AUTH;
+        /* If the client has already authenticated this connection, then
+         * silently ignore this additional auth request, per recommendation
+         * in RFC4252.
+         */
+        if (sftp_sess_state & SFTP_SESS_STATE_HAVE_AUTH) {
+          (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+            "ignoring %s (%d) message: Connection already authenticated",
+            sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type);
 
-        } else if (ok < 0) {
-          SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_BY_APPLICATION, NULL);
+        } else {
+          int ok;
+
+          ok = sftp_auth_handle(pkt);
+          if (ok == 1) {
+            sftp_sess_state |= SFTP_SESS_STATE_HAVE_AUTH;
+
+          } else if (ok < 0) {
+            SFTP_DISCONNECT_CONN(SFTP_SSH2_DISCONNECT_BY_APPLICATION, NULL);
+          }
         }
 
         break;
@@ -1284,6 +1540,8 @@ int sftp_ssh2_packet_handle(void) {
       }
 
     default:
+      pr_event_generate("ssh2.invalid-packet", pkt);
+
       (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
         "unhandled %s (%d) message, disconnecting",
         sftp_ssh2_packet_get_mesg_type_desc(mesg_type), mesg_type);
@@ -1301,14 +1559,14 @@ int sftp_ssh2_packet_rekey_reset(void) {
   /* Add the rekey seqno limit to the current sequence numbers. */
 
   if (rekey_client_seqno > 0) {
-    rekey_client_seqno = packet_client_seqno + SFTP_PACKET_REKEY_SEQNO_LIMIT;
+    rekey_client_seqno = packet_client_seqno + SFTP_PACKET_CLIENT_REKEY_SEQNO_LIMIT;
 
     if (rekey_client_seqno == 0)
       rekey_client_seqno++;
   }
 
   if (rekey_server_seqno > 0) {
-    rekey_server_seqno = packet_client_seqno + SFTP_PACKET_REKEY_SEQNO_LIMIT;
+    rekey_server_seqno = packet_client_seqno + SFTP_PACKET_SERVER_REKEY_SEQNO_LIMIT;
 
     if (rekey_server_seqno == 0)
       rekey_server_seqno++;
@@ -1351,8 +1609,23 @@ int sftp_ssh2_packet_send_version(void) {
     }
 
     sent_version_id = TRUE;
+    session.total_raw_out += res;
+
+    (void) pr_log_writefile(sftp_logfd, MOD_SFTP_VERSION,
+      "sent server version '%s'", server_version);
   }
 
   return 0;
 }
 
+int sftp_ssh2_packet_set_version(const char *version) {
+  if (server_version == NULL) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  server_version = version;
+  version_id = pstrcat(sftp_pool, version, "\r\n", NULL);
+  return 0;
+}
+
