@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_wrap2 -- tcpwrappers-like access control
  *
- * Copyright (c) 2000-2008 TJ Saunders
+ * Copyright (c) 2000-2014 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders gives permission to link this program
  * with OpenSSL, and distribute the resulting executable, without including
@@ -24,8 +24,13 @@
 
 #include "mod_wrap2.h"
 
+#ifdef WRAP2_USE_NIS
+#include <rpc/rpc.h>
+#include <rpcsvc/ypclnt.h>
+#endif /* WRAP2_USE_NIS */
+
 typedef struct regtab_obj {
-  struct regtab_obj *next;
+  struct regtab_obj *prev, *next;
 
   /* Table source type name */
   const char *regtab_name;
@@ -36,6 +41,7 @@ typedef struct regtab_obj {
 } wrap2_regtab_t;
 
 module wrap2_module;
+unsigned long wrap2_opts = 0UL;
 
 /* Wrap tables for the current session */
 static char *wrap2_allow_table = NULL;
@@ -51,7 +57,7 @@ static wrap2_regtab_t *wrap2_regtab_list = NULL;
 static int wrap2_logfd = -1;
 static char *wrap2_logname = NULL;
 
-static unsigned char wrap2_engine = FALSE;
+static int wrap2_engine = FALSE;
 static char *wrap2_service_name = WRAP2_DEFAULT_SERVICE_NAME;
 static char *wrap2_client_name = NULL;
 static config_rec *wrap2_ctxt = NULL;
@@ -115,48 +121,21 @@ static int wrap2_closelog(void) {
 }
 
 int wrap2_log(const char *fmt, ...) {
-  char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
-  time_t timestamp = time(NULL);
-  struct tm *t = NULL;
   va_list msg;
+  int res;
 
   if (!wrap2_logname)
     return 0;
 
-  t = pr_localtime(NULL, &timestamp);
-
-  /* Prepend the timestamp. */
-  strftime(buf, sizeof(buf), "%b %d %H:%M:%S ", t);
-  buf[sizeof(buf) - 1] = '\0';
-
-  /* Prepend a small header */
-  snprintf(buf + strlen(buf), sizeof(buf) - strlen(buf),
-           MOD_WRAP2_VERSION "[%u]: ", (unsigned int) getpid());
-
-  buf[sizeof(buf) - 1] = '\0';
-
-  /* Affix the message. */
   va_start(msg, fmt);
-  vsnprintf(buf + strlen(buf), sizeof(buf) - strlen(buf), fmt, msg);
+  res = pr_log_vwritefile(wrap2_logfd, MOD_WRAP2_VERSION, fmt, msg);
   va_end(msg);
 
-  buf[strlen(buf)] = '\n';
-  buf[sizeof(buf) - 1] = '\0';
-
-  while (write(wrap2_logfd, buf, strlen(buf)) < 0) {
-    if (errno == EINTR) {
-      pr_signals_handle();
-      continue;
-    }
-
-    return -1;
-  }
-
-  return 0;
+  return res;
 }
 
 static int wrap2_openlog(void) {
-  int res = 0;
+  int res = 0, xerrno;
 
   /* Sanity check */
   wrap2_logname = get_param_ptr(main_server->conf, "WrapLog", FALSE);
@@ -171,10 +150,12 @@ static int wrap2_openlog(void) {
 
   pr_signals_block();
   PRIVS_ROOT
-  res = pr_log_openfile(wrap2_logname, &wrap2_logfd, 0640);
+  res = pr_log_openfile(wrap2_logname, &wrap2_logfd, PR_LOG_SYSTEM_MODE);
+  xerrno = errno;
   PRIVS_RELINQUISH
   pr_signals_unblock();
 
+  errno = xerrno;
   return res;
 }
 
@@ -185,22 +166,25 @@ static int wrap2_close_table(wrap2_table_t *tab) {
 }
 
 static wrap2_table_t *wrap2_open_table(char *name) {
-  char *info = NULL;
+  char *info = NULL, *ptr = NULL;
   unsigned char have_type = FALSE;
   register wrap2_regtab_t *regtab = NULL;
   wrap2_table_t *tab = NULL;
 
-  info = strchr(name, ':');
+  info = ptr = strchr(name, ':');
   *info++ = '\0';
 
   /* Look up the table source open routine by name, and invoke it */
   for (regtab = wrap2_regtab_list; regtab; regtab = regtab->next) {
     if (strcmp(regtab->regtab_name, name) == 0) {
       tab = regtab->regtab_open(wrap2_pool, info);
-      if (tab == NULL)
+      if (tab == NULL) {
+        *ptr = ':';
         return NULL;
+      }
 
       have_type = TRUE;
+      *ptr = ':';
       break;
     }
   }
@@ -307,7 +291,7 @@ static char *wrap2_get_user(wrap2_conn_t *conn) {
 
     /* RFC1413 lookups may have already been done by the mod_ident module.
      * If so, use the ident name stashed; otherwise, use the user name issued
-     * by the client, C_USER.
+     * by the client.
      */
 
     rfc1413_ident = pr_table_get(session.notes, "mod_ident.rfc1413-ident",
@@ -316,8 +300,12 @@ static char *wrap2_get_user(wrap2_conn_t *conn) {
       sstrncpy(conn->user, rfc1413_ident, sizeof(conn->user));
 
     } else {
-      char *user = get_param_ptr(main_server->conf, C_USER, FALSE);
-      sstrncpy(conn->user, user, sizeof(conn->user));
+      char *user;
+
+      user = pr_table_get(session.notes, "mod_auth.orig-user", NULL);
+      if (user) {
+        sstrncpy(conn->user, user, sizeof(conn->user));
+      }
     }
   }
 
@@ -325,32 +313,70 @@ static char *wrap2_get_user(wrap2_conn_t *conn) {
 }
 
 static char *wrap2_get_hostaddr(wrap2_host_t *host) {
-  if (*host->addr == '\0')
+  if (*host->addr == '\0') {
     sstrncpy(host->addr, pr_netaddr_get_ipstr(session.c->remote_addr),
       sizeof(host->addr));
+  }
 
   return host->addr;
 }
 
 static char *wrap2_get_hostname(wrap2_host_t *host) {
-  if (*host->name == '\0')
-    sstrncpy(host->name, session.c->remote_name, sizeof(host->name));
+
+  if (*host->name == '\0') {
+    int reverse_dns;
+    size_t namelen;
+
+    reverse_dns = pr_netaddr_set_reverse_dns(TRUE);
+    if (reverse_dns) {
+      /* If UseReverseDNS is on, then clear any caches, so that we really do
+       * use the DNS name here if possible.
+       */
+      pr_netaddr_clear_cache();
+
+      session.c->remote_addr->na_have_dnsstr = FALSE;
+      sstrncpy(host->name, pr_netaddr_get_dnsstr(session.c->remote_addr),
+        sizeof(host->name));
+
+      /* If the retrieved hostname ends in a trailing period, trim it off. */
+      namelen = strlen(host->name); 
+      if (host->name[namelen-1] == '.') {
+        host->name[namelen-1] = '\0';
+        namelen--;
+      }
+
+      pr_netaddr_set_reverse_dns(reverse_dns);
+      session.c->remote_addr->na_have_dnsstr = TRUE;
+
+    } else {
+      wrap2_log("'UseReverseDNS off' in effect, NOT resolving %s to DNS name "
+        "for comparison", pr_netaddr_get_ipstr(session.c->remote_addr));
+
+      sstrncpy(host->name, pr_netaddr_get_dnsstr(session.c->remote_addr),
+        sizeof(host->name));
+      pr_netaddr_set_reverse_dns(reverse_dns);
+    }
+  }
 
   return host->name;
 }
 
 static char *wrap2_get_hostinfo(wrap2_host_t *host) {
-  char *hostname = wrap2_get_hostname(host);
+  char *hostname;
+
+  hostname = wrap2_get_hostname(host);
 
   if (WRAP2_IS_KNOWN_HOSTNAME(hostname))
-    return host->name;
+    return hostname;
 
   return wrap2_get_hostaddr(host);
 }
 
 static char *wrap2_get_client(wrap2_conn_t *conn) {
   static char both[WRAP2_BUFFER_SIZE] = {'\0'};
-  char *hostinfo = wrap2_get_hostinfo(conn->client);
+  char *hostinfo;
+
+  hostinfo = wrap2_get_hostinfo(conn->client);
 
   if (strcasecmp(wrap2_get_user(conn), WRAP2_UNKNOWN) != 0) {
     snprintf(both, sizeof(both), "%s@%s", conn->user, hostinfo);
@@ -380,7 +406,7 @@ static char *wrap2_skip_whitespace(char *str) {
   /* Skip any leading whitespace. */
   tmp = str;
   for (i = 0; str[i]; i++) {
-    if (isspace((int) str[i])) {
+    if (PR_ISSPACE(str[i])) {
       tmp = &str[i+1];
       continue;
     }
@@ -413,14 +439,10 @@ static unsigned char wrap2_match_string(const char *tok, const char *str) {
 
     /* Prefix */
     return (strncasecmp(tok, str, len) == 0);
-
-  } else {
-
-    /* Exact match */
-    return (strcasecmp(tok, str) == 0);
   }
 
-  return FALSE;
+  /* Exact match */
+  return (strcasecmp(tok, str) == 0);
 }
 
 static unsigned long wrap2_addr_a2n(const char *str) {
@@ -455,11 +477,15 @@ static unsigned char wrap2_match_netmask(const char *net_tok,
    * access control language. John P. Rouillard <rouilj@cs.umb.edu>.
    */
 
-  if ((addr = wrap2_addr_a2n(str)) == INADDR_NONE)
+  addr = wrap2_addr_a2n(str);
+  if (addr == INADDR_NONE) {
     return FALSE;
+  }
 
-  if ((net = wrap2_addr_a2n(net_tok)) == INADDR_NONE ||
-      (mask = wrap2_addr_a2n(mask_tok)) == INADDR_NONE) {
+  net = wrap2_addr_a2n(net_tok);
+  mask = wrap2_addr_a2n(mask_tok);
+  if (net == INADDR_NONE ||
+      mask == INADDR_NONE) {
     wrap2_log("warning: bad net/mask expression: '%s/%s'", net_tok, mask_tok);
     return FALSE;
   }
@@ -469,6 +495,7 @@ static unsigned char wrap2_match_netmask(const char *net_tok,
 
 static unsigned char wrap2_match_host(char *tok, wrap2_host_t *host) {
   char *mask = NULL;
+  size_t len;
 
   tok = wrap2_skip_whitespace(tok);
 
@@ -494,19 +521,81 @@ static unsigned char wrap2_match_host(char *tok, wrap2_host_t *host) {
     return FALSE;
 #endif
 
+  } else if (strcasecmp(tok, "ALL") == 0) {
+    /* Matches everything */
+    return TRUE;
+
   } else if (strcasecmp(tok, "KNOWN") == 0) {
+    char *name;
 
     /* Check address and name. */
-    char *name = wrap2_get_hostname(host);
+    name = wrap2_get_hostname(host);
     return ((strcasecmp(wrap2_get_hostaddr(host), WRAP2_UNKNOWN) != 0) &&
       WRAP2_IS_KNOWN_HOSTNAME(name));
 
   } else if (strcasecmp(tok, "LOCAL") == 0) {
+    char *name;
 
     /* Local: no dots in name. */
-    char *name = wrap2_get_hostname(host);
+    name = wrap2_get_hostname(host);
     return (strchr(name, '.') == NULL && WRAP2_IS_KNOWN_HOSTNAME(name));
 
+  } else if (tok[(len = strlen(tok)) - 1] == '.') {
+    const char *ip_str;
+ 
+    /* Prefix */
+
+    ip_str = wrap2_get_hostaddr(host);
+    return (strncasecmp(tok, ip_str, len) == 0);
+
+  } else if (tok[0] == '.') {
+    register unsigned int i;
+    char *primary_name;
+    array_header *dns_names;
+
+    /* Suffix */
+    primary_name = wrap2_get_hostname(host);
+    len = strlen(primary_name) - strlen(tok);
+
+    wrap2_log("comparing client hostname '%s' (part %s) against DNS "
+      "pattern '%s'", primary_name, primary_name+len, tok);
+
+    if (len > 0 &&
+        strcasecmp(tok, primary_name + len) == 0) {
+      return TRUE;
+    }
+
+    if (!(wrap2_opts & WRAP_OPT_CHECK_ALL_NAMES)) {
+      return FALSE;
+    }
+
+    dns_names = pr_netaddr_get_dnsstr_list(session.pool,
+      session.c->remote_addr);
+    if (dns_names != NULL &&
+        dns_names->nelts > 0) {
+      char **names;
+
+      names = dns_names->elts;
+      for (i = 0; i < dns_names->nelts; i++) {
+        char *name;
+
+        name = names[i];
+        if (name != NULL) {
+          len = strlen(name) - strlen(tok);
+
+          wrap2_log("comparing client hostname '%s' (part %s) against DNS "
+            "pattern '%s'", name, name+len, tok);
+
+          if (len > 0 &&
+              strcasecmp(tok, name + len) == 0) {
+            return TRUE;
+          }
+        }
+      }
+    }
+
+    return FALSE;
+
 #ifdef PR_USE_IPV6 
   } else if (pr_netaddr_use_ipv6() &&
              *tok == '[') {
@@ -516,24 +605,34 @@ static unsigned char wrap2_match_host(char *tok, wrap2_host_t *host) {
 
     /* IPv6 address */
 
-    if (pr_netaddr_get_family(session.c->remote_addr) == AF_INET)
+    if (pr_netaddr_get_family(session.c->remote_addr) == AF_INET) {
       /* No need to try to match an IPv6 address against an IPv4 client. */
       return FALSE;
+    }
 
     /* Find the terminating ']'. */
     cp = strchr(tok, ']');
-    if (cp)
-      *cp = '\0';
+    if (cp == NULL) {
+      wrap2_log("bad IPv6 address syntax: '%s'", tok);
+      return FALSE;
+    }
+
+    *cp = '\0';
 
     /* Lookup a netaddr for the IPv6 address. */
-    acl_addr = pr_netaddr_get_addr(wrap2_pool, tok, NULL);
+    acl_addr = pr_netaddr_get_addr(wrap2_pool, tok + 1, NULL);
     if (!acl_addr) {
-      wrap2_log("unable to resolve IPv6 address '%s'", tok);
+      wrap2_log("unable to resolve IPv6 address '%s'", tok + 1);
+      return FALSE;
+    }
+
+    if (*(cp + 1) != '/') {
+      wrap2_log("bad mask syntax: '%s'", cp + 1);
       return FALSE;
     }
 
     /* Determine the number of mask bits. */
-    nmaskbits = strtol(cp + 1, &tmp, 10);
+    nmaskbits = strtol(cp + 2, &tmp, 10);
     if (tmp && *tmp) {
       wrap2_log("bad mask syntax: '%s'", tmp);
       return FALSE;
@@ -548,11 +647,69 @@ static unsigned char wrap2_match_host(char *tok, wrap2_host_t *host) {
     return (wrap2_match_netmask(tok, mask, wrap2_get_hostaddr(host)));
 
   } else {
+    pr_netaddr_t *acl_addr;
 
-    /* Anything else */
-    return (wrap2_match_string(tok, wrap2_get_hostaddr(host)) ||
-      (WRAP2_IS_NOT_INADDR(tok) &&
-       wrap2_match_string(tok, wrap2_get_hostname(host))));
+    /* Anything else.
+     *
+     * In order to properly compare IP addresses (and to handle cases of
+     * handling IPv4-mapped IPv6 addresses compared against IPv4 addresses),
+     * we need to use pr_netaddr_cmp(), rather than doing the string
+     * comparison that libwrap used.
+     */
+
+    acl_addr = pr_netaddr_get_addr(wrap2_pool, tok, NULL);
+    if (acl_addr == NULL) {
+      if (wrap2_match_string(tok, wrap2_get_hostname(host))) {
+        return TRUE;
+      }
+
+      wrap2_log("unable to handle address '%s'", tok);
+
+    } else {
+      if (pr_netaddr_cmp(session.c->remote_addr, acl_addr) == 0) {
+        return TRUE;
+      }
+    }
+
+    if (WRAP2_IS_NOT_INADDR(tok)) {
+      register unsigned int i;
+      char *primary_name;
+      array_header *dns_names;
+
+      primary_name = wrap2_get_hostname(host);
+      wrap2_log("comparing client hostname '%s' against DNS name '%s'",
+        primary_name, tok);
+
+      if (wrap2_match_string(tok, primary_name)) {
+        return TRUE;
+      }
+
+      if (!(wrap2_opts & WRAP_OPT_CHECK_ALL_NAMES)) {
+        return FALSE;
+      }
+
+      dns_names = pr_netaddr_get_dnsstr_list(session.pool,
+        session.c->remote_addr);
+      if (dns_names != NULL &&
+          dns_names->nelts > 0) {
+        char **names;
+
+        names = dns_names->elts;
+        for (i = 0; i < dns_names->nelts; i++) {
+          char *name;
+
+          name = names[i];
+          if (name != NULL) {
+            wrap2_log("comparing client hostname '%s' against DNS name '%s'",
+              name, tok);
+
+            if (wrap2_match_string(tok, name)) {
+              return TRUE;
+            }
+          }
+        }
+      }
+    }
   }
 
   return FALSE;
@@ -568,8 +725,9 @@ static unsigned char wrap2_match_client(char *tok, wrap2_conn_t *conn) {
     /* Plain host */
     match = wrap2_match_host(tok, conn->client);
 
-    if (match)
+    if (match) {
       wrap2_log("client matches '%s'", tok);
+    }
 
   } else {
 
@@ -577,8 +735,9 @@ static unsigned char wrap2_match_client(char *tok, wrap2_conn_t *conn) {
     match = (wrap2_match_host(host, conn->client) &&
       wrap2_match_string(tok, wrap2_get_user(conn)));
 
-    if (match)
+    if (match) {
       wrap2_log("client matches '%s@%s'", tok, host);
+    }
   }
 
   return match;
@@ -628,7 +787,16 @@ static unsigned char wrap2_match_list(array_header *list, wrap2_conn_t *conn,
    */
 
   for (i = list_idx; i < list->nelts; i++) {
-    char *token = wrap2_skip_whitespace(tokens[i]);
+    char *token;
+
+    /* It's possible that the token string is actually NULL; handle this
+     * case gracefully.
+     */
+    if (tokens[i] == NULL) {
+      continue;
+    }
+
+    token = wrap2_skip_whitespace(tokens[i]);
 
     if (strcasecmp(token, "EXCEPT") == 0) {
       /* EXCEPT -- give up now. */
@@ -639,7 +807,7 @@ static unsigned char wrap2_match_list(array_header *list, wrap2_conn_t *conn,
       register unsigned int j;
 
       /* If yes, look for exceptions */
-      for (j = i + 1; j < list->nelts; j++) {      
+      for (j = i + 1; j < list->nelts; j++) {
         token = wrap2_skip_whitespace(tokens[j]);
         if (strcasecmp(token, "EXCEPT") == 0) {
           return (wrap2_match_list(list, conn, match_token, j+1) == 0);
@@ -682,9 +850,10 @@ static char *wrap2_opt_trim_string(char *string) {
   char *start = NULL, *end = NULL, *cp = NULL;
 
   for (cp = string; *cp; cp++) {
-    if (!isspace((int) *cp)) {
-      if (start == '\0')
+    if (!PR_ISSPACE(*cp)) {
+      if (start == '\0') {
         start = cp;
+      }
       end = cp;
     }
   }
@@ -884,6 +1053,7 @@ static int wrap2_handle_opts(array_header *options, wrap2_conn_t *conn) {
 
 static int wrap2_match_table(wrap2_table_t *tab, wrap2_conn_t *conn) {
   register unsigned int i;
+  int res;
   array_header *daemon_list = NULL, *client_list = NULL, *options_list = NULL;
 
   /* Build daemon list. */
@@ -897,7 +1067,7 @@ static int wrap2_match_table(wrap2_table_t *tab, wrap2_conn_t *conn) {
   wrap2_log("table daemon list:");
   for (i = 0; i < daemon_list->nelts; i++) {
     char **daemons = daemon_list->elts;
-    wrap2_log("  %s", daemons[i]);
+    wrap2_log("  %s", daemons[i] ? daemons[i] : "<null>");
   }
 
   /* Build client list. */
@@ -911,7 +1081,7 @@ static int wrap2_match_table(wrap2_table_t *tab, wrap2_conn_t *conn) {
   wrap2_log("table client list:");
   for (i = 0; i < client_list->nelts; i++) {
     char **clients = client_list->elts;
-    wrap2_log("  %s", clients[i]);
+    wrap2_log("  %s", clients[i] ? clients[i] : "<null>");
   }
 
   /* Build options list. */
@@ -921,26 +1091,32 @@ static int wrap2_match_table(wrap2_table_t *tab, wrap2_conn_t *conn) {
     wrap2_log("table options list:");
     for (i = 0; i < options_list->nelts; i++) {
       char **opts = options_list->elts;
-      wrap2_log("  %s", opts[i]);
+      wrap2_log("  %s", opts[i] ? opts[i] : "<null>");
     }
   }
 
-  if (wrap2_match_list(daemon_list, conn, wrap2_match_daemon, 0) &&
-      wrap2_match_list(client_list, conn, wrap2_match_client, 0)) {
-#ifdef WRAP2_USE_OPTIONS
-    int res = 0;
-    res = wrap2_handle_opts(options_list, conn);
-
-    if (res == WRAP2_OPT_ALLOW)
-      return WRAP2_TAB_ALLOW;
-
-    if (res == WRAP2_OPT_DENY)
-      return WRAP2_TAB_DENY;
-#endif
-    return WRAP2_TAB_MATCH;
+  res = wrap2_match_list(daemon_list, conn, wrap2_match_daemon, 0);
+  if (res == FALSE) {
+    return 0;
   }
 
-  return 0;
+  res = wrap2_match_list(client_list, conn, wrap2_match_client, 0);
+  if (res == FALSE) {
+    return 0;
+  }
+
+#ifdef WRAP2_USE_OPTIONS
+  res = wrap2_handle_opts(options_list, conn);
+  if (res == WRAP2_OPT_ALLOW) {
+    return WRAP2_TAB_ALLOW;
+  }
+
+  if (res == WRAP2_OPT_DENY) {
+    return WRAP2_TAB_DENY;
+  }
+#endif
+
+  return WRAP2_TAB_MATCH;
 }
 
 static unsigned char wrap2_allow_access(wrap2_conn_t *conn) {
@@ -979,8 +1155,9 @@ static unsigned char wrap2_allow_access(wrap2_conn_t *conn) {
       return FALSE;
     }
 
-  } else
+  } else {
     wrap2_log("error opening allow table: %s", strerror(errno));
+  }
 
   /* Open deny table. */
   deny_tab = wrap2_open_table(wrap2_deny_table);
@@ -1001,8 +1178,9 @@ static unsigned char wrap2_allow_access(wrap2_conn_t *conn) {
       return FALSE; 
     }
 
-  } else
+  } else {
     wrap2_log("error opening deny table: %s", strerror(errno));
+  }
 
   wrap2_allow_table = wrap2_deny_table = NULL;
   return TRUE;
@@ -1085,9 +1263,11 @@ int wrap2_register(const char *srcname,
 
   /* Note: I know that use of permanent_pool is discouraged as much as
    * possible, but in this particular instance, I need a pool that
-   * persists across rehashes.  The registration of a table type only
-   * happens once, on module init when the server first starts up, so
-   * this will not constitute a memory leak.
+   * persists across rehashes.
+   *
+   * Ideally, the wrap2_regtab_t struct would have a subpool member;
+   * the objects would have their own pools which could then be
+   * destroyed upon unregistration.
    */
   wrap2_regtab_t *regtab = pcalloc(permanent_pool, sizeof(wrap2_regtab_t));
 
@@ -1095,12 +1275,51 @@ int wrap2_register(const char *srcname,
   regtab->regtab_open = srcopen;
 
   /* Add this object to the list. */
-  regtab->next = wrap2_regtab_list;
+  if (wrap2_regtab_list) {
+    wrap2_regtab_list->prev = regtab;
+    regtab->next = wrap2_regtab_list;
+  }
+
   wrap2_regtab_list = regtab;
 
   return 0;
 }
 
+int wrap2_unregister(const char *srcname) {
+  if (wrap2_regtab_list) {
+    register wrap2_regtab_t *regtab = NULL;
+
+    for (regtab = wrap2_regtab_list; regtab; regtab = regtab->next) {
+      if (strcmp(regtab->regtab_name, srcname) == 0) {
+
+        if (regtab->prev) {
+          regtab->prev->next = regtab->next;
+
+        } else {
+          wrap2_regtab_list = regtab->next;
+        }
+
+        if (regtab->next) {
+          regtab->next->prev = regtab->prev;
+        }
+
+        regtab->prev = regtab->next = NULL;
+
+        /* NOTE: a counter should be kept of the number of unregistrations,
+         * as the memory for a registration is not freed on unregistration.
+         */
+        return 0;
+      }
+    }
+
+    errno = ENOENT;
+    return -1;
+  }
+
+  errno = EPERM;
+  return -1;
+}
+
 /* "builtin" source callbacks. */
 
 static int builtin_close_cb(wrap2_table_t *tab) {
@@ -1178,12 +1397,13 @@ MODRET set_wrapengine(cmd_rec *cmd) {
   CHECK_ARGS(cmd, 1);
   CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
 
-  if ((bool = get_boolean(cmd, 1)) == -1)
+  bool = get_boolean(cmd, 1);
+  if (bool == -1)
     CONF_ERROR(cmd, "expecting Boolean parameter");
 
   c = add_config_param(cmd->argv[0], 1, NULL);
-  c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
-  *((unsigned char *) c->argv[0]) = (unsigned char) bool;
+  c->argv[0] = pcalloc(c->pool, sizeof(int));
+  *((int *) c->argv[0]) = bool;
 
   return PR_HANDLED(cmd);
 }
@@ -1267,6 +1487,38 @@ MODRET set_wraplog(cmd_rec *cmd) {
   return PR_HANDLED(cmd);
 }
 
+/* usage: WrapOptions opt1 ... optN */
+MODRET set_wrapoptions(cmd_rec *cmd) {
+  config_rec *c = NULL;
+  register unsigned int i = 0;
+  unsigned long opts = 0UL;
+
+  if (cmd->argc-1 == 0)
+    CONF_ERROR(cmd, "wrong number of parameters");
+
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
+
+  c = add_config_param(cmd->argv[0], 1, NULL);
+
+  for (i = 1; i < cmd->argc; i++) {
+    if (strcmp(cmd->argv[i], "CheckOnConnect") == 0) {
+      opts |= WRAP_OPT_CHECK_ON_CONNECT;
+
+    } else if (strcmp(cmd->argv[i], "CheckAllNames") == 0) {
+      opts |= WRAP_OPT_CHECK_ALL_NAMES;
+
+    } else {
+      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown WrapOption '",
+        cmd->argv[i], "'", NULL));
+    }
+  }
+
+  c->argv[0] = pcalloc(c->pool, sizeof(unsigned long));
+  *((unsigned long *) c->argv[0]) = opts;
+
+  return PR_HANDLED(cmd);
+}
+
 /* usage: WrapServiceName <name> */
 MODRET set_wrapservicename(cmd_rec *cmd) {
   CHECK_ARGS(cmd, 1);
@@ -1392,7 +1644,7 @@ MODRET set_wrapusertables(cmd_rec *cmd) {
  */
 
 MODRET wrap2_pre_pass(cmd_rec *cmd) {
-  wrap2_conn_t con;
+  wrap2_conn_t conn;
   unsigned char have_tables = FALSE;
   char *user = NULL;
   config_rec *c = NULL;
@@ -1403,8 +1655,7 @@ MODRET wrap2_pre_pass(cmd_rec *cmd) {
   /* Hide passwords */
   session.hide_password = TRUE;
 
-  user = get_param_ptr(cmd->server->conf, C_USER, FALSE);
-
+  user = pr_table_get(session.notes, "mod_auth.orig-user", NULL);
   if (!user)
     return PR_DECLINED(cmd);
 
@@ -1435,11 +1686,12 @@ MODRET wrap2_pre_pass(cmd_rec *cmd) {
       session.login_gid = pw->pw_gid;
 
       gr = pr_auth_getgrgid(cmd->pool, session.login_gid);
-      if (gr != NULL)
+      if (gr != NULL) {
         session.group = pstrdup(cmd->pool, gr->gr_name);
 
-      else
+      } else {
         wrap2_log("unable to resolve GID for '%s'", user);
+      }
 
     } else {
       wrap2_log("unable to resolve UID for '%s'", user);
@@ -1453,9 +1705,12 @@ MODRET wrap2_pre_pass(cmd_rec *cmd) {
   
   c = find_config(wrap2_ctxt ? wrap2_ctxt->subset : main_server->conf,
     CONF_PARAM, "WrapUserTables", FALSE);
-
   while (c) {
-    array_header *user_array = make_array(cmd->tmp_pool, 0, sizeof(char *));
+    array_header *user_array;
+
+    pr_signals_handle();
+
+    user_array = make_array(cmd->tmp_pool, 0, sizeof(char *));
     *((char **) push_array(user_array)) = pstrdup(cmd->tmp_pool, user);
 
     /* Check the user OR expression. Do not forget the offset, to skip
@@ -1480,13 +1735,18 @@ MODRET wrap2_pre_pass(cmd_rec *cmd) {
   /* Next, search for group-specific access tables.  Multiple WrapGroupTables
    * directives are allowed.
    */ 
-  if (!have_tables)
+  if (!have_tables) {
     c = find_config(wrap2_ctxt ? wrap2_ctxt->subset : main_server->conf,
       CONF_PARAM, "WrapGroupTables", FALSE);
+  }
 
   while (c) {
-    array_header *gid_array = make_array(cmd->pool, 0, sizeof(gid_t));
-    array_header *group_array = make_array(cmd->pool, 0, sizeof(char *));
+    array_header *gid_array, *group_array;
+
+    pr_signals_handle();
+
+    gid_array = make_array(cmd->pool, 0, sizeof(gid_t));
+    group_array = make_array(cmd->pool, 0, sizeof(char *));
 
     if (pr_auth_getgroups(cmd->pool, user, &gid_array, &group_array) < 1) {
       wrap2_log("no supplemental groups found for user '%s'", user);
@@ -1516,9 +1776,10 @@ MODRET wrap2_pre_pass(cmd_rec *cmd) {
   /* Finally for globally-applicable access files.  Only one such directive
    * is allowed.
    */
-  if (!have_tables)
+  if (!have_tables) {
     c = find_config(wrap2_ctxt ? wrap2_ctxt->subset : main_server->conf,
       CONF_PARAM, "WrapTables", FALSE);
+  }
 
   if (c) {
     wrap2_allow_table = c->argv[0];
@@ -1534,40 +1795,44 @@ MODRET wrap2_pre_pass(cmd_rec *cmd) {
     wrap2_log("using '%s' for deny table", wrap2_deny_table);
 
   } else {
-
     wrap2_log("no tables configured, allowing connection");
     return PR_DECLINED(cmd);
   }
 
   wrap2_log("looking under service name '%s'", wrap2_service_name);
 
-  memset(&con, '\0', sizeof(con));
+  memset(&conn, '\0', sizeof(conn));
 
-  wrap2_conn_set(&con, WRAP2_CONN_DAEMON, wrap2_service_name,
+  wrap2_conn_set(&conn, WRAP2_CONN_DAEMON, wrap2_service_name,
     WRAP2_CONN_SOCK_FD, session.c->rfd, 0);
 
   wrap2_log("%s", "checking access rules for connection");
 
-  if (strcasecmp(wrap2_get_hostname(con.client), WRAP2_PARANOID) == 0 ||
-      !wrap2_allow_access(&con)) {
+  if (wrap2_allow_access(&conn) == FALSE) {
     char *msg = NULL;
 
     /* Log the denied connection */
-    wrap2_log("refused connection from %s", wrap2_get_client(&con));
+    wrap2_log("refused connection from %s", wrap2_get_client(&conn));
+
+    /* Broadcast this event to any interested listeners.  We use the same
+     * event name as mod_wrap for consistency.
+     */
+    pr_event_generate("mod_wrap.connection-denied", NULL);
 
     /* Check for a configured WrapDenyMsg.  If not present, then use the
      * default denied message.
      */
     msg = get_param_ptr(wrap2_ctxt ? wrap2_ctxt->subset : main_server->conf,
       "WrapDenyMsg", FALSE);
-    if (msg != NULL)
+    if (msg != NULL) {
       msg = sreplace(cmd->tmp_pool, msg, "%u", user, NULL);
+    }
 
-    pr_response_send(R_530, msg ? msg : _("Access denied"));
-    end_login(0);
+    pr_response_send(R_530, "%s", msg ? msg : _("Access denied"));
+    pr_session_disconnect(&wrap2_module, PR_SESS_DISCONNECT_MODULE_ACL, NULL);
   }
 
-  wrap2_log("allowed connection from %s", wrap2_get_client(&con));
+  wrap2_log("allowed connection from %s", wrap2_get_client(&conn));
   return PR_DECLINED(cmd);
 }
 
@@ -1583,7 +1848,9 @@ MODRET wrap2_post_pass(cmd_rec *cmd) {
   msg = get_param_ptr(wrap2_ctxt ? wrap2_ctxt->subset : main_server->conf,
     "WrapAllowMsg", FALSE);
   if (msg != NULL) {
-    char *user = get_param_ptr(cmd->server->conf, C_USER, FALSE);
+    char *user;
+
+    user = pr_table_get(session.notes, "mod_auth.orig-user", NULL);
     msg = sreplace(cmd->tmp_pool, msg, "%u", user, NULL);
     pr_response_add(R_DUP, "%s", msg);
   }
@@ -1591,6 +1858,24 @@ MODRET wrap2_post_pass(cmd_rec *cmd) {
   return PR_DECLINED(cmd);
 }
 
+MODRET wrap2_post_pass_err(cmd_rec *cmd) {
+  if (!wrap2_engine)
+    return PR_DECLINED(cmd);
+
+  /* Clear the values from the session struct as well, specifically
+   * session.user.  Failure to do so caused Bug#3727.
+   */
+  session.user = NULL;
+  session.group = NULL;
+   
+  wrap2_ctxt = NULL;
+  wrap2_allow_table = NULL;
+  wrap2_deny_table = NULL;
+  wrap2_client_name = NULL;
+
+  return PR_DECLINED(cmd);
+}
+
 /* Event handlers
  */
 
@@ -1605,10 +1890,15 @@ static void wrap2_mod_unload_ev(const void *event_data, void *user_data) {
     /* Unregister ourselves from all events. */
     pr_event_unregister(&wrap2_module, NULL, NULL);
 
+    wrap2_unregister("builtin");
+
     if (wrap2_pool) {
       destroy_pool(wrap2_pool);
       wrap2_pool = NULL;
     }
+
+    close(wrap2_logfd);
+    wrap2_logfd = -1;
   }
 }
 
@@ -1650,29 +1940,85 @@ static int wrap2_init(void) {
 }
 
 static int wrap2_sess_init(void) {
-  unsigned char *engine = NULL;
+  config_rec *c;
 
-  engine = get_param_ptr(main_server->conf, "WrapEngine", FALSE);
-  if (engine != NULL &&
-      *engine == TRUE)
-    wrap2_engine = TRUE;
+  c = find_config(main_server->conf, CONF_PARAM, "WrapEngine", FALSE);
+  if (c) {
+    wrap2_engine = *((int *) c->argv[0]);
+  }
 
-  else {
-    wrap2_engine = FALSE;
+  if (!wrap2_engine) {
     return 0;
   }
 
+  wrap2_openlog();
+
   /* Look up any configured WrapServiceName */
   wrap2_service_name = get_param_ptr(main_server->conf, "WrapServiceName",
     FALSE);
   if (wrap2_service_name == NULL)
     wrap2_service_name = WRAP2_DEFAULT_SERVICE_NAME;
 
-  wrap2_openlog();
-
   /* Make sure that tables will be closed when the child exits. */
   pr_event_register(&wrap2_module, "core.exit", wrap2_exit_ev, NULL);
 
+  c = find_config(main_server->conf, CONF_PARAM, "WrapOptions", FALSE);
+  while (c != NULL) {
+    unsigned long opts;
+
+    pr_signals_handle();
+
+    opts = *((unsigned long *) c->argv[0]);
+    wrap2_opts |= opts;
+
+    c = find_config_next(c, c->next, CONF_PARAM, "WrapOptions", FALSE);
+  }
+
+  if (wrap2_opts & WRAP_OPT_CHECK_ON_CONNECT) {
+    c = find_config(main_server->conf, CONF_PARAM, "WrapTables", FALSE);
+    if (c) {
+      wrap2_conn_t conn;
+
+      wrap2_allow_table = c->argv[0];
+      wrap2_deny_table = c->argv[1];
+      wrap2_client_name = "";
+
+      wrap2_log("using '%s' for allow table", wrap2_allow_table);
+      wrap2_log("using '%s' for deny table", wrap2_deny_table);
+      wrap2_log("looking under service name '%s'", wrap2_service_name);
+
+      memset(&conn, '\0', sizeof(conn));
+      wrap2_conn_set(&conn, WRAP2_CONN_DAEMON, wrap2_service_name,
+        WRAP2_CONN_SOCK_FD, session.c->rfd, 0);
+
+      wrap2_log("%s", "checking access rules for connection");
+
+      if (wrap2_allow_access(&conn) == FALSE) {
+        char *msg = NULL;
+
+        /* Log the denied connection */
+        wrap2_log("refused connection from %s", wrap2_get_client(&conn));
+
+        /* Broadcast this event to any interested listeners.  We use the same
+         * event name as mod_wrap for consistency.
+         */
+        pr_event_generate("mod_wrap.connection-denied", NULL);
+
+        /* Check for a configured WrapDenyMsg.  If not present, then use the
+         * default denied message.
+         */
+        msg = get_param_ptr(main_server->conf, "WrapDenyMsg", FALSE);
+        if (msg != NULL) {
+          msg = sreplace(session.pool, msg, "%u", "unknown", NULL);
+        }
+
+        pr_response_send(R_530, "%s", msg ? msg : _("Access denied"));
+        pr_session_disconnect(&wrap2_module, PR_SESS_DISCONNECT_MODULE_ACL,
+          NULL);
+      }
+    }
+  }
+
   return 0;
 }
 
@@ -1685,6 +2031,7 @@ static conftable wrap2_conftab[] = {
   { "WrapEngine",		set_wrapengine,		NULL },
   { "WrapGroupTables",		set_wrapgrouptables,	NULL },
   { "WrapLog",			set_wraplog,		NULL },
+  { "WrapOptions",		set_wrapoptions,	NULL },
   { "WrapServiceName",		set_wrapservicename,	NULL },
   { "WrapTables",		set_wraptables,		NULL },
   { "WrapUserTables",		set_wrapusertables,	NULL },
@@ -1692,8 +2039,9 @@ static conftable wrap2_conftab[] = {
 };
 
 static cmdtable wrap2_cmdtab[] = {
-  { PRE_CMD,	C_PASS,	G_NONE,	wrap2_pre_pass,	FALSE,	FALSE },
-  { POST_CMD,	C_PASS,	G_NONE,	wrap2_post_pass,FALSE,	FALSE },
+  { PRE_CMD,	C_PASS,	G_NONE,	wrap2_pre_pass,		FALSE,	FALSE },
+  { POST_CMD,	C_PASS,	G_NONE,	wrap2_post_pass,	FALSE,	FALSE },
+  { POST_CMD_ERR,C_PASS,G_NONE,	wrap2_post_pass_err,	FALSE,	FALSE },
   { 0, NULL }
 };
 
