@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2001, 2002, 2003 The ProFTPD Project team
+ * Copyright (c) 2001-2012 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,8 +23,7 @@
  */
 
 /* ProFTPD bindings support routines.
- *
- * $Id: bindings.h,v 1.10 2004/11/20 20:32:19 castaglia Exp $
+ * $Id: bindings.h,v 1.14 2012-04-15 18:04:14 castaglia Exp $
  */
 
 #include "conf.h"
@@ -60,23 +59,24 @@ typedef struct ipbind_rec {
    */
   array_header *ib_namebinds;
 
-  /* if this binding is the DefaultServer binding */
+  /* If this binding is the DefaultServer binding */
   unsigned char ib_isdefault;
 
-  /* if this binding handles localhost requests */
+  /* If this binding handles localhost requests */
   unsigned char ib_islocalhost;
 
-  /* if this binding is active */
+  /* If this binding is active */
   unsigned char ib_isactive;
 
 } pr_ipbind_t;
 
 /* Structure associating a name to a server_rec */
 typedef struct namebind_rec {
-  server_rec *nb_server;
-  conn_t *nb_listener;
   const char *nb_name;
+  unsigned char nb_iswildcard;
   unsigned char nb_isactive;
+  server_rec *nb_server;
+
 } pr_namebind_t;
 
 /* Define the size of the hash table used to store server configurations.
@@ -94,7 +94,7 @@ conn_t *pr_ipbind_accept_conn(fd_set *readfds, int *listenfd);
  * arguments. The new binding is added the list maintained by the bindings
  * layer.  Returns 0 on success, -1 on failure.
  */
-int pr_ipbind_create(server_rec *server, pr_netaddr_t *addr);
+int pr_ipbind_create(server_rec *server, pr_netaddr_t *addr, unsigned int port);
 
 /* Close all IP bindings associated with the given IP address/port combination.
  * The bindings are then marked as inactive, so that future lookups via
@@ -146,6 +146,9 @@ int pr_ipbind_open(pr_netaddr_t *addr, unsigned int port, conn_t *listen_conn,
   unsigned char isdefault, unsigned char islocalhost,
   unsigned char open_namebinds);
 
+conn_t *pr_ipbind_get_listening_conn(server_rec *server, pr_netaddr_t *addr,
+  unsigned int port);
+
 /* Close the pr_namebind_t with the given name.
  */
 int pr_namebind_close(const char *name, pr_netaddr_t *addr, unsigned int port);
@@ -196,11 +199,11 @@ void free_bindings(void);
       "%s:%d: notice: unable to close ipbind: %s", \
       __FILE__, __LINE__, strerror(errno))
 
-#define PR_CREATE_IPBIND(s, a) \
-  if ((res = pr_ipbind_create((s), (a))) < 0) \
+#define PR_CREATE_IPBIND(s, a, p) \
+  if ((res = pr_ipbind_create((s), (a), (p))) < 0) \
     pr_log_pri(PR_LOG_NOTICE, \
-      "%s:%d: notice: unable to create ipbind '%s': %s", \
-      __FILE__, __LINE__, (s)->ServerAddress, strerror(errno))
+      "%s:%d: notice: unable to create ipbind '%s#%u': %s", \
+      __FILE__, __LINE__, (s)->ServerAddress, (p), strerror(errno))
 
 #define PR_OPEN_IPBIND(a, p, c, d, l, o) \
   if ((res = pr_ipbind_open((a), (p), (c), (d), (l), (o))) < 0) \
