@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2006 The ProFTPD Project
+ * Copyright (c) 2001-2014 The ProFTPD Project
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -25,8 +25,7 @@
  */
 
 /* ProFTPD virtual/modular filesystem support.
- *
- * $Id: fsio.h,v 1.15 2007/01/11 04:05:07 castaglia Exp $
+ * $Id: fsio.h,v 1.37 2014/01/31 16:52:33 castaglia Exp $
  */
 
 #ifndef PR_FSIO_H
@@ -34,10 +33,6 @@
 
 #include "conf.h"
 
-#ifdef HAVE_REGEX_H
-#include <regex.h>
-#endif
-
 /* This is a Tru64-specific hack, to work around some macro funkiness
  * in their /usr/include/sys/mount.h header.
  */
@@ -62,6 +57,7 @@
 #define FSIO_FILE_CHMOD		(1 << 13)
 #define FSIO_FILE_CHOWN		(1 << 14)
 #define FSIO_FILE_ACCESS	(1 << 15)
+#define FSIO_FILE_UTIMES	(1 << 23)
 
 /* Macro that defines the most common file ops */
 #define FSIO_FILE_COMMON	(FSIO_FILE_OPEN|FSIO_FILE_READ|FSIO_FILE_WRITE|\
@@ -105,7 +101,7 @@ struct fs_rec {
   void *fs_data;
 
   /* Pool for this object's use */
-  pool *fs_pool;
+  struct pool_rec *fs_pool;
 
   /* FS function pointers */
   int (*stat)(pr_fs_t *, const char *, struct stat *);
@@ -125,9 +121,14 @@ struct fs_rec {
   int (*ftruncate)(pr_fh_t *, int, off_t);
   int (*truncate)(pr_fs_t *, const char *, off_t);
   int (*chmod)(pr_fs_t *, const char *, mode_t);
+  int (*fchmod)(pr_fh_t *, int, mode_t);
   int (*chown)(pr_fs_t *, const char *, uid_t, gid_t);
+  int (*fchown)(pr_fh_t *, int, uid_t, gid_t);
+  int (*lchown)(pr_fs_t *, const char *, uid_t, gid_t);
   int (*access)(pr_fs_t *, const char *, int, uid_t, gid_t, array_header *);
   int (*faccess)(pr_fh_t *, int, uid_t, gid_t, array_header *);
+  int (*utimes)(pr_fs_t *, const char *, struct timeval *);
+  int (*futimes)(pr_fh_t *, int, struct timeval *);
 
   /* For actual operations on the directory (or subdirs)
    * we cast the return from opendir to DIR* in src/fs.c, so
@@ -141,6 +142,29 @@ struct fs_rec {
   struct dirent *(*readdir)(pr_fs_t *, void *);
   int (*mkdir)(pr_fs_t *, const char *, mode_t);
   int (*rmdir)(pr_fs_t *, const char *);
+
+  /* This flag determines whether this FS handler allows cross-FS hardlinks,
+   * either from this FS to another FS, or from another FS to this FS.
+   *
+   * If the flag is set to FALSE by the FS registrant, then a hardlink
+   * across FS handlers will fail, with errno set to EXDEV.  The caller
+   * will then have to handle the EXDEV error appropriately.
+   */
+  int allow_xdev_link;
+
+  /* This flag determines whether this FS handler allows cross-FS renames,
+   * either from this FS to another FS, or from another FS to this FS.
+   *
+   * If the flag is set to FALSE by the FS registrant, then a rename
+   * across FS handlers will fail, with errno set to EXDEV.  The caller
+   * will then have to handle the EXDEV error appropriately.
+   *
+   * In the core engine, a RNFR/RNTO sequence which encounters an EXDEV
+   * error will cause a copy/delete of the file.  This can be more IO
+   * intensive than expected, and lead to longer times for the RNTO
+   * command to complete.
+   */
+  int allow_xdev_rename;
 };
 
 struct fh_rec {
@@ -159,15 +183,20 @@ struct fh_rec {
 
   /* For buffer I/O on this file, should anything choose to use it. */
   pr_buffer_t *fh_buf;
+
+  /* Hint of the optimal buffer size for IO on this file. */
+  size_t fh_iosz;
 };
 
+/* Maximum symlink count, for loop detection. */
+#define PR_FSIO_MAX_LINK_COUNT		32
+
 /* Macros for that code that needs to get into the internals of pr_fs_t.
  * (These will help keep the internals as opaque as possible).
  */
 #define PR_FH_FD(f)	((f)->fh_fd)
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-#ifdef PR_FS_MATCH
+#if defined(PR_USE_REGEX) && defined(PR_FS_MATCH)
 typedef struct fs_match_rec pr_fs_match_t;
 struct fs_match_rec {
 
@@ -199,8 +228,7 @@ struct fs_match_rec {
    */
   array_header *fsm_fs_objs;
 };
-#endif /* PR_FS_MATCH */
-#endif /* HAVE_REGEX_H && HAVE_REGCOMP */
+#endif /* PR_USE_REGEX and PR_FS_MATCH */
 
 int pr_fsio_stat(const char *, struct stat *);
 int pr_fsio_stat_canon(const char *, struct stat *);
@@ -218,6 +246,7 @@ int pr_fsio_mkdir(const char *, mode_t);
 int pr_fsio_rmdir(const char *);
 int pr_fsio_rename(const char *, const char *);
 int pr_fsio_rename_canon(const char *, const char *);
+int pr_fsio_smkdir(pool *, const char *, mode_t, uid_t, gid_t);
 int pr_fsio_unlink(const char *);
 int pr_fsio_unlink_canon(const char *);
 pr_fh_t *pr_fsio_open(const char *, int);
@@ -235,19 +264,36 @@ int pr_fsio_ftruncate(pr_fh_t *, off_t);
 int pr_fsio_truncate(const char *, off_t);
 int pr_fsio_truncate_canon(const char *, off_t);
 int pr_fsio_chmod(const char *, mode_t);
+int pr_fsio_fchmod(pr_fh_t *, mode_t);
 int pr_fsio_chmod_canon(const char *, mode_t);
 int pr_fsio_chown(const char *, uid_t, gid_t);
+int pr_fsio_fchown(pr_fh_t *, uid_t, gid_t);
+int pr_fsio_lchown(const char *, uid_t, gid_t);
 int pr_fsio_chown_canon(const char *, uid_t, gid_t);
 int pr_fsio_chroot(const char *);
 int pr_fsio_access(const char *, int, uid_t, gid_t, array_header *);
 int pr_fsio_faccess(pr_fh_t *, int, uid_t, gid_t, array_header *);
+int pr_fsio_utimes(const char *, struct timeval *);
+int pr_fsio_futimes(pr_fh_t *, struct timeval *);
 off_t pr_fsio_lseek(pr_fh_t *, off_t, int);
 
+/* Set a flag determining whether we guard against write operations in
+ * certain sensitive directories while we are chrooted, e.g. "Roaring Beast"
+ * style attacks.
+ */
+int pr_fsio_guard_chroot(int);
+
+/* Set a flag determining whether to use mkdtemp(3) (if available) or not.
+ * Returns the previously-set value.
+ */
+int pr_fsio_set_use_mkdtemp(int);
+
 /* FS-related functions */
 
 char *pr_fsio_getline(char *, int, pr_fh_t *, unsigned int *);
 char *pr_fsio_gets(char *, size_t, pr_fh_t *);
 int pr_fsio_puts(const char *, pr_fh_t *);
+int pr_fsio_set_block(pr_fh_t *);
 
 pr_fs_t *pr_register_fs(pool *, const char *, const char *);
 pr_fs_t *pr_create_fs(pool *, const char *);
@@ -257,8 +303,7 @@ pr_fs_t *pr_remove_fs(const char *);
 pr_fs_t *pr_unmount_fs(const char *, const char *);
 int pr_unregister_fs(const char *);
 
-#if defined(HAVE_REGEX_H) && defined(HAVE_RECOMP)
-#ifdef PR_FS_MATCH
+#if defined(PR_USE_REGEX) && defined(PR_FS_MATCH)
 pr_fs_match_t *pr_register_fs_match(pool *, const char *, const char *, int);
 void pr_associate_fs(pr_fs_match_t *, pr_fs_t *);
 pr_fs_match_t *pr_create_fs_match(pool *, const char *, const char *, int);
@@ -266,8 +311,7 @@ pr_fs_match_t *pr_get_fs_match(const char *, int);
 pr_fs_match_t *pr_get_next_fs_match(pr_fs_match_t *, const char *, int);
 int pr_insert_fs_match(pr_fs_match_t *);
 int pr_unregister_fs_match(const char *);
-#endif /* PR_FS_MATCH */
-#endif /* HAVE_REGEX_H && HAVE_REGCOMP */
+#endif /* PR_USE_REGEX and PR_FS_MATCH */
 
 void pr_fs_clear_cache(void);
 int pr_fs_copy_file(const char *, const char *);
@@ -280,19 +324,52 @@ int pr_fs_resolve_partial(const char *, char *, size_t, int);
 int pr_fs_resolve_path(const char *, char *, size_t, int);
 char *pr_fs_decode_path(pool *, const char *);
 char *pr_fs_encode_path(pool *, const char *);
-int pr_fs_use_utf8(int bool);
+int pr_fs_use_encoding(int);
 int pr_fs_valid_path(const char *);
 void pr_fs_virtual_path(const char *, char *, size_t);
+
 void pr_fs_clean_path(const char *, char *, size_t);
+int pr_fs_clean_path2(const char *, char *, size_t, int);
+#define PR_FSIO_CLEAN_PATH_FL_MAKE_ABS_PATH	0x001
+
 int pr_fs_glob(const char *, int, int (*errfunc)(const char *, int), glob_t *);
 void pr_fs_globfree(glob_t *);
 void pr_resolve_fs_map(void);
 
+/* The main three fds (stdin, stdout, stderr) need to be protected, reserved
+ * for use.  This function uses dup(2) to open new fds on the given fd
+ * until the new fd is not one of the big three.
+ */
+int pr_fs_get_usable_fd(int);
+
+/* Similar to pr_fs_get_usable_fd(), except that it automatically closes the
+ * old (given) fd if a usable fd was found.  Returns -1 (with errno set) if
+ * a usable fd could not be found.
+ */
+int pr_fs_get_usable_fd2(int *);
+
 #if defined(HAVE_STATFS) || defined(HAVE_SYS_STATVFS_H) || \
   defined(HAVE_SYS_VFS_H)
 off_t pr_fs_getsize(char *);
 #endif
 
+/* Unlike pr_fs_getsize(), this function is always present, and is also
+ * capable of returning an error when there is a problem checking the
+ * filesystem stats.
+ */
+int pr_fs_getsize2(char *, off_t *);
+
+/* Similar to pr_fs_getsize2(), except that this operates on an already-opened
+ * file descriptor, rather than a path.
+ */
+int pr_fs_fgetsize(int, off_t *);
+
+/* Returns TRUE if the given path is on an NFS-mounted filesystem, FALSE
+ * if not on an NFS-mounted filesystem, and -1 if there was an error
+ * determining which (with errno set appropriately).
+ */
+int pr_fs_is_nfs(const char *path);
+
 /* For internal use only. */
 int init_fs(void);
 
