@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004 The ProFTPD Project team
+ * Copyright (c) 2004-2011 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,7 +23,7 @@
  */
 
 /* Variables API definition
- * $Id: var.h,v 1.2 2004/11/01 17:12:27 castaglia Exp $
+ * $Id: var.h,v 1.4 2011-05-23 20:35:35 castaglia Exp $
  */
 
 #ifndef PR_VAR_H
@@ -54,11 +54,6 @@ const char *pr_var_get(const char *name);
  */
 const char *pr_var_next(const char **desc);
 
-/* Remove the given variable name from the table.  Returns 0 on success, -1
- * if there was an error.
- */
-int pr_var_remove(const char *name);
-
 /* Rewinds the iterator used by pr_var_next() to the start of the list.
  */
 void pr_var_rewind(void);
