@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_facts -- a module for handling "facts" [RFC3659]
  *
- * Copyright (c) 2007-2009 The ProFTPD Project
+ * Copyright (c) 2007-2013 The ProFTPD Project
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,19 +15,20 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
  * resulting executable, without including the source code for OpenSSL in the
  * source distribution.
  *
- * $Id: mod_facts.c,v 1.26.2.1 2010/03/02 18:08:07 castaglia Exp $
+ * $Id: mod_facts.c,v 1.60 2013-04-16 16:04:43 castaglia Exp $
  */
 
 #include "conf.h"
+#include "privs.h"
 
-#define MOD_FACTS_VERSION		"mod_facts/0.1"
+#define MOD_FACTS_VERSION		"mod_facts/0.3"
 
 #if PROFTPD_VERSION_NUMBER < 0x0001030101
 # error "ProFTPD 1.3.1rc1 or later required"
@@ -45,6 +46,10 @@ static unsigned long facts_opts = 0;
 #define FACTS_OPT_SHOW_UNIX_MODE	0x00040
 #define FACTS_OPT_SHOW_UNIX_OWNER	0x00080
 
+static unsigned long facts_mlinfo_opts = 0;
+#define FACTS_MLINFO_FL_SHOW_SYMLINKS			0x00001
+#define FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK		0x00002
+
 struct mlinfo {
   pool *pool;
   struct stat st;
@@ -56,25 +61,24 @@ struct mlinfo {
 
 /* Necessary prototypes */
 static void facts_mlinfobuf_flush(void);
+static int facts_sess_init(void);
 
 /* Support functions
  */
 
 static int facts_filters_allow_path(cmd_rec *cmd, const char *path) {
-#if defined(HAVE_REGEX_H) && defined(HAVE_REGCOMP)
-  regex_t *preg = get_param_ptr(CURRENT_CONF, "PathAllowFilter", FALSE);
-
-  if (preg &&
-      regexec(preg, path, 0, NULL, 0) != 0) {
+#ifdef PR_USE_REGEX
+  pr_regex_t *pre = get_param_ptr(CURRENT_CONF, "PathAllowFilter", FALSE);
+  if (pre != NULL &&
+      pr_regexp_exec(pre, path, 0, NULL, 0, 0, 0) != 0) {
     pr_log_debug(DEBUG2, MOD_FACTS_VERSION
       ": %s denied by PathAllowFilter on '%s'", cmd->argv[0], cmd->arg);
     return -1;
   }
 
-  preg = get_param_ptr(CURRENT_CONF, "PathDenyFilter", FALSE);
-
-  if (preg &&
-      regexec(preg, path, 0, NULL, 0) == 0) {
+  pre = get_param_ptr(CURRENT_CONF, "PathDenyFilter", FALSE);
+  if (pre != NULL &&
+      pr_regexp_exec(pre, path, 0, NULL, 0, 0, 0) == 0) {
     pr_log_debug(DEBUG2, MOD_FACTS_VERSION
       ": %s denied by PathDenyFilter on '%s'", cmd->argv[0], cmd->arg);
     return -1;
@@ -84,73 +88,117 @@ static int facts_filters_allow_path(cmd_rec *cmd, const char *path) {
   return 0;
 }
 
-static time_t facts_mktime(unsigned int year, unsigned int month,
-    unsigned int mday, unsigned int hour, unsigned int min, unsigned int sec) {
-  struct tm tm;
-  time_t res;
-  char *env;
+#define FACTS_SECS_PER_MIN	(60)
+#define FACTS_SECS_PER_HOUR	(60 * FACTS_SECS_PER_MIN)
+#define FACTS_SECS_PER_DAY	(24 * FACTS_SECS_PER_HOUR)
+#define FACTS_EPOCH_YEAR	1970
 
-#ifdef HAVE_TZNAME
-  char *tzname_dup[2];
+/* How many days come before each month (0-12).  */
+static const unsigned short int facts_ydays_for_mon[2][13] = {
+  /* Normal years.  */
+  { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
 
-  /* The mktime(3) function has a nasty habit of changing the tzname global
-   * variable as a side-effect.  This can cause problems, as when the process
-   * has become chrooted, and mktime(3) sets/changes tzname wrong.  (For more
-   * information on the tzname global variable, see the tzset(3) man page.)
-   *
-   * The best way to deal with this issue (which is especially prominent
-   * on systems running glibc-2.3 or later, which is particularly ill-behaved
-   * in a chrooted environment, as it assumes the ability to find system
-   * timezone files at paths which are no longer valid within the chroot)
-   * is to set the TZ environment variable explicitly, before starting
-   * proftpd.  You can also use the SetEnv configuration directive within
-   * the proftpd.conf to set the TZ environment variable, e.g.:
-   *
-   *  SetEnv TZ PST
-   *
-   * To try to help sites which fail to do this, the tzname global variable
-   * will be copied prior to the mktime(3) call, and the copy restored after
-   * the call.  (Note that calling the ctime(3) and localtime(3) functions also
-   * causes a similar overwriting/setting of the tzname environment variable.)
-   */
-  memcpy(&tzname_dup, tzname, sizeof(tzname_dup));
-#endif /* HAVE_TZNAME */
+  /* Leap years.  */
+  { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
+};
 
-  env = pr_env_get(session.pool, "TZ");
+static unsigned long facts_secs_per_min(unsigned int min) {
+  unsigned long nsecs;
 
-  /* Set the TZ environment to be GMT, so that mktime(3) treats the timestamp
-   * provided by the client as being in GMT/UTC.
-   */
-  if (pr_env_set(session.pool, "TZ", "GMT") < 0) {
-    pr_log_debug(DEBUG8, MOD_FACTS_VERSION
-      ": error setting TZ environment variable to 'GMT': %s", strerror(errno));
-  }
+  nsecs = (min * FACTS_SECS_PER_MIN);
+  return nsecs;
+}
 
-  tm.tm_sec = sec;
-  tm.tm_min = min;
-  tm.tm_hour = hour;
-  tm.tm_mday = mday;
-  tm.tm_mon = (month - 1);
-  tm.tm_year = (year - 1900);
-  tm.tm_wday = 0;
-  tm.tm_yday = 0;
-  tm.tm_isdst = -1;
+static unsigned long facts_secs_per_hour(unsigned int hour) {
+  unsigned long nsecs;
 
-  res = mktime(&tm);
+  nsecs = (hour * FACTS_SECS_PER_HOUR);
+  return nsecs;
+}
 
-  /* Restore the old TZ setting, if any. */
-  if (env) {
-    if (pr_env_set(session.pool, "TZ", env) < 0) {
-      pr_log_debug(DEBUG8, MOD_FACTS_VERSION
-        ": error setting TZ environment variable to '%s': %s", env,
-        strerror(errno));
+static unsigned long facts_secs_per_day(unsigned long ndays) {
+  unsigned long nsecs;
+
+  nsecs = (ndays * FACTS_SECS_PER_DAY);
+  return nsecs;
+}
+
+/* Every 4th year is a leap year, except for every 100th year, but including
+ * every 400th year.
+ */
+static int facts_leap_year(unsigned int year) {
+  int leap_year = 0;
+
+  if ((year % 4) == 0) {
+    leap_year = 1;
+
+    if ((year % 100) == 0) {
+      leap_year = 0;
+
+      if ((year % 400) == 0) {
+        leap_year = 1;
+      }
     }
   }
 
-#ifdef HAVE_TZNAME
-  /* Restore the old tzname values prior to returning. */
-  memcpy(tzname, tzname_dup, sizeof(tzname_dup));
-#endif /* HAVE_TZNAME */
+  return leap_year;
+}
+
+static unsigned long facts_secs_per_mon(unsigned int mon, unsigned int year) {
+  int leap_year;
+  static unsigned int ndays;
+  static unsigned long nsecs;
+
+  leap_year = facts_leap_year(year);
+  ndays = facts_ydays_for_mon[leap_year][mon-1];
+
+  nsecs = facts_secs_per_day(ndays);
+  return nsecs;
+}
+
+static unsigned long facts_secs_per_year(unsigned int year) {
+  unsigned long ndays, nsecs;
+
+  ndays = (year - FACTS_EPOCH_YEAR) * 365;
+
+  /* Compute the number of leap days between 1970 and the given year
+   * (exclusive).  There is a leap day every 4th year...
+   */
+  ndays += (((year - 1) / 4) - (FACTS_EPOCH_YEAR / 4));
+
+  /* ...except every 100th year...*/
+  ndays -= (((year - 1) / 100) - (FACTS_EPOCH_YEAR / 100));
+
+  /* ...but still every 400th year. */
+  ndays += (((year - 1) / 400) - (FACTS_EPOCH_YEAR / 400));
+
+  nsecs = facts_secs_per_day(ndays);
+  return nsecs;
+}
+
+static time_t facts_mktime(unsigned int year, unsigned int month,
+    unsigned int mday, unsigned int hour, unsigned int min, unsigned int sec) {
+  time_t res;
+
+  /* Rather than using the system mktime(3) function (which requires external
+   * files such as /etc/localtime and the timezone definition files, depending
+   * on the TZ environment value setting), we use a custom mktime collection
+   * of functions.
+   *
+   * Fortunately, our homegrown collection of time conversion functions
+   * ONLY needs to generate GMT seconds here, and so we don't have to worry
+   * about DST, timezones, etc (Bug#3790).
+   */
+
+  res = facts_secs_per_year(year) +
+        facts_secs_per_mon(month, year) +
+
+        /* Subtract one day to make the mday zero-based. */
+        facts_secs_per_day(mday - 1) +
+
+        facts_secs_per_hour(hour) +
+        facts_secs_per_min(min) +
+        sec;
 
   return res;
 }
@@ -161,10 +209,10 @@ static size_t facts_mlinfo_fmt(struct mlinfo *info, char *buf, size_t bufsz) {
 
   memset(buf, '\0', bufsz);
 
-  ptr = buf + buflen;
+  ptr = buf;
 
   if (facts_opts & FACTS_OPT_SHOW_MODIFY) {
-    snprintf(ptr, bufsz - buflen, "modify=%04d%02d%02d%02d%02d%02d;",
+    snprintf(ptr, bufsz, "modify=%04d%02d%02d%02d%02d%02d;",
       info->tm->tm_year+1900, info->tm->tm_mon+1, info->tm->tm_mday,
       info->tm->tm_hour, info->tm->tm_min, info->tm->tm_sec);
     buflen = strlen(buf);
@@ -246,17 +294,20 @@ static size_t facts_mlinfo_fmt(struct mlinfo *info, char *buf, size_t bufsz) {
  * channel, wherease MLSD's output is sent via a data transfer, much like
  * LIST or NLST.
  */
-static char *mlinfo_buf = NULL;
+static char *mlinfo_buf = NULL, *mlinfo_bufptr = NULL;
 static size_t mlinfo_bufsz = 0;
 static size_t mlinfo_buflen = 0;
 
 static void facts_mlinfobuf_init(void) {
   if (mlinfo_buf == NULL) {
-    mlinfo_bufsz = pr_config_get_xfer_bufsz();
+    mlinfo_bufsz = pr_config_get_server_xfer_bufsz(PR_NETIO_IO_WR);
     mlinfo_buf = palloc(session.pool, mlinfo_bufsz);
+    pr_trace_msg("data", 8, "allocated facts buffer of %lu bytes",
+      (unsigned long) mlinfo_bufsz);
   }
 
   memset(mlinfo_buf, '\0', mlinfo_bufsz);
+  mlinfo_bufptr = mlinfo_buf;
   mlinfo_buflen = 0;
 }
 
@@ -273,7 +324,8 @@ static void facts_mlinfobuf_add(struct mlinfo *info) {
     (void) facts_mlinfobuf_flush();
   }
 
-  sstrcat(mlinfo_buf, buf, mlinfo_bufsz);
+  sstrcat(mlinfo_bufptr, buf, mlinfo_bufsz - mlinfo_buflen);
+  mlinfo_bufptr += buflen;
   mlinfo_buflen += buflen;
 }
 
@@ -281,31 +333,143 @@ static void facts_mlinfobuf_flush(void) {
   if (mlinfo_buflen > 0) {
     int res;
 
+    /* Make sure the ASCII flags are cleared from the session flags,
+     * so that the pr_data_xfer() function does not try to perform
+     * ASCII translation on this data.
+     */
+    session.sf_flags &= ~SF_ASCII_OVERRIDE;
+
     res = pr_data_xfer(mlinfo_buf, mlinfo_buflen);
     if (res < 0 &&
         errno != 0) {
       pr_log_debug(DEBUG3, MOD_FACTS_VERSION
         ": error transferring data: [%d] %s", errno, strerror(errno));
     }
+
+    session.sf_flags |= SF_ASCII_OVERRIDE;
   }
 
   facts_mlinfobuf_init();
 }
 
 static int facts_mlinfo_get(struct mlinfo *info, const char *path,
-    const char *dent_name) {
+    const char *dent_name, int flags, uid_t uid, gid_t gid, mode_t *mode) {
   char *perm = "";
+  int res;
 
-  if (pr_fsio_stat(path, &(info->st)) < 0) {
-    pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": error stat'ing '%s': %s",
-      path, strerror(errno));
+  res = pr_fsio_lstat(path, &(info->st));
+  if (res < 0) {
+    int xerrno = errno;
+
+    pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": error lstat'ing '%s': %s",
+      path, strerror(xerrno));
+
+    errno = xerrno;
     return -1;
   }
 
+  if (uid != (uid_t) -1) {
+    info->st.st_uid = uid;
+  }
+
+  if (gid != (gid_t) -1) {
+    info->st.st_gid = gid;
+  }
+
   info->tm = pr_gmtime(info->pool, &(info->st.st_mtime));
 
   if (!S_ISDIR(info->st.st_mode)) {
+#ifdef S_ISLNK
+    if (S_ISLNK(info->st.st_mode)) {
+      struct stat target_st;
+
+      /* Now we need to use stat(2) on the path (versus lstat(2)) to get the
+       * info for the target, and copy its st_dev and st_ino values to our
+       * stat in order to ensure that the unique fact values are the same.
+       */
+
+      pr_fs_clear_cache();
+      res = pr_fsio_stat(path, &target_st);
+      if (res < 0) {
+        int xerrno = errno;
+
+        pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": error stat'ing '%s': %s",
+          path, strerror(xerrno));
+
+        errno = xerrno;
+        return -1;
+      }
+
+      info->st.st_dev = target_st.st_dev;
+      info->st.st_ino = target_st.st_ino;
+
+      if (flags & FACTS_MLINFO_FL_SHOW_SYMLINKS) {
+
+        /* Do we use the proper RFC 3659 syntax (i.e. following the BNF rules
+         * of RFC 3659), which would be:
+         *
+         *   type=OS.unix=symlink
+         *
+         * See:
+         *   http://www.rfc-editor.org/errata_search.php?rfc=3659
+         *
+         * and search for "OS.unix=slink".
+         *
+         * Or do we use the syntax in the _examples_ presented in RFC 3659,
+         * which is what clients such as FileZilla expect:
+         *
+         *   type=OS.unix=slink:<target>
+         *
+         * See:
+         *   http://trac.filezilla-project.org/ticket/4490
+         */
+
+        if (flags & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
+          char target[PR_TUNABLE_PATH_MAX+1];
+          int targetlen;
+
+          targetlen = pr_fsio_readlink(path, target, sizeof(target)-1);
+          if (targetlen < 0) { 
+            int xerrno = errno;
+
+            pr_log_debug(DEBUG4, MOD_FACTS_VERSION
+              ": error reading symlink '%s': %s", path, strerror(xerrno));
+
+            errno = xerrno;
+            return -1;
+          }
+
+          if (targetlen >= sizeof(target)-1) {
+            targetlen = sizeof(target)-1;
+          }
+
+          target[targetlen] = '\0';
+
+          info->type = pstrcat(info->pool, "OS.unix=slink:",
+            dir_best_path(info->pool, target), NULL);
+
+        } else {
+          /* Use the proper syntax.  Too bad for the not-really-compliant
+           * FileZilla.
+           */
+          info->type = "OS.unix=symlink";
+        }
+
+      } else {
+        if (S_ISDIR(target_st.st_mode)) {
+          info->type = "dir";
+
+        } else {
+          info->type = "file";
+        }
+      }
+
+    } else {
+      info->type = "file";
+    }
+#else
     info->type = "file";
+#endif
 
     if (pr_fsio_access(path, R_OK, session.uid, session.gid,
         session.gids) == 0) {
@@ -362,6 +526,15 @@ static int facts_mlinfo_get(struct mlinfo *info, const char *path,
   }
 
   info->perm = perm;
+
+  if (mode != NULL) {
+    /* We cheat here by simply overwriting the entire st.st_mode value with
+     * the DirFakeMode.  This works because later operations on this data
+     * don't pay attention to the file type.
+     */
+    info->st.st_mode = *mode;
+  }
+
   return 0;
 }
 
@@ -468,6 +641,7 @@ static int facts_modify_mtime(pool *p, const char *path, char *timestamp) {
   char c, *ptr;
   unsigned int year, month, day, hour, min, sec;
   struct timeval tvs[2];
+  int res;
 
   (void) p;
 
@@ -477,6 +651,13 @@ static int facts_modify_mtime(pool *p, const char *path, char *timestamp) {
   year = atoi(ptr);
   timestamp[4] = c;
 
+  if (year < FACTS_EPOCH_YEAR) {
+    pr_log_debug(DEBUG8, MOD_FACTS_VERSION
+      ": bad year value (%d) in timestamp '%s'", year, timestamp);
+    errno = EINVAL;
+    return -1;
+  }
+
   ptr = &(timestamp[4]);
   c = timestamp[6];
   timestamp[6] = '\0';
@@ -548,10 +729,81 @@ static int facts_modify_mtime(pool *p, const char *path, char *timestamp) {
   tvs[0].tv_sec = tvs[1].tv_sec = facts_mktime(year, month, day, hour, min,
     sec);
 
-  if (pr_fsio_utimes(path, tvs) < 0) {
-    pr_log_debug(DEBUG2, MOD_FACTS_VERSION
-      ": error modifying modify fact for '%s': %s", path, strerror(errno));
-    return -1;
+  res = pr_fsio_utimes(path, tvs);
+  if (res < 0) {
+    int xerrno = errno;
+
+    if (xerrno == EPERM) {
+      struct stat st;
+      int matching_gid = FALSE;
+
+      /* If the utimes(2) call failed because the process UID doesn't
+       * match the file UID, then check to see if the GIDs match (and that
+       * the file has group write permissions).
+       *
+       * This can be alleviated in two ways: a) if mod_cap is present,
+       * enable the CAP_FOWNER capability for the session, or b) use root
+       * privs.
+       */
+      pr_fs_clear_cache();
+      if (pr_fsio_stat(path, &st) < 0) {
+        errno = xerrno;
+        return -1;
+      }
+
+      /* Be sure to check the primary and all the supplemental groups to
+       * which this session belongs.
+       */
+      if (st.st_gid == session.gid) {
+        matching_gid = TRUE;
+
+      } else {
+        register unsigned int i;
+        gid_t *gids;
+
+        gids = session.gids->elts;
+        for (i = 0; i < session.gids->nelts; i++) {
+          if (st.st_gid == gids[i]) {
+            matching_gid = TRUE;
+            break;
+          }
+        }
+      }
+
+      if (matching_gid == TRUE &&
+          (st.st_mode & S_IWGRP)) {
+        int merrno = 0;
+
+        /* Try the utimes(2) call again, this time with root privs. */
+
+        pr_signals_block();
+        PRIVS_ROOT
+        res = pr_fsio_utimes(path, tvs);
+        if (res < 0) {
+          merrno = errno;
+        }
+        PRIVS_RELINQUISH
+        pr_signals_unblock();
+
+        if (res == 0)
+          return 0;
+
+        pr_log_debug(DEBUG2, MOD_FACTS_VERSION
+          ": error modifying modify fact for '%s': %s", path,
+          strerror(merrno));
+        errno = xerrno;
+        return -1;
+      }
+
+      errno = xerrno;
+      return -1;
+
+    } else {
+      pr_log_debug(DEBUG2, MOD_FACTS_VERSION
+        ": error modifying modify fact for '%s': %s", path, strerror(xerrno));
+      errno = xerrno;
+      return -1;
+    }
   }
 
   return 0;
@@ -608,7 +860,7 @@ static int facts_modify_unix_mode(pool *p, const char *path, char *mode_str) {
  */
 
 MODRET facts_mff(cmd_rec *cmd) {
-  const char *path, *decoded_path;
+  const char *path, *canon_path, *decoded_path;
   char *facts, *ptr;
 
   if (cmd->argc < 3) {
@@ -623,11 +875,22 @@ MODRET facts_mff(cmd_rec *cmd) {
    * find the FIRST space in cmd->arg; the path is everything past that space.
    */
   ptr = strchr(cmd->arg, ' ');
+  if (ptr == NULL) {
+    pr_response_add_err(R_501, _("Invalid command syntax"));
+    return PR_ERROR(cmd);
+  }
+
   path = pstrdup(cmd->tmp_pool, ptr + 1);
 
   decoded_path = pr_fs_decode_path(cmd->tmp_pool, path);
 
-  if (!dir_check(cmd->tmp_pool, cmd, cmd->group, (char *) decoded_path, NULL)) {
+  canon_path = dir_canonical_path(cmd->tmp_pool, decoded_path);
+  if (canon_path == NULL) {
+    pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EINVAL));
+    return PR_ERROR(cmd);
+  }
+
+  if (!dir_check(cmd->tmp_pool, cmd, cmd->group, canon_path, NULL)) {
     pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": %s command denied by <Limit>",
       cmd->argv[0]);
     pr_response_add_err(R_550, _("Unable to handle command"));
@@ -677,8 +940,12 @@ MODRET facts_mff(cmd_rec *cmd) {
       }
 
       if (facts_modify_mtime(cmd->tmp_pool, decoded_path, timestamp) < 0) {
-        pr_response_add_err(errno == ENOENT ? R_550 : R_501, "%s: %s", path,
-          strerror(errno));
+        int xerrno = errno;
+
+        pr_response_add_err(xerrno == ENOENT ? R_550 : R_501, "%s: %s", path,
+          strerror(xerrno));
+
+        errno = xerrno;
         return PR_ERROR(cmd);
       }
 
@@ -749,7 +1016,7 @@ MODRET facts_mff(cmd_rec *cmd) {
 }
 
 MODRET facts_mfmt(cmd_rec *cmd) {
-  const char *path, *decoded_path;
+  const char *path, *canon_path, *decoded_path;
   char *timestamp, *ptr;
   int res;
 
@@ -765,11 +1032,22 @@ MODRET facts_mfmt(cmd_rec *cmd) {
    * find the FIRST space in cmd->arg; the path is everything past that space.
    */
   ptr = strchr(cmd->arg, ' ');
+  if (ptr == NULL) {
+    pr_response_add_err(R_501, _("Invalid command syntax"));
+    return PR_ERROR(cmd);
+  }
+
   path = pstrdup(cmd->tmp_pool, ptr + 1);
 
   decoded_path = pr_fs_decode_path(cmd->tmp_pool, path);
 
-  if (!dir_check(cmd->tmp_pool, cmd, cmd->group, (char *) decoded_path, NULL)) {
+  canon_path = dir_canonical_path(cmd->tmp_pool, decoded_path);
+  if (canon_path == NULL) {
+    pr_response_add_err(R_550, "%s: %s", cmd->arg, strerror(EINVAL));
+    return PR_ERROR(cmd);
+  }
+
+  if (!dir_check(cmd->tmp_pool, cmd, cmd->group, canon_path, NULL)) {
     pr_log_debug(DEBUG4, MOD_FACTS_VERSION ": %s command denied by <Limit>",
       cmd->argv[0]);
     pr_response_add_err(R_550, _("Unable to handle command"));
@@ -796,12 +1074,16 @@ MODRET facts_mfmt(cmd_rec *cmd) {
 
   res = facts_modify_mtime(cmd->tmp_pool, decoded_path, timestamp);
   if (res < 0) {
+    int xerrno = errno;
+
     if (ptr) {
       *ptr = '.';
     }
 
-    pr_response_add_err(errno == ENOENT ? R_550 : R_501, "%s: %s", path,
-      strerror(errno));
+    pr_response_add_err(xerrno == ENOENT ? R_550 : R_501, "%s: %s", path,
+      strerror(xerrno));
+
+    errno = xerrno;
     return PR_ERROR(cmd);
   }
 
@@ -823,7 +1105,13 @@ MODRET facts_mfmt(cmd_rec *cmd) {
 
 MODRET facts_mlsd(cmd_rec *cmd) {
   const char *path, *decoded_path, *best_path;
+  config_rec *c;
+  uid_t fake_uid = -1;
+  gid_t fake_gid = -1;
+  mode_t *fake_mode = NULL;
   struct mlinfo info;
+  unsigned char *ptr;
+  int flags = 0;
   DIR *dirh;
   struct dirent *dent;
 
@@ -867,8 +1155,73 @@ MODRET facts_mlsd(cmd_rec *cmd) {
     return PR_ERROR(cmd);
   }
 
+  /* Determine whether to display symlinks as such. */
+  ptr = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE);
+  if (ptr != NULL) {
+    if (*ptr == TRUE) {
+      flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS;
+
+      if (facts_mlinfo_opts & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
+        flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK;
+      }
+    }
+
+  } else {
+    /* ShowSymlinks is documented as being 'on' by default. */
+    flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS;
+
+    if (facts_mlinfo_opts & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
+      flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK;
+    }
+  }
+
   best_path = dir_best_path(cmd->tmp_pool, decoded_path);
 
+  fake_mode = get_param_ptr(get_dir_ctxt(cmd->tmp_pool, (char *) best_path),
+    "DirFakeMode", FALSE);
+ 
+  c = find_config(get_dir_ctxt(cmd->tmp_pool, (char *) best_path), CONF_PARAM,
+    "DirFakeUser", FALSE);
+  if (c) {
+    if (c->argc > 0) {
+      const char *fake_user = NULL;
+
+      fake_user = c->argv[0];
+      if (fake_user != NULL &&
+          strncmp(fake_user, "~", 2) != 0) {
+        fake_uid = pr_auth_name2uid(cmd->tmp_pool, fake_user);
+
+      } else {
+        fake_uid = session.uid;
+      }
+
+    } else {
+      /* Handle the "DirFakeUser off" case (Bug#3715). */
+      fake_uid = (uid_t) -1;
+    }
+  }
+
+  c = find_config(get_dir_ctxt(cmd->tmp_pool, (char *) best_path), CONF_PARAM,
+    "DirFakeGroup", FALSE);
+  if (c) {
+    if (c->argc > 0) {
+      const char *fake_group = NULL;
+
+      fake_group = c->argv[0];
+      if (fake_group != NULL &&
+          strncmp(fake_group, "~", 2) != 0) {
+        fake_gid = pr_auth_name2gid(cmd->tmp_pool, fake_group);
+
+      } else {
+        fake_gid = session.gid;
+      }
+
+    } else {
+      /* Handle the "DirFakeGroup off" case (Bug#3715). */
+      fake_gid = (gid_t) -1;
+    }
+  }
+
   dirh = pr_fsio_opendir(best_path);
   if (dirh == NULL) {
     int xerrno = errno;
@@ -879,15 +1232,17 @@ MODRET facts_mlsd(cmd_rec *cmd) {
       best_path, strerror(xerrno));
 
     pr_response_add_err(R_550, "%s: %s", path, strerror(xerrno));
+
+    errno = xerrno;
     return PR_ERROR(cmd);
   }
 
   /* Open data connection */
-  session.sf_flags |= SF_ASCII_OVERRIDE;
   if (pr_data_open(NULL, C_MLSD, PR_NETIO_IO_WR, 0) < 0) {
     pr_fsio_closedir(dirh);
     return PR_ERROR(cmd);
   }
+  session.sf_flags |= SF_ASCII_OVERRIDE;
 
   pr_fs_clear_cache();
   facts_mlinfobuf_init();
@@ -899,6 +1254,10 @@ MODRET facts_mlsd(cmd_rec *cmd) {
     pr_signals_handle();
 
     rel_path = pdircat(cmd->tmp_pool, best_path, dent->d_name, NULL);
+    res = dir_check(cmd->tmp_pool, cmd, cmd->group, rel_path, &hidden);
+    if (!res || hidden) {
+      continue;
+    }
 
     /* Check that the file can be listed. */
     abs_path = dir_realpath(cmd->tmp_pool, rel_path);
@@ -907,8 +1266,9 @@ MODRET facts_mlsd(cmd_rec *cmd) {
       
     } else {
       abs_path = dir_canonical_path(cmd->tmp_pool, rel_path);
-      if (abs_path == NULL)
+      if (abs_path == NULL) {
         abs_path = rel_path;
+      }
 
       res = dir_check_canon(cmd->tmp_pool, cmd, cmd->group, abs_path, &hidden);
     }
@@ -920,7 +1280,8 @@ MODRET facts_mlsd(cmd_rec *cmd) {
     memset(&info, 0, sizeof(struct mlinfo));
 
     info.pool = cmd->tmp_pool;
-    if (facts_mlinfo_get(&info, rel_path, dent->d_name) < 0) {
+    if (facts_mlinfo_get(&info, rel_path, dent->d_name, flags,
+        fake_uid, fake_gid, fake_mode) < 0) {
       pr_log_debug(DEBUG3, MOD_FACTS_VERSION
         ": MLSD: unable to get info for '%s': %s", abs_path, strerror(errno));
       continue;
@@ -953,6 +1314,15 @@ MODRET facts_mlsd(cmd_rec *cmd) {
 }
 
 MODRET facts_mlsd_cleanup(cmd_rec *cmd) {
+  const char *proto;
+
+  proto = pr_session_get_protocol(0);
+
+  /* Ignore this for SFTP connections. */
+  if (strncmp(proto, "sftp", 5) == 0) {
+    return PR_DECLINED(cmd);
+  }
+
   if (session.xfer.p) {
     destroy_pool(session.xfer.p);
   }
@@ -962,7 +1332,12 @@ MODRET facts_mlsd_cleanup(cmd_rec *cmd) {
 }
 
 MODRET facts_mlst(cmd_rec *cmd) {
-  int hidden = FALSE;
+  int flags = 0, hidden = FALSE;
+  config_rec *c;
+  uid_t fake_uid = -1;
+  gid_t fake_gid = -1;
+  mode_t *fake_mode = NULL;
+  unsigned char *ptr;
   const char *path, *decoded_path;
   struct mlinfo info;
 
@@ -992,18 +1367,95 @@ MODRET facts_mlst(cmd_rec *cmd) {
     return PR_HANDLED(cmd);
   }
 
+  /* Determine whether to display symlinks as such. */
+  ptr = get_param_ptr(TOPLEVEL_CONF, "ShowSymlinks", FALSE);
+  if (ptr &&
+      *ptr == TRUE) {
+    flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS;
+
+    if (facts_mlinfo_opts & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
+      flags |= FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK;
+    }
+  }
+
+  fake_mode = get_param_ptr(get_dir_ctxt(cmd->tmp_pool, (char *) decoded_path),
+    "DirFakeMode", FALSE);
+
+  c = find_config(get_dir_ctxt(cmd->tmp_pool, (char *) decoded_path),
+    CONF_PARAM, "DirFakeUser", FALSE);
+  if (c) {
+    if (c->argc > 0) {
+      const char *fake_user;
+
+      fake_user = c->argv[0];
+      if (fake_user != NULL &&
+          strncmp(fake_user, "~", 2) != 0) {
+        fake_uid = pr_auth_name2uid(cmd->tmp_pool, fake_user);
+
+      } else {
+        fake_uid = session.uid;
+      }
+
+    } else {
+      /* Handle the "DirFakeUser off" case (Bug#3715). */
+      fake_uid = session.uid;
+    }
+  }
+
+  c = find_config(get_dir_ctxt(cmd->tmp_pool, (char *) decoded_path),
+    CONF_PARAM, "DirFakeGroup", FALSE);
+  if (c) {
+    if (c->argc > 0) {
+      const char *fake_group;
+
+      fake_group = c->argv[0];
+      if (fake_group != NULL &&
+          strncmp(fake_group, "~", 2) != 0) {
+        fake_gid = pr_auth_name2gid(cmd->tmp_pool, fake_group);
+
+      } else {
+        fake_gid = session.gid;
+      }
+
+    } else {
+      /* Handle the "DirFakeGroup off" case (Bug#3715). */
+      fake_gid = session.gid;
+    }
+  }
+
   info.pool = cmd->tmp_pool;
 
   pr_fs_clear_cache();
-  if (facts_mlinfo_get(&info, decoded_path, decoded_path) < 0) {
+  if (facts_mlinfo_get(&info, decoded_path, decoded_path, flags, fake_uid,
+      fake_gid, fake_mode) < 0) {
     pr_response_add_err(R_550, _("'%s' cannot be listed"), path);
     return PR_ERROR(cmd);
   }
 
   /* No need to re-encode the path here as UTF8, since 'path' is the
    * original parameter as sent by the client.
+   *
+   * However, as per RFC3659 Section 7.3.1, since we advertise TVFS in our
+   * FEAT output, the path here should be the full path (as seen by the
+   * client).
    */
-  info.path = path;
+
+  /* XXX What about chroots? */
+
+  if (flags & FACTS_MLINFO_FL_SHOW_SYMLINKS) {
+    if (flags & FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK) {
+      info.path = dir_canonical_path(cmd->tmp_pool, path);
+
+    } else {
+      /* If we are supposed to show symlinks, then use dir_best_path() to get
+       * the full path, including dereferencing the symlink.
+       */
+      info.path = dir_best_path(cmd->tmp_pool, path);
+    }
+
+  } else {
+    info.path = dir_canonical_path(cmd->tmp_pool, path);
+  }
 
   pr_response_add(R_250, _("Start of list for %s"), path);
   facts_mlinfo_add(&info);
@@ -1107,6 +1559,27 @@ MODRET facts_opts_mlst(cmd_rec *cmd) {
   return PR_HANDLED(cmd);
 }
 
+MODRET facts_post_host(cmd_rec *cmd) {
+
+  /* If the HOST command changed the main_server pointer, reinitialize
+   * ourselves.
+   */
+  if (session.prev_server != NULL) {
+    int res;
+
+    facts_opts = 0;
+    facts_mlinfo_opts = 0;
+
+    res = facts_sess_init();
+    if (res < 0) {
+      pr_session_disconnect(&facts_module,
+        PR_SESS_DISCONNECT_SESSION_INIT_FAILED, NULL);
+    }
+  }
+
+  return PR_DECLINED(cmd);
+}
+
 /* Configuration handlers
  */
 
@@ -1129,12 +1602,42 @@ MODRET set_factsadvertise(cmd_rec *cmd) {
   return PR_HANDLED(cmd);
 }
 
+/* usage: FactsOptions opt1 ... optN */
+MODRET set_factsoptions(cmd_rec *cmd) {
+  register unsigned int i;
+  config_rec *c;
+  unsigned long opts = 0UL;
+
+  if (cmd->argc-1 == 0) {
+    CONF_ERROR(cmd, "wrong number of parameters");
+  }
+
+  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);
+
+  c = add_config_param(cmd->argv[0], 1, NULL);
+
+  for (i = 1; i < cmd->argc; i++) {
+    if (strncmp(cmd->argv[i], "UseSlink", 9) == 0) {
+      opts |= FACTS_MLINFO_FL_SHOW_SYMLINKS_USE_SLINK;
+
+    } else {
+      CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown FactsOption '",
+        cmd->argv[i], "'", NULL));
+    }
+  }
+
+  c->argv[0] = palloc(c->pool, sizeof(unsigned long));
+  *((unsigned long *) c->argv[0]) = opts;
+
+  return PR_HANDLED(cmd);
+}
+
 /* Initialization functions
  */
 
 static int facts_init(void) {
-  pr_help_add("MLSD", "[<sp> pathname]", TRUE);
-  pr_help_add("MLST", "[<sp> pathname]", TRUE);
+  pr_help_add(C_MLSD, _("[<sp> pathname]"), TRUE);
+  pr_help_add(C_MLST, _("[<sp> pathname]"), TRUE);
 
   return 0;
 }
@@ -1148,8 +1651,21 @@ static int facts_sess_init(void) {
     advertise = *((int *) c->argv[0]);
   }
 
-  if (advertise == FALSE)
+  if (advertise == FALSE) {
     return 0;
+  }
+
+  c = find_config(main_server->conf, CONF_PARAM, "FactsOptions", FALSE);
+  while (c != NULL) {
+    unsigned long opts;
+
+    pr_signals_handle();
+  
+    opts = *((unsigned long *) c->argv[0]);
+    facts_mlinfo_opts |= opts;
+
+    c = find_config_next(c, c->next, CONF_PARAM, "FactsOptions", FALSE);
+  }
 
   facts_opts = FACTS_OPT_SHOW_MODIFY|FACTS_OPT_SHOW_PERM|FACTS_OPT_SHOW_SIZE|
     FACTS_OPT_SHOW_TYPE|FACTS_OPT_SHOW_UNIQUE|FACTS_OPT_SHOW_UNIX_GROUP|
@@ -1173,17 +1689,19 @@ static int facts_sess_init(void) {
 
 static conftable facts_conftab[] = {
   { "FactsAdvertise",	set_factsadvertise,	NULL },
+  { "FactsOptions",	set_factsoptions,	NULL },
   { NULL }
 };
 
 static cmdtable facts_cmdtab[] = {
-  { CMD,	"MFF",		G_WRITE,facts_mff,  TRUE, FALSE, CL_WRITE },
-  { CMD,	"MFMT",		G_WRITE,facts_mfmt, TRUE, FALSE, CL_WRITE },
+  { CMD,	C_MFF,		G_WRITE,facts_mff,  TRUE, FALSE, CL_WRITE },
+  { CMD,	C_MFMT,		G_WRITE,facts_mfmt, TRUE, FALSE, CL_WRITE },
   { CMD,	C_MLSD,		G_DIRS,	facts_mlsd, TRUE, FALSE, CL_DIRS },
   { LOG_CMD,	C_MLSD,		G_NONE,	facts_mlsd_cleanup, FALSE, FALSE },
   { LOG_CMD_ERR,C_MLSD,		G_NONE,	facts_mlsd_cleanup, FALSE, FALSE },
   { CMD,	C_MLST,		G_DIRS,	facts_mlst, TRUE, FALSE, CL_DIRS },
   { CMD,	C_OPTS "_MLST", G_NONE, facts_opts_mlst, FALSE, FALSE },
+  { POST_CMD,	C_HOST,		G_NONE,	facts_post_host, FALSE, FALSE },
   { 0, NULL }
 };
 
