@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004-2010 The ProFTPD Project team
+ * Copyright (c) 2004-2014 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,30 +23,124 @@
  */
 
 /* Display of files
- * $Id: display.c,v 1.16.2.2 2010/04/17 17:11:14 castaglia Exp $
+ * $Id: display.c,v 1.32 2013-10-07 05:51:30 castaglia Exp $
  */
 
 #include "conf.h"
 
+static int first_msg_sent = FALSE;
+static const char *first_msg = NULL;
+static const char *prev_msg = NULL;
+
+/* Note: The size provided by pr_fs_getsize2() is in KB, not bytes. */
 static void format_size_str(char *buf, size_t buflen, off_t size) {
-  char units[] = {'K', 'M', 'G', 'T', 'P'};
+  char *units[] = {"K", "M", "G", "T", "P"};
+  unsigned int nunits = 5;
   register unsigned int i = 0;
+  int res;
 
   /* Determine the appropriate units label to use. */
-  while (size > 1024) {
+  while (size > 1024 &&
+         i < nunits) {
+    pr_signals_handle();
+
     size /= 1024;
     i++;
   }
 
   /* Now, prepare the buffer. */
-  snprintf(buf, buflen, "%.3" PR_LU "%cB", (pr_off_t) size, units[i]);
+  res = snprintf(buf, buflen, "%.3" PR_LU "%sB", (pr_off_t) size, units[i]);
+
+  if (res > 2) {
+    /* Check for leading zeroes; it's an aethetic choice. */
+    if (buf[0] == '0' && buf[1] != '.') {
+      memmove(&buf[0], &buf[1], res-1);
+      buf[res-1] = '\0';
+    }
+  }
+}
+
+static int display_add_line(pool *p, const char *resp_code,
+    const char *resp_msg) {
+
+  /* Handle the case where the data to Display might contain only one line. */
+
+  if (first_msg_sent == FALSE &&
+      first_msg == NULL) {
+      first_msg = pstrdup(p, resp_msg);
+    return 0;
+  }
+
+  if (first_msg != NULL) {
+    pr_response_send_raw("%s-%s", resp_code, first_msg);
+    first_msg = NULL;
+    first_msg_sent = TRUE;
+
+    prev_msg = pstrdup(p, resp_msg);
+    return 0; 
+  }
+
+  if (prev_msg != NULL) {
+    if (session.multiline_rfc2228) {
+      pr_response_send_raw("%s-%s", resp_code, prev_msg);
+
+    } else {
+      pr_response_send_raw(" %s", prev_msg);
+    }
+  }
+
+  prev_msg = pstrdup(p, resp_msg);
+  return 0;
+}
+
+static int display_flush_lines(pool *p, const char *resp_code, int flags) {
+  if (first_msg != NULL) {
+    if (session.auth_mech != NULL) {
+      if (flags & PR_DISPLAY_FL_NO_EOM) {
+        pr_response_send_raw("%s-%s", resp_code, first_msg);
+
+      } else {
+        pr_response_send_raw("%s %s", resp_code, first_msg);
+      }
+
+    } else {
+      /* There is a special case if the client has not yet authenticated; it
+       * means we are handling a DisplayConnect file.  The server will send
+       * a banner as well, so we need to treat this is the start of a multiline
+       * response.
+       */
+      pr_response_send_raw("%s-%s", resp_code, first_msg);
+    }
+
+  } else {
+    if (prev_msg) {
+      if (session.multiline_rfc2228) {
+        pr_response_send_raw("%s-%s", resp_code, prev_msg);
+
+      } else {
+        if (flags & PR_DISPLAY_FL_NO_EOM) {
+          pr_response_send_raw(" %s", prev_msg);
+
+        } else {
+          pr_response_send_raw("%s %s", resp_code, prev_msg);
+        }
+      }
+    }
+  }
+
+  /* Reset state for the next set of lines. */
+  first_msg_sent = FALSE;
+  first_msg = NULL;
+  prev_msg = NULL;
+
+  return 0;
 }
 
 static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
     int flags) {
   struct stat st;
   char buf[PR_TUNABLE_BUFFER_SIZE] = {'\0'};
-  int len;
+  int len, res;
   unsigned int *current_clients = NULL;
   unsigned int *max_clients = NULL;
   off_t fs_size = 0;
@@ -64,23 +158,23 @@ static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
   char mg_xfer_units[12] = {'\0'}, *user;
   const char *mg_time;
   char *rfc1413_ident = NULL;
-  char *first = NULL, *prev = NULL;
-  int sent_first = FALSE;
 
   /* Stat the opened file to determine the optimal buffer size for IO. */
   memset(&st, 0, sizeof(st));
   pr_fsio_fstat(fh, &st);
   fh->fh_iosz = st.st_blksize;
 
-#if defined(HAVE_STATFS) || defined(HAVE_SYS_STATVFS_H) || \
-   defined(HAVE_SYS_VFS_H)
-  fs_size = pr_fs_getsize((fs ? (char *) fs : (char *) fh->fh_path));
+  /* Note: The size provided by pr_fs_getsize() is in KB, not bytes. */
+  res = pr_fs_fgetsize(fh->fh_fd, &fs_size);
+  if (res < 0 &&
+      errno != ENOSYS) {
+    (void) pr_log_debug(DEBUG7, "error getting filesystem size for '%s': %s",
+      fh->fh_path, strerror(errno));
+    fs_size = 0;
+  }
+
   snprintf(mg_size, sizeof(mg_size), "%" PR_LU, (pr_off_t) fs_size);
   format_size_str(mg_size_units, sizeof(mg_size_units), fs_size);
-#else
-  snprintf(mg_size, sizeof(mg_size), "%" PR_LU, (pr_off_t) fs_size);
-  format_size_str(mg_size_units, sizeof(mg_size_units), fs_size);
-#endif
 
   p = make_sub_pool(session.pool);
   pr_pool_tag(p, "Display Pool");
@@ -98,7 +192,8 @@ static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
 
   snprintf(mg_cur, sizeof(mg_cur), "%u", current_clients ? *current_clients: 1);
 
-  if (session.class && session.class->cls_name) {
+  if (session.conn_class != NULL &&
+      session.conn_class->cls_name != NULL) {
     unsigned int *class_clients = NULL;
     config_rec *maxc = NULL;
     unsigned int maxclients = 0;
@@ -120,7 +215,9 @@ static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
       FALSE);
 
     while (maxc) {
-      if (strcmp(maxc->argv[0], session.class->cls_name) != 0) {
+      pr_signals_handle();
+
+      if (strcmp(maxc->argv[0], session.conn_class->cls_name) != 0) {
         maxc = find_config_next(maxc, maxc->next, CONF_PARAM,
           "MaxClientsPerClass", FALSE);
         continue;
@@ -201,7 +298,8 @@ static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
     buf[sizeof(buf)-1] = '\0';
     len = strlen(buf);
 
-    while(len && (buf[len-1] == '\r' || buf[len-1] == '\n')) {
+    while (len &&
+           (buf[len-1] == '\r' || buf[len-1] == '\n')) {
       buf[len-1] = '\0';
       len--;
     }
@@ -263,7 +361,8 @@ static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
         val = pr_var_get(key);
         if (val == NULL) {
           pr_trace_msg("var", 4,
-            "no value set for name '%s', using \"(none)\"", key);
+            "no value set for name '%s' [%s], using \"(none)\"", key,
+            strerror(errno));
           val = "(none)";
         }
       }
@@ -293,82 +392,29 @@ static int display_fh(pr_fh_t *fh, const char *fs, const char *code,
       "%U", user,
       "%u", rfc1413_ident,
       "%V", main_server->ServerName,
-      "%x", session.class ? session.class->cls_name : "(unknown)",
+      "%x", session.conn_class ? session.conn_class->cls_name : "(unknown)",
       "%y", mg_cur_class,
       "%z", mg_class_limit,
       NULL);
 
     sstrncpy(buf, outs, sizeof(buf));
 
-    /* Handle the case where the Display file might contain only one
-     * line.
-     *
-     * We _could_ just use pr_response_add(), and let the response code
-     * automatically handle all of the multiline response formatting.
-     * However, some of the Display files are at times waiting for the
-     * response chains to be flushed, which won't work (e.g. login, logout).
-     * Thus we have to deal with multiline files appropriately here.
-     */
-
-    if (sent_first == FALSE &&
-        first == NULL) {
-      first = pstrdup(p, outs);
-      continue;
-    }
-
-    if (first) {
-      pr_response_send_raw("%s-%s", code, first);
-      first = NULL;
-      sent_first = TRUE;
-
-      prev = pstrdup(p, outs);
-      continue;
-    }
-
-    if (prev) {
-      if (MultilineRFC2228) {
-        pr_response_send_raw("%s-%s", code, prev);
-
-      } else {
-        pr_response_send_raw(" %s", prev);
-      }
-    }
-
-    prev = pstrdup(p, outs);
-  }
-
-  if (first != NULL) {
-    if (session.auth_mech != NULL) {
-      if (flags & PR_DISPLAY_FL_NO_EOM) {
-        pr_response_send_raw("%s-%s", code, first);
-
-      } else {
-        pr_response_send_raw("%s %s", code, first);
-      }
+    if (flags & PR_DISPLAY_FL_SEND_NOW) {
+      /* Normally we use pr_response_add(), and let the response code
+       * automatically handle all of the multiline response formatting.
+       * However, some of the Display files are at times waiting for the
+       * response chains to be flushed, which won't work (i.e. DisplayConnect
+       * and DisplayQuit).
+       */
+      display_add_line(p, code, outs);
 
     } else {
-      /* There is a special case if the client has not yet authenticated; it
-       * means we are handling a DisplayConnect file.  The server will send
-       * a banner as well, so we need to treat this is the start of a multiline
-       * response.
-       */
-      pr_response_send_raw("%s-%s", code, first);
+      pr_response_add(code, "%s", outs);
     }
+  }
 
-  } else {
-    if (prev) {
-      if (MultilineRFC2228) {
-        pr_response_send_raw("%s-%s", code, prev);
-
-      } else {
-        if (flags & PR_DISPLAY_FL_NO_EOM) {
-          pr_response_send_raw(" %s", prev);
-
-        } else {
-          pr_response_send_raw("%s %s", code, prev);
-        }
-      }
-    }
+  if (flags & PR_DISPLAY_FL_SEND_NOW) {
+    display_flush_lines(p, code, flags);
   }
 
   destroy_pool(p);
@@ -388,15 +434,30 @@ int pr_display_file(const char *path, const char *fs, const char *code,
     int flags) {
   pr_fh_t *fh = NULL;
   int res, xerrno;
+  struct stat st;
 
-  if (!path || !code) {
+  if (path == NULL ||
+      code == NULL) {
     errno = EINVAL;
     return -1;
   }
 
   fh = pr_fsio_open_canon(path, O_RDONLY);
-  if (fh == NULL)
+  if (fh == NULL) {
     return -1;
+  }
+
+  res = pr_fsio_fstat(fh, &st);
+  if (res < 0) {
+    pr_fsio_close(fh);
+    return -1;
+  }
+
+  if (S_ISDIR(st.st_mode)) {
+    pr_fsio_close(fh);
+    errno = EISDIR;
+    return -1;
+  }
 
   res = display_fh(fh, fs, code, flags);
   xerrno = errno;
