@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2004-2009 The ProFTPD Project team
+ * Copyright (c) 2004-2013 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -23,7 +23,7 @@
  */
 
 /* Configuration parser
- * $Id: parser.c,v 1.19 2009/03/05 06:01:51 castaglia Exp $
+ * $Id: parser.c,v 1.40 2013-12-30 06:38:59 castaglia Exp $
  */
 
 #include "conf.h"
@@ -38,7 +38,7 @@ static config_rec **parser_curr_config = NULL;
 
 static array_header *parser_servstack = NULL;
 static server_rec **parser_curr_server = NULL;
-static unsigned int parser_sid = 0;
+static unsigned int parser_sid = 1;
 
 static xaset_t **parser_server_list = NULL;
 
@@ -65,10 +65,10 @@ static struct config_src *parser_sources = NULL;
  */
 
 static void add_config_ctxt(config_rec *c) {
-  if (!*parser_curr_config)
+  if (!*parser_curr_config) {
     *parser_curr_config = c;
 
-  else {
+  } else {
     parser_curr_config = (config_rec **) push_array(parser_confstack);
     *parser_curr_config = c;
   }
@@ -84,10 +84,10 @@ static struct config_src *add_config_source(pr_fh_t *fh) {
   cs->cs_fh = fh;
   cs->cs_lineno = 0;
 
-  if (!parser_sources)
+  if (!parser_sources) {
     parser_sources = cs;
 
-  else {
+  } else {
     cs->cs_next = parser_sources;
     parser_sources = cs;
   }
@@ -96,23 +96,75 @@ static struct config_src *add_config_source(pr_fh_t *fh) {
 }
 
 static char *get_config_word(pool *p, char *word) {
+  size_t wordlen;
 
   /* Should this word be replaced with a value from the environment?
    * If so, tmp will contain the expanded value, otherwise tmp will
    * contain a string duped from the given pool.
    */
 
-  /* Does the given word use the environment syntax? */
-  if (strlen(word) > 7 &&
-      strncmp(word, "%{env:", 6) == 0 &&
-      word[strlen(word)-1] == '}') {
-    char *env;
+  wordlen = strlen(word);
 
-    word[strlen(word)-1] = '\0';
+  if (wordlen > 7) {
+    char *ptr = NULL;
 
-    env = pr_env_get(p, word + 6);
+    /* Does the given word use the environment syntax?
+     *
+     * In the simple (and most common) case, the entire word is the variable
+     * syntax.  But we also need to check for cases where the environment
+     * variable syntax is embedded within the word string.
+     */
 
-    return env ? pstrdup(p, env) : "";
+    if (strncmp(word, "%{env:", 6) == 0 &&
+        word[wordlen-1] == '}') {
+      char *env;
+
+      word[wordlen-1] = '\0';
+
+      env = pr_env_get(p, word + 6);
+
+      return env ? pstrdup(p, env) : "";
+    }
+
+    /* This is in a while loop in order to handle a) multiple different
+     * variables, and b) cases where the substituted value is itself a
+     * variable.   (Hopefully no one is so clever as to want to actually
+     * _use_ the latter approach.)
+     */
+    ptr = strstr(word, "%{env:");
+    while (ptr != NULL) {
+      char *env, *key, *ptr2, *var;
+      unsigned int keylen;
+
+      pr_signals_handle();
+
+      ptr2 = strchr(ptr + 6, '}');
+      if (ptr2 == NULL) {
+        /* No terminating marker; continue on to the next potential
+         * variable in the word.
+         */
+        ptr2 = ptr + 6;
+        ptr = strstr(ptr2, "%{env:");
+        continue;
+      }
+
+      keylen = (ptr2 - ptr - 6);
+      var = pstrndup(p, ptr, (ptr2 - ptr) + 1);
+
+      key = pstrndup(p, ptr + 6, keylen);
+
+      env = pr_env_get(p, key);
+      if (env == NULL) {
+        /* No value in the environment; continue on to the next potential
+         * variable in the word.
+         */
+        ptr = strstr(ptr2, "%{env:");
+        continue;
+      }
+
+      word = sreplace(p, word, var, env, NULL);
+      ptr = strstr(word, "%{env:");
+    }
   }
 
   return pstrdup(p, word);
@@ -151,7 +203,7 @@ int pr_parser_cleanup(void) {
   parser_curr_config = NULL;
 
   /* Reset the SID counter. */
-  parser_sid = 0;
+  parser_sid = 1;
 
   return 0;
 }
@@ -165,26 +217,31 @@ config_rec *pr_parser_config_ctxt_close(int *empty) {
    */
 
   if (parser_curr_config == (config_rec **) parser_confstack->elts) {
-    if (!c->subset || !c->subset->xas_list) {
+    if (c != NULL &&
+        (!c->subset || !c->subset->xas_list)) {
       xaset_remove(c->set, (xasetmember_t *) c);
       destroy_pool(c->pool);
 
-      if (empty)
+      if (empty) {
         *empty = TRUE;
+      }
     }
 
-    if (*parser_curr_config)
+    if (*parser_curr_config) {
       *parser_curr_config = NULL;
+    }
 
     return NULL;
   }
 
-  if (!c->subset || !c->subset->xas_list) {
+  if (c != NULL &&
+      (!c->subset || !c->subset->xas_list)) {
     xaset_remove(c->set, (xasetmember_t *) c);
     destroy_pool(c->pool);
 
-    if (empty)
+    if (empty) {
       *empty = TRUE;
+    }
   }
 
   parser_curr_config--;
@@ -227,7 +284,7 @@ config_rec *pr_parser_config_ctxt_open(const char *name) {
    * parent server.  This keeps <Global> config recs from being freed
    * prematurely, and helps to avoid memory leaks.
    */
-  if (strcmp(name, "<Global>") == 0) {
+  if (strncasecmp(name, "<Global>", 9) == 0) {
     if (!global_config_pool) {
       global_config_pool = make_sub_pool(permanent_pool);
       pr_pool_tag(global_config_pool, "<Global> Pool");
@@ -276,7 +333,7 @@ int pr_parser_parse_file(pool *p, const char *path, config_rec *start,
   pool *tmp_pool;
   char *report_path;
 
-  if (!path) {
+  if (path == NULL) {
     errno = EINVAL;
     return -1;
   }
@@ -293,13 +350,44 @@ int pr_parser_parse_file(pool *p, const char *path, config_rec *start,
 
   fh = pr_fsio_open(path, O_RDONLY);
   if (fh == NULL) {
+    int xerrno = errno;
+
     destroy_pool(tmp_pool);
+
+    errno = xerrno;
     return -1;
   }
 
   /* Stat the opened file to determine the optimal buffer size for IO. */
   memset(&st, 0, sizeof(st));
-  pr_fsio_fstat(fh, &st);
+  if (pr_fsio_fstat(fh, &st) < 0) {
+    int xerrno = errno;
+
+    pr_fsio_close(fh);
+    destroy_pool(tmp_pool);
+
+    errno = xerrno;
+    return -1;
+  }
+
+  if (S_ISDIR(st.st_mode)) {
+    pr_fsio_close(fh);
+    destroy_pool(tmp_pool);
+
+    errno = EISDIR;
+    return -1;
+  }
+
+  /* Check for world-writable files (and later, files in world-writable
+   * directories).
+   *
+   * For now, just warn about these; later, we will be more draconian.
+   */
+  if (st.st_mode & S_IWOTH) {
+    pr_log_pri(PR_LOG_WARNING, "warning: config file '%s' is world-writable",
+     path); 
+  }
+
   fh->fh_iosz = st.st_blksize;
 
   /* Push the configuration information onto the stack of configuration
@@ -339,18 +427,20 @@ int pr_parser_parse_file(pool *p, const char *path, config_rec *start,
           if (MODRET_ISERROR(mr)) {
 
             if (!(flags & PR_PARSER_FL_DYNAMIC_CONFIG)) {
-              pr_log_pri(PR_LOG_ERR, "Fatal: %s on line %u of '%s'",
+              pr_log_pri(PR_LOG_WARNING, "fatal: %s on line %u of '%s'",
                 MODRET_ERRMSG(mr), cs->cs_lineno, report_path);
               exit(1);
 
-            } else
+            } else {
               pr_log_pri(PR_LOG_WARNING, "warning: %s on line %u of '%s'",
                 MODRET_ERRMSG(mr), cs->cs_lineno, report_path);
+            }
           }
         }
 
-        if (!MODRET_ISDECLINED(mr))
+        if (!MODRET_ISDECLINED(mr)) {
           found = TRUE;
+        }
 
         conftab = pr_stash_get_symbol(PR_SYM_CONF, cmd->argv[0], conftab,
           &cmd->stash_index);
@@ -362,15 +452,16 @@ int pr_parser_parse_file(pool *p, const char *path, config_rec *start,
       if (!found) {
 
         if (!(flags & PR_PARSER_FL_DYNAMIC_CONFIG)) {
-          pr_log_pri(PR_LOG_ERR, "Fatal: unknown configuration directive "
+          pr_log_pri(PR_LOG_WARNING, "fatal: unknown configuration directive "
             "'%s' on line %u of '%s'", cmd->argv[0], cs->cs_lineno,
             report_path);
           exit(1);
 
-        } else 
+        } else {
           pr_log_pri(PR_LOG_WARNING, "warning: unknown configuration directive "
             "'%s' on line %u of '%s'", cmd->argv[0], cs->cs_lineno,
             report_path);
+        }
       }
     }
 
@@ -387,18 +478,19 @@ int pr_parser_parse_file(pool *p, const char *path, config_rec *start,
 }
 
 cmd_rec *pr_parser_parse_line(pool *p) {
-  char buf[PR_TUNABLE_BUFFER_SIZE], *word = NULL;
+  register unsigned int i;
+  char buf[PR_TUNABLE_BUFFER_SIZE+1], *arg = "", *word = NULL;
   cmd_rec *cmd = NULL;
   pool *sub_pool = NULL;
   array_header *arr = NULL;
 
-  if (!p) {
+  if (p == NULL) {
     errno = EINVAL;
     return NULL;
   }
 
   memset(buf, '\0', sizeof(buf));
-
+  
   while (pr_parser_read_line(buf, sizeof(buf)-1) != NULL) {
     char *bufp = buf;
 
@@ -415,7 +507,9 @@ cmd_rec *pr_parser_parse_line(pool *p) {
     /* Add each word to the array */
     arr = make_array(cmd->pool, 4, sizeof(char **));
     while ((word = pr_str_get_word(&bufp, 0)) != NULL) {
-      char *tmp = get_config_word(cmd->pool, word);
+      char *tmp;
+
+      tmp = get_config_word(cmd->pool, word);
 
       *((char **) push_array(arr)) = tmp;
       cmd->argc++;
@@ -448,19 +542,30 @@ cmd_rec *pr_parser_parse_line(pool *p) {
       if (*(cp + strlen(cp)-1) == '>' &&
           cmd->argc > 1) {
 
-        if (strcmp(cp, ">") == 0) {
+        if (strncmp(cp, ">", 2) == 0) {
           cmd->argv[cmd->argc-1] = NULL;
           cmd->argc--;
 
-        } else
+        } else {
           *(cp + strlen(cp)-1) = '\0';
+        }
 
         cp = cmd->argv[0];
-        if (*(cp + strlen(cp)-1) != '>')
+        if (*(cp + strlen(cp)-1) != '>') {
           cmd->argv[0] = pstrcat(cmd->pool, cp, ">", NULL);
+        }
       }
     }
 
+    if (cmd->argc < 2) {
+      arg = pstrdup(cmd->pool, arg);
+    }
+
+    for (i = 1; i < cmd->argc; i++) {
+      arg = pstrcat(cmd->pool, arg, *arg ? " " : "", cmd->argv[i], NULL);
+    }
+
+    cmd->arg = arg;
     return cmd;
   }
 
@@ -478,11 +583,12 @@ int pr_parser_prepare(pool *p, xaset_t **parsed_servers) {
     p = parser_pool;
   }
 
-  if (!parsed_servers)
+  if (parsed_servers == NULL) {
     parser_server_list = &server_list;
 
-  else
+  } else {
     parser_server_list = parsed_servers;
+  }
 
   parser_servstack = make_array(p, 1, sizeof(server_rec *));
   parser_curr_server = (server_rec **) push_array(parser_servstack);
@@ -533,15 +639,15 @@ char *pr_parser_read_line(char *buf, size_t bufsz) {
     if (buflen &&
         buf[buflen - 1] == '\n') {
       have_eol = TRUE;
-      buf[buflen - 1] = '\0';
-      buflen = strlen(buf);
+      buf[buflen-1] = '\0';
+      buflen--;
     }
 
     while (buflen &&
            buf[buflen - 1] == '\r') {
       pr_signals_handle();
-      buf[buflen - 1] = '\0';
-      buflen = strlen(buf);
+      buf[buflen-1] = '\0';
+      buflen--;
     }
 
     if (!have_eol) {
@@ -551,7 +657,7 @@ char *pr_parser_read_line(char *buf, size_t bufsz) {
     }
 
     /* Advance past any leading whitespace. */
-    for (bufp = buf; *bufp && isspace((int) *bufp); bufp++);
+    for (bufp = buf; *bufp && PR_ISSPACE(*bufp); bufp++);
 
     /* Check for commented or blank lines at this point, and just continue on
      * to the next configuration line if found.  If not, return the
@@ -611,6 +717,14 @@ server_rec *pr_parser_server_ctxt_open(const char *addrstr) {
   s->pool = p;
   s->config_type = CONF_VIRTUAL;
   s->sid = ++parser_sid;
+  s->notes = pr_table_nalloc(p, 0, 8);
+
+  /* TCP KeepAlive is enabled by default, with the system defaults. */
+  s->tcp_keepalive = palloc(s->pool, sizeof(struct tcp_keepalive));
+  s->tcp_keepalive->keepalive_enabled = TRUE;
+  s->tcp_keepalive->keepalive_idle = -1;
+  s->tcp_keepalive->keepalive_count = -1;
+  s->tcp_keepalive->keepalive_intvl = -1;
 
   /* Have to make sure it ends up on the end of the chain, otherwise
    * main_server becomes useless.
