@@ -1,6 +1,6 @@
 /*
  * ProFTPD - FTP server daemon
- * Copyright (c) 2007 The ProFTPD Project team
+ * Copyright (c) 2007-2012 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, The ProFTPD Project team and other respective
  * copyright holders give permission to link this program with OpenSSL, and
@@ -22,9 +22,8 @@
  * OpenSSL in the source distribution.
  */
 
-/*
- * Proctitle management
- * $Id: proctitle.c,v 1.5 2007/07/19 18:12:42 castaglia Exp $
+/* Proctitle management
+ * $Id: proctitle.c,v 1.13 2012-04-15 18:04:15 castaglia Exp $
  */
 
 #include "conf.h"
@@ -59,10 +58,15 @@ static char *prog_last_argv = NULL;
 #endif /* PR_DEVEL_STACK_TRACE */
 
 static int prog_argc = -1;
+static char proc_title_buf[BUFSIZ];
+
+static unsigned int proc_flags = 0;
+#define PR_PROCTITLE_FL_USE_STATIC		0x001
 
 void pr_proctitle_init(int argc, char *argv[], char *envp[]) {
 #ifndef PR_DEVEL_STACK_TRACE
-  register int i, envpsize;
+  register int i;
+  register size_t envpsize;
   char **p;
 
   /* Move the environment so setproctitle can use the space. */
@@ -72,9 +76,19 @@ void pr_proctitle_init(int argc, char *argv[], char *envp[]) {
   if ((p = (char **) malloc((i + 1) * sizeof(char *))) != NULL) {
     environ = p;
 
-    for (i = 0; envp[i] != NULL; i++)
-      if ((environ[i] = malloc(strlen(envp[i]) + 1)) != NULL)
-        strcpy(environ[i], envp[i]);
+    for (i = 0; envp[i] != NULL; i++) {
+      size_t envp_len = strlen(envp[i]);
+
+      if (envp_len > PR_TUNABLE_ENV_MAX) {
+        /* Skip any environ variables that are too long. */
+        continue;
+      }
+
+      environ[i] = malloc(envp_len + 1);
+      if (environ[i] != NULL) {
+        sstrncpy(environ[i], envp[i], envp_len + 1);
+      }
+    }
 
     environ[i] = NULL;
   }
@@ -99,7 +113,13 @@ void pr_proctitle_init(int argc, char *argv[], char *envp[]) {
   __progname = strdup("proftpd");
   __progname_full = strdup(argv[0]);
 # endif /* HAVE___PROGNAME */
+#else
+  /* Silence compiler warning about unused variable when stacktrace
+   * developer mode is configured.
+   */
+  prog_argc = -1;
 #endif /* !PR_DEVEL_STACK_TRACE */
+  memset(proc_title_buf, '\0', sizeof(proc_title_buf));
 }
 
 void pr_proctitle_free(void) {
@@ -124,45 +144,103 @@ void pr_proctitle_free(void) {
 #endif /* PR_USE_DEVEL */
 }
 
+void pr_proctitle_set_str(const char *str) {
+#ifndef PR_DEVEL_STACK_TRACE
+
+# ifndef HAVE_SETPROCTITLE
+  char *p;
+  int i, procbuflen, maxlen = (prog_last_argv - prog_argv[0]) - 2;
+
+#  if PF_ARGV_TYPE == PF_ARGV_PSTAT
+  union pstun pst;
+#  endif /* PF_ARGV_PSTAT */
+
+  if (proc_flags & PR_PROCTITLE_FL_USE_STATIC) {
+    return;
+  }
+
+  sstrncpy(proc_title_buf, str, sizeof(proc_title_buf));
+  procbuflen = strlen(proc_title_buf);
+
+#  if PF_ARGV_TYPE == PF_ARGV_NEW
+  /* We can just replace argv[] arguments.  Nice and easy. */
+  prog_argv[0] = proc_title_buf;
+  for (i = 1; i < prog_argc; i++) {
+    prog_argv[i] = "";
+  }
+#  endif /* PF_ARGV_NEW */
+
+#  if PF_ARGV_TYPE == PF_ARGV_WRITEABLE
+  /* We can overwrite individual argv[] arguments.  Semi-nice. */
+  snprintf(prog_argv[0], maxlen, "%s", proc_title_buf);
+  p = &prog_argv[0][procbuflen];
+
+  while (p < prog_last_argv)
+    *p++ = '\0';
+
+  for (i = 1; i < prog_argc; i++) {
+    prog_argv[i] = "";
+  }
+
+#  endif /* PF_ARGV_WRITEABLE */
+
+#  if PF_ARGV_TYPE == PF_ARGV_PSSTRINGS
+  PS_STRINGS->ps_nargvstr = 1;
+  PS_STRINGS->ps_argvstr = proc_title_buf;
+#  endif /* PF_ARGV_PSSTRINGS */
+
+# else
+  if (proc_flags & PR_PROCTITLE_FL_USE_STATIC) {
+    return;
+  }
+
+  setproctitle("%s", str);
+# endif /* HAVE_SETPROCTITLE */
+#endif /* PR_DEVEL_STACK_TRACE */
+}
+
 void pr_proctitle_set(const char *fmt, ...) {
 #ifndef PR_DEVEL_STACK_TRACE
   va_list msg;
-  static char statbuf[BUFSIZ];
 
 # ifndef HAVE_SETPROCTITLE
 #  if PF_ARGV_TYPE == PF_ARGV_PSTAT
   union pstun pst;
 #  endif /* PF_ARGV_PSTAT */
   char *p;
-  int i, statbuflen, maxlen = (prog_last_argv - prog_argv[0]) - 2;
+  int i, procbuflen, maxlen = (prog_last_argv - prog_argv[0]) - 2;
 # endif /* HAVE_SETPROCTITLE */
 
+  if (proc_flags & PR_PROCTITLE_FL_USE_STATIC) {
+    return;
+  }
+
   if (!fmt)
     return;
 
   va_start(msg, fmt);
 
-  memset(statbuf, 0, sizeof(statbuf));
+  memset(proc_title_buf, 0, sizeof(proc_title_buf));
 
 # ifdef HAVE_SETPROCTITLE
 #  if __FreeBSD__ >= 4 && !defined(FREEBSD4_0) && !defined(FREEBSD4_1)
   /* FreeBSD's setproctitle() automatically prepends the process name. */
-  vsnprintf(statbuf, sizeof(statbuf), fmt, msg);
+  vsnprintf(proc_title_buf, sizeof(proc_title_buf), fmt, msg);
 
 #  else /* FREEBSD4 */
   /* Manually append the process name for non-FreeBSD platforms. */
-  snprintf(statbuf, sizeof(statbuf), "%s", "proftpd: ");
-  vsnprintf(statbuf + strlen(statbuf), sizeof(statbuf) - strlen(statbuf),
-    fmt, msg);
+  snprintf(proc_title_buf, sizeof(proc_title_buf), "%s", "proftpd: ");
+  vsnprintf(proc_title_buf + strlen(proc_title_buf),
+    sizeof(proc_title_buf) - strlen(proc_title_buf), fmt, msg);
 
 #  endif /* FREEBSD4 */
-  setproctitle("%s", statbuf);
+  setproctitle("%s", proc_title_buf);
 
 # else /* HAVE_SETPROCTITLE */
   /* Manually append the process name for non-setproctitle() platforms. */
-  snprintf(statbuf, sizeof(statbuf), "%s", "proftpd: ");
-  vsnprintf(statbuf + strlen(statbuf), sizeof(statbuf) - strlen(statbuf),
-    fmt, msg);
+  snprintf(proc_title_buf, sizeof(proc_title_buf), "%s", "proftpd: ");
+  vsnprintf(proc_title_buf + strlen(proc_title_buf),
+    sizeof(proc_title_buf) - strlen(proc_title_buf), fmt, msg);
 
 # endif /* HAVE_SETPROCTITLE */
 
@@ -171,11 +249,11 @@ void pr_proctitle_set(const char *fmt, ...) {
 # ifdef HAVE_SETPROCTITLE
   return;
 # else
-  statbuflen = strlen(statbuf);
+  procbuflen = strlen(proc_title_buf);
 
 #  if PF_ARGV_TYPE == PF_ARGV_NEW
   /* We can just replace argv[] arguments.  Nice and easy. */
-  prog_argv[0] = statbuf;
+  prog_argv[0] = proc_title_buf;
   for (i = 1; i < prog_argc; i++) {
     prog_argv[i] = "";
   }
@@ -183,8 +261,8 @@ void pr_proctitle_set(const char *fmt, ...) {
 
 #  if PF_ARGV_TYPE == PF_ARGV_WRITEABLE
   /* We can overwrite individual argv[] arguments.  Semi-nice. */
-  snprintf(prog_argv[0], maxlen, "%s", statbuf);
-  p = &prog_argv[0][statbuflen];
+  snprintf(prog_argv[0], maxlen, "%s", proc_title_buf);
+  p = &prog_argv[0][procbuflen];
 
   while (p < prog_last_argv)
     *p++ = '\0';
@@ -196,15 +274,42 @@ void pr_proctitle_set(const char *fmt, ...) {
 #  endif /* PF_ARGV_WRITEABLE */
 
 #  if PF_ARGV_TYPE == PF_ARGV_PSTAT
-  pst.pst_command = statbuf;
-  pstat(PSTAT_SETCMD, pst, statbuflen, 0, 0);
+  pst.pst_command = proc_title_buf;
+  pstat(PSTAT_SETCMD, pst, procbuflen, 0, 0);
+
 #  endif /* PF_ARGV_PSTAT */
 
 #  if PF_ARGV_TYPE == PF_ARGV_PSSTRINGS
   PS_STRINGS->ps_nargvstr = 1;
-  PS_STRINGS->ps_argvstr = statbuf;
+  PS_STRINGS->ps_argvstr = proc_title_buf;
 #  endif /* PF_ARGV_PSSTRINGS */
 
 # endif /* HAVE_SETPROCTITLE */
 #endif /* !PR_DEVEL_STACK_TRACE */
 }
+
+void pr_proctitle_set_static_str(const char *buf) {
+  if (buf != NULL) {
+    pr_proctitle_set_str(buf);
+    proc_flags |= PR_PROCTITLE_FL_USE_STATIC;
+
+  } else {
+    /* Reset. */
+    proc_flags = 0;
+  }
+}
+
+int pr_proctitle_get(char *buf, size_t bufsz) {
+  if (buf == NULL ||
+      bufsz == 0) {
+
+    /* If the caller didn't provide an input buffer, they are simply
+     * querying for the length of the current process title.  Easy enough
+     * to acquiesce to that request.
+     */
+    return strlen(proc_title_buf);
+  }
+
+  sstrncpy(buf, proc_title_buf, bufsz);
+  return strlen(buf);
+}
