@@ -2,7 +2,7 @@
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
  * Copyright (c) 1999, 2000 MacGyver aka Habeeb J. Dihu <macgyver@tos.net>
- * Copyright (c) 2001-2010 The ProFTPD Project team
+ * Copyright (c) 2001-2013 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -27,25 +27,11 @@
 /* Various basic support routines for ProFTPD, used by all modules
  * and not specific to one or another.
  *
- * $Id: support.c,v 1.104.2.1 2010/06/15 16:58:38 castaglia Exp $
+ * $Id: support.c,v 1.120 2013-08-07 16:35:27 castaglia Exp $
  */
 
 #include "conf.h"
 
-#include <signal.h>
-
-#ifdef HAVE_SYS_STATVFS_H
-# include <sys/statvfs.h>
-#elif defined(HAVE_SYS_VFS_H)
-# include <sys/vfs.h>
-#elif defined(HAVE_SYS_MOUNT_H)
-# include <sys/mount.h>
-#endif
-
-#ifdef AIX3
-# include <sys/statfs.h>
-#endif
-
 #ifdef PR_USE_OPENSSL
 # include <openssl/crypto.h>
 #endif /* PR_USE_OPENSSL */
@@ -88,10 +74,16 @@ static void mask_signals(unsigned char block) {
 #endif
     sigaddset(&mask_sigset, SIGHUP);
 
-    sigprocmask(SIG_BLOCK, &mask_sigset, NULL);
+    if (sigprocmask(SIG_BLOCK, &mask_sigset, NULL) < 0) {
+      pr_log_pri(PR_LOG_NOTICE,
+        "unable to block signal set: %s", strerror(errno));
+    }
 
   } else {
-    sigprocmask(SIG_UNBLOCK, &mask_sigset, NULL);
+    if (sigprocmask(SIG_UNBLOCK, &mask_sigset, NULL) < 0) {
+      pr_log_pri(PR_LOG_NOTICE,
+        "unable to unblock signal set: %s", strerror(errno));
+    }
   }
 }
 
@@ -131,15 +123,17 @@ void schedule(void (*f)(void*,void*,void*,void*),int nloops, void *a1,
   pool *p, *sub_pool;
   sched_t *s;
 
-  if (!scheds) {
-   p = make_sub_pool(permanent_pool);
-   pr_pool_tag(p, "Schedules Pool");
-   scheds = xaset_create(p, NULL);
+  if (scheds == NULL) {
+    p = make_sub_pool(permanent_pool);
+    pr_pool_tag(p, "Schedules Pool");
+    scheds = xaset_create(p, NULL);
 
-  } else
-   p = scheds->pool;
+  } else {
+    p = scheds->pool;
+  }
 
   sub_pool = make_sub_pool(p);
+  pr_pool_tag(sub_pool, "schedule pool");
 
   s = pcalloc(sub_pool, sizeof(sched_t));
   s->pool = sub_pool;
@@ -149,21 +143,23 @@ void schedule(void (*f)(void*,void*,void*,void*),int nloops, void *a1,
   s->a3 = a3;
   s->a4 = a4;
   s->loops = nloops;
-  xaset_insert(scheds, (xasetmember_t*)s);
+  xaset_insert(scheds, (xasetmember_t *) s);
 }
 
 void run_schedule(void) {
-  sched_t *s,*snext;
+  sched_t *s, *snext;
 
-  if (!scheds || !scheds->xas_list)
+  if (scheds == NULL ||
+      scheds->xas_list == NULL) {
     return;
+  }
 
-  for (s = (sched_t*)scheds->xas_list; s; s=snext) {
+  for (s = (sched_t *) scheds->xas_list; s; s = snext) {
     snext = s->next;
 
     if (s->loops-- <= 0) {
-      s->f(s->a1,s->a2,s->a3,s->a4);
-      xaset_remove(scheds, (xasetmember_t*)s);
+      s->f(s->a1, s->a2, s->a3, s->a4);
+      xaset_remove(scheds, (xasetmember_t *) s);
       destroy_pool(s->pool);
     }
   }
@@ -179,38 +175,38 @@ void run_schedule(void) {
  * Alas, current (Jul 2000) Linux systems define NAME_MAX anyway.
  * NB: NAME_MAX_GUESS is defined in support.h.
  */
-int get_name_max(char *dirname, int dir_fd) {
-  int name_max = 0;
+size_t get_name_max(char *dirname, int dir_fd) {
+  long res = 0;
 #if defined(HAVE_FPATHCONF) || defined(HAVE_PATHCONF)
   char *msgfmt = "";
 
 # if defined(HAVE_FPATHCONF)
   if (dir_fd > 0) {
-    name_max = fpathconf(dir_fd, _PC_NAME_MAX);
-    msgfmt = "fpathconf(%s, _PC_NAME_MAX) = %d, errno = %d";
+    res = fpathconf(dir_fd, _PC_NAME_MAX);
+    msgfmt = "fpathconf(%s, _PC_NAME_MAX) = %ld, errno = %d";
   } else
 # endif
 # if defined(HAVE_PATHCONF)
   if (dirname != NULL) {
-    name_max = pathconf(dirname, _PC_NAME_MAX);
-    msgfmt = "pathconf(%s, _PC_NAME_MAX) = %d, errno = %d";
+    res = pathconf(dirname, _PC_NAME_MAX);
+    msgfmt = "pathconf(%s, _PC_NAME_MAX) = %ld, errno = %d";
   } else
 # endif
   /* No data provided to use either pathconf() or fpathconf() */
   return -1;
 
-  if (name_max < 0) {
+  if (res < 0) {
     /* NB: errno may not be set if the failure is due to a limit or option
      * not being supported.
      */
-    pr_log_debug(DEBUG1, msgfmt, dirname ? dirname : "(NULL)", name_max, errno);
+    pr_log_debug(DEBUG1, msgfmt, dirname ? dirname : "(NULL)", res, errno);
   }
 
 #else
-  name_max = NAME_MAX_GUESS;
+  res = NAME_MAX_GUESS;
 #endif /* HAVE_FPATHCONF or HAVE_PATHCONF */
 
-  return name_max;
+  return (size_t) res;
 }
 
 
@@ -415,7 +411,7 @@ char *dir_abs_path(pool *p, const char *path, int interpolate) {
  * PATH, or 0 if it doesn't exist. Catch symlink loops using LAST_INODE and
  * RCOUNT.
  */
-static mode_t _symlink(char *path, ino_t last_inode, int rcount) {
+static mode_t _symlink(const char *path, ino_t last_inode, int rcount) {
   char buf[PR_TUNABLE_PATH_MAX + 1];
   struct stat sbuf;
   int i;
@@ -438,15 +434,17 @@ static mode_t _symlink(char *path, ino_t last_inode, int rcount) {
       return 0;
     }
 
-    if (S_ISLNK(sbuf.st_mode))
+    if (S_ISLNK(sbuf.st_mode)) {
       return _symlink(buf, (ino_t) sbuf.st_ino, rcount);
+    }
+
     return sbuf.st_mode;
   }
 
   return 0;
 }
 
-mode_t file_mode(char *path) {
+mode_t file_mode(const char *path) {
   struct stat sbuf;
   mode_t res = 0;
 
@@ -455,12 +453,14 @@ mode_t file_mode(char *path) {
     if (S_ISLNK(sbuf.st_mode)) {
       res = _symlink(path, (ino_t) 0, 0);
 
-      if (res == 0)
+      if (res == 0) {
 	/* a dangling symlink, but it exists to rename or delete. */
 	res = sbuf.st_mode;
+      }
 
-    } else
+    } else {
       res = sbuf.st_mode;
+    }
   }
 
   return res;
@@ -471,15 +471,19 @@ mode_t file_mode(char *path) {
  * If DIRP == -1, fail unless PATH exists; the caller doesn't care whether
  * PATH is a file or a directory.
  */
-static int _exists(char *path, int dirp) {
+static int _exists(const char *path, int dirp) {
   mode_t fmode;
 
-  if ((fmode = file_mode(path)) != 0) {
-    if (dirp == 1 && !S_ISDIR(fmode))
+  fmode = file_mode(path);
+  if (fmode != 0) {
+    if (dirp == 1 &&
+        !S_ISDIR(fmode)) {
       return FALSE;
 
-    else if (dirp == 0 && S_ISDIR(fmode))
+    } else if (dirp == 0 &&
+               S_ISDIR(fmode)) {
       return FALSE;
+    }
 
     return TRUE;
   }
@@ -487,15 +491,15 @@ static int _exists(char *path, int dirp) {
   return FALSE;
 }
 
-int file_exists(char *path) {
+int file_exists(const char *path) {
   return _exists(path, 0);
 }
 
-int dir_exists(char *path) {
+int dir_exists(const char *path) {
   return _exists(path, 1);
 }
 
-int exists(char *path) {
+int exists(const char *path) {
   return _exists(path, -1);
 }
 
@@ -510,19 +514,19 @@ char *safe_token(char **s) {
   if (!s || !*s)
     return res;
 
-  while (isspace((int) **s) && **s)
+  while (PR_ISSPACE(**s) && **s)
     (*s)++;
 
   if (**s) {
     res = *s;
 
-    while (!isspace((int) **s) && **s)
+    while (!PR_ISSPACE(**s) && **s)
       (*s)++;
 
     if (**s)
       *(*s)++ = '\0';
 
-    while (isspace((int) **s) && **s)
+    while (PR_ISSPACE(**s) && **s)
       (*s)++;
   }
 
@@ -600,6 +604,7 @@ int check_shutmsg(time_t *shut, time_t *deny, time_t *disc, char *msg,
   return 0;
 }
 
+#if !defined(PR_USE_OPENSSL) || OPENSSL_VERSION_NUMBER <= 0x000907000L
 /* "safe" memset() (code borrowed from OpenSSL).  This function should be
  * used to clear/scrub sensitive memory areas instead of memset() for the
  * reasons mentioned in this BugTraq thread:
@@ -607,7 +612,8 @@ int check_shutmsg(time_t *shut, time_t *deny, time_t *disc, char *msg,
  *  http://online.securityfocus.com/archive/1/298598
  */
 
-unsigned char memscrub_ctr = 0;
+static unsigned char memscrub_ctr = 0;
+#endif
 
 void pr_memscrub(void *ptr, size_t ptrlen) {
 #if defined(PR_USE_OPENSSL) && OPENSSL_VERSION_NUMBER > 0x000907000L
@@ -643,6 +649,30 @@ void pr_memscrub(void *ptr, size_t ptrlen) {
 #endif
 }
 
+void pr_getopt_reset(void) {
+#if defined(FREEBSD4) || defined(FREEBSD5) || defined(FREEBSD6) || \
+    defined(FREEBSD7) || defined(FREEBSD8) || defined(FREEBSD9) || \
+    defined(FREEBSD10) || \
+    defined(DARWIN7) || defined(DARWIN8) || defined(DARWIN9) || \
+    defined(DARWIN10) || defined(DARWIN11)
+  optreset = 1;
+  opterr = 1;
+  optind = 1;
+
+#elif defined(SOLARIS2) || defined(HPUX11)
+  opterr = 0;
+  optind = 1;
+
+#else
+  opterr = 0;
+  optind = 0;
+#endif /* !FreeBSD, !Mac OSX and !Solaris2 */
+
+  if (pr_env_get(permanent_pool, "POSIXLY_CORRECT") == NULL) {
+    pr_env_set(permanent_pool, "POSIXLY_CORRECT", "1");
+  }
+}
+
 struct tm *pr_gmtime(pool *p, const time_t *t) {
   struct tm *sys_tm, *dup_tm;
 
