@@ -1,7 +1,7 @@
 /*
  * ProFTPD - FTP server daemon
  * Copyright (c) 1997, 1998 Public Flood Software
- * Copyright (c) 2001, 2002, 2003, 2004 The ProFTPD Project team
+ * Copyright (c) 2001-2013 The ProFTPD Project team
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, Public Flood Software/MacGyver aka Habeeb J. Dihu
  * and other respective copyright holders give permission to link this program
@@ -23,15 +23,12 @@
  * the source code for OpenSSL in the source distribution.
  */
 
-/*
- * Timer system, based on alarm() and SIGALRM
- * $Id: timers.c,v 1.24 2004/11/20 22:35:46 castaglia Exp $
+/* Timer system, based on alarm() and SIGALRM
+ * $Id: timers.c,v 1.39 2013-10-09 06:36:20 castaglia Exp $
  */
 
 #include "conf.h"
 
-#include <signal.h>
-
 /* From src/main.c */
 volatile extern unsigned int recvd_signal_flags;
 
@@ -45,20 +42,26 @@ struct timer {
   module *mod;                  /* Module owning this timer */
   callback_t callback;          /* Function to callback */
   char remove;                  /* Internal use */
+
+  const char *desc;		/* Description of timer, provided by caller */
 };
 
+#define PR_TIMER_DYNAMIC_TIMERNO	1024
+
 static int _current_timeout = 0;
 static int _total_time = 0;
 static int _sleep_sem = 0;
-static int alarms_blocked = 0,alarm_pending = 0;
+static int alarms_blocked = 0, alarm_pending = 0;
 static xaset_t *timers = NULL;
 static xaset_t *recycled = NULL;
 static xaset_t *free_timers = NULL;
 static int _indispatch = 0;
-static int dynamic_timerno = 1024;
+static int dynamic_timerno = PR_TIMER_DYNAMIC_TIMERNO;
 static unsigned int nalarms = 0;
 static time_t _alarmed_time = 0;
 
+static pool *timer_pool = NULL;
+
 static int timer_cmp(struct timer *t1, struct timer *t2) {
   if (t1->count < t2->count)
     return -1;
@@ -78,10 +81,16 @@ static int process_timers(int elapsed) {
   struct timer *t = NULL, *next = NULL;
 
   if (!recycled)
-    recycled = xaset_create(NULL, NULL);
+    recycled = xaset_create(timer_pool, NULL);
+
+  if (!elapsed &&
+      !recycled->xas_list) {
+
+    if (!timers)
+      return 0;
 
-  if (!elapsed && !recycled->xas_list)
     return (timers->xas_list ? ((struct timer *) timers->xas_list)->count : 0);
+  }
 
   /* Critical code, no interruptions please */
   if (_indispatch)
@@ -91,7 +100,7 @@ static int process_timers(int elapsed) {
   _indispatch++;
 
   if (elapsed) {
-    for (t = (struct timer *) timers->xas_list; t; t=next) {
+    for (t = (struct timer *) timers->xas_list; t; t = next) {
       /* If this timer has already been handled, skip */
       next = t->next;
 
@@ -101,8 +110,15 @@ static int process_timers(int elapsed) {
         xaset_insert(free_timers, (xasetmember_t *) t);
 
       } else if ((t->count -= elapsed) <= 0) {
-
         /* This timer's interval has elapsed, so trigger its callback. */
+
+        pr_trace_msg("timer", 4,
+          "%ld %s for timer ID %d ('%s', for module '%s') elapsed, invoking "
+          "callback (%p)", t->interval,
+          t->interval != 1 ? "seconds" : "second", t->timerno,
+          t->desc ? t->desc : "<unknown>",
+          t->mod ? t->mod->name : "<none>", t->callback);
+
         if (t->callback(t->interval, t->timerno, t->interval - t->count,
             t->mod) == 0) {
 
@@ -116,6 +132,9 @@ static int process_timers(int elapsed) {
           /* A non-zero return value from a timer callback signals that
            * the timer should be reused/restarted.
            */
+          pr_trace_msg("timer", 6, "restarting timer ID %d ('%s'), as per "
+            "callback", t->timerno, t->desc ? t->desc : "<unknown>");
+
           xaset_remove(timers, (xasetmember_t *) t);
           t->count = t->interval;
           xaset_insert(recycled, (xasetmember_t *) t);
@@ -151,10 +170,17 @@ static RETSIGTYPE sig_alarm(int signo) {
 #endif
 
   /* Install this handler for SIGALRM. */
-  sigaction(SIGALRM, &act, NULL);
+  if (sigaction(SIGALRM, &act, NULL) < 0) {
+    pr_log_pri(PR_LOG_WARNING,
+      "unable to install SIGALRM handler via sigaction(2): %s",
+      strerror(errno));
+  }
 
 #ifdef HAVE_SIGINTERRUPT
-  siginterrupt(SIGALRM, 1);
+  if (siginterrupt(SIGALRM, 1) < 0) {
+    pr_log_pri(PR_LOG_WARNING,
+      "unable to allow SIGALRM to interrupt system calls: %s", strerror(errno));
+  }
 #endif
 
   recvd_signal_flags |= RECEIVED_SIG_ALRM;
@@ -179,10 +205,17 @@ static void set_sig_alarm(void) {
 #endif
 
   /* Install this handler for SIGALRM. */
-  sigaction(SIGALRM, &act, NULL);
+  if (sigaction(SIGALRM, &act, NULL) < 0) {
+    pr_log_pri(PR_LOG_WARNING,
+      "unable to install SIGALRM handler via sigaction(2): %s",
+      strerror(errno));
+  }
 
 #ifdef HAVE_SIGINTERRUPT
-  siginterrupt(SIGALRM, 1);
+  if (siginterrupt(SIGALRM, 1) < 0) {
+    pr_log_pri(PR_LOG_WARNING,
+      "unable to allow SIGALRM to interrupt system calls: %s", strerror(errno));
+  }
 #endif
 }
 
@@ -223,19 +256,27 @@ void handle_alarm(void) {
 int pr_timer_reset(int timerno, module *mod) {
   struct timer *t = NULL;
 
-  if (_indispatch)
+  if (!timers) {
+    errno = EPERM;
     return -1;
+  }
+
+  if (_indispatch) {
+    errno = EINTR;
+    return -1;
+  }
 
   pr_alarms_block();
 
   if (!recycled)
-    recycled = xaset_create(NULL, NULL);
+    recycled = xaset_create(timer_pool, NULL);
 
-  for (t = (struct timer *) timers->xas_list; t; t=t->next)
-    if (t->timerno == timerno && (t->mod == mod || mod == ANY_MODULE)) {
+  for (t = (struct timer *) timers->xas_list; t; t = t->next) {
+    if (t->timerno == timerno &&
+        (t->mod == mod || mod == ANY_MODULE)) {
       t->count = t->interval;
-      xaset_remove(timers, (xasetmember_t*)t);
-      xaset_insert(recycled, (xasetmember_t*)t);
+      xaset_remove(timers, (xasetmember_t *) t);
+      xaset_insert(recycled, (xasetmember_t *) t);
       nalarms++;
 
       /* The handle_alarm() function also readjusts the timers lists
@@ -245,14 +286,22 @@ int pr_timer_reset(int timerno, module *mod) {
       handle_alarm();
       break;
     }
+  }
 
   pr_alarms_unblock();
 
+  if (t) {
+    pr_trace_msg("timer", 7, "reset timer ID %d ('%s', for module '%s')",
+      t->timerno, t->desc, t->mod ? t->mod->name : "[none]");
+    return t->timerno;
+  }
+
   return (t ? t->timerno : 0);
 }
 
 int pr_timer_remove(int timerno, module *mod) {
-  struct timer *t = NULL;
+  struct timer *t = NULL, *tnext = NULL;
+  int nremoved = 0;
 
   /* If there are no timers currently registered, do nothing. */
   if (!timers)
@@ -260,14 +309,19 @@ int pr_timer_remove(int timerno, module *mod) {
 
   pr_alarms_block();
 
-  for (t = (struct timer *) timers->xas_list; t; t = t->next)
-    if (t->timerno == timerno && (t->mod == mod || mod == ANY_MODULE)) {
+  for (t = (struct timer *) timers->xas_list; t; t = tnext) {
+    tnext = t->next;
+
+    if ((timerno < 0 || t->timerno == timerno) &&
+        (mod == ANY_MODULE || t->mod == mod)) {
+      nremoved++;
+
       if (_indispatch) {
         t->remove++;
 
       } else {
-        xaset_remove(timers, (xasetmember_t*)t);
-        xaset_insert(free_timers, (xasetmember_t*)t);
+        xaset_remove(timers, (xasetmember_t *) t);
+        xaset_insert(free_timers, (xasetmember_t *) t);
 	nalarms++;
 
         /* The handle_alarm() function also readjusts the timers lists
@@ -276,27 +330,53 @@ int pr_timer_remove(int timerno, module *mod) {
          */
         handle_alarm();
       }
+
+      pr_trace_msg("timer", 7, "removed timer ID %d ('%s', for module '%s')",
+        t->timerno, t->desc, t->mod ? t->mod->name : "[none]");
+    }
+
+    /* If we are removing a specific timer, break out of the loop now.
+     * Otherwise, keep removing any matching timers.
+     */
+    if (nremoved > 0 &&
+        timerno >= 0) {
       break;
     }
+  }
 
   pr_alarms_unblock();
 
-  return (t ? t->timerno : 0);
+  if (nremoved == 0) {
+    errno = ENOENT;
+    return -1;
+  }
+
+  /* If we removed a specific timer because of the given timerno, return
+   * that timerno value.
+   */
+  if (timerno >= 0) {
+    return timerno;
+  }
+
+  return nremoved;
 }
 
-int pr_timer_add(int seconds, int timerno, module *mod, callback_t cb) {
+int pr_timer_add(int seconds, int timerno, module *mod, callback_t cb,
+    const char *desc) {
   struct timer *t = NULL;
 
-  if (seconds < 0) {
+  if (seconds <= 0 ||
+      cb == NULL ||
+      desc == NULL) {
     errno = EINVAL;
     return -1;
   }
 
   if (!timers)
-    timers = xaset_create(NULL, (XASET_COMPARE) timer_cmp);
+    timers = xaset_create(timer_pool, (XASET_COMPARE) timer_cmp);
 
   /* Check to see that, if specified, the timerno is not already in use. */
-  if (timerno != -1) {
+  if (timerno >= 0) {
     for (t = (struct timer *) timers->xas_list; t; t = t->next) {
       if (t->timerno == timerno) {
         errno = EPERM;
@@ -306,26 +386,31 @@ int pr_timer_add(int seconds, int timerno, module *mod, callback_t cb) {
   }
 
   if (!free_timers)
-    free_timers = xaset_create(NULL, NULL);
+    free_timers = xaset_create(timer_pool, NULL);
 
   /* Try to use an old timer first */
   pr_alarms_block();
   t = (struct timer *) free_timers->xas_list;
-  if (t != NULL)
+  if (t != NULL) {
     xaset_remove(free_timers, (xasetmember_t *) t);
 
-  else
-    /* XXX The Timer API uses the permanent pool when it should be using
-     * its own subpool.
-     */
+  } else {
+
+    if (timer_pool == NULL) {
+      timer_pool = make_sub_pool(permanent_pool);
+      pr_pool_tag(timer_pool, "Timer Pool");
+    }
 
     /* Must allocate a new one */
-    t = palloc(permanent_pool, sizeof(struct timer));
+    t = palloc(timer_pool, sizeof(struct timer));
+  }
 
-  if (timerno == -1) {
+  if (timerno < 0) {
     /* Dynamic timer */
-    if (dynamic_timerno < 1024)
-      dynamic_timerno = 1024;
+    if (dynamic_timerno < PR_TIMER_DYNAMIC_TIMERNO) {
+      dynamic_timerno = PR_TIMER_DYNAMIC_TIMERNO;
+    }
+
     timerno = dynamic_timerno++;
   }
 
@@ -334,6 +419,7 @@ int pr_timer_add(int seconds, int timerno, module *mod, callback_t cb) {
   t->callback = cb;
   t->mod = mod;
   t->remove = 0;
+  t->desc = desc;
 
   /* If called while _indispatch, add to the recycled list to prevent
    * list corruption
@@ -341,7 +427,7 @@ int pr_timer_add(int seconds, int timerno, module *mod, callback_t cb) {
 
   if (_indispatch) {
     if (!recycled)
-      recycled = xaset_create(NULL, NULL);
+      recycled = xaset_create(timer_pool, NULL);
     xaset_insert(recycled, (xasetmember_t *) t);
 
   } else {
@@ -358,6 +444,10 @@ int pr_timer_add(int seconds, int timerno, module *mod, callback_t cb) {
 
   pr_alarms_unblock();
 
+  pr_trace_msg("timer", 7, "added timer ID %d ('%s', for module '%s'), "
+    "triggering in %ld %s", t->timerno, t->desc,
+    t->mod ? t->mod->name : "[none]", t->interval,
+    t->interval != 1 ? "seconds" : "second");
   return timerno;
 }
 
@@ -396,7 +486,7 @@ int pr_timer_sleep(int seconds) {
     return -1;
   }
 
-  timerno = pr_timer_add(seconds, -1, NULL, sleep_cb);
+  timerno = pr_timer_add(seconds, -1, NULL, sleep_cb, "sleep");
   if (timerno == -1)
     return -1;
 
@@ -409,6 +499,24 @@ int pr_timer_sleep(int seconds) {
   return 0;
 }
 
+int pr_timer_usleep(unsigned long usecs) {
+  struct timeval tv;
+
+  if (usecs == 0) {
+    errno = EINVAL;
+    return -1;
+  }
+
+  tv.tv_sec = (usecs / 1000000);
+  tv.tv_usec = (usecs - (tv.tv_sec * 1000000));
+
+  pr_signals_block();
+  (void) select(0, NULL, NULL, NULL, &tv);
+  pr_signals_unblock();
+
+  return 0;
+}
+
 void timers_init(void) {
 
   /* Reset some of the key static variables. */
@@ -416,11 +524,19 @@ void timers_init(void) {
   _total_time = 0;
   nalarms = 0;
   _alarmed_time = 0;
+  dynamic_timerno = PR_TIMER_DYNAMIC_TIMERNO;
 
   /* Don't inherit the parent's timer lists. */
   timers = NULL;
   recycled = NULL;
   free_timers = NULL;
 
- return;
+  /* Reset the timer pool. */
+  if (timer_pool)
+    destroy_pool(timer_pool);
+
+  timer_pool = make_sub_pool(permanent_pool);
+  pr_pool_tag(timer_pool, "Timer Pool");
+
+  return;
 }
